// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use cpu::{
    config::CpuConfig,
    instruction::{
        AddSubMOp, AluCommonMOp, CommonMOp, LogicalMOp, MOp, MOpDestReg, MOpRegNum,
        OutputIntegerMode, COMMON_MOP_2_IMM_WIDTH, COMMON_MOP_3_IMM_WIDTH,
    },
    reg_alloc::{reg_alloc, FetchedDecodedMOp},
    unit::UnitKind,
};
use fayalite::{
    assert_export_firrtl,
    firrtl::ExportOptions,
    prelude::*,
    sim::{time::SimDuration, vcd::VcdWriterDecls, Simulation},
    util::RcWriter,
};
use std::num::NonZeroUsize;

#[hdl]
#[test]
fn test_reg_alloc() {
    let _n = SourceLocation::normalize_files_for_tests();
    let mut config = CpuConfig::new(vec![UnitKind::AluBranch, UnitKind::AluBranch]);
    config.fetch_width = NonZeroUsize::new(2).unwrap();
    let m = reg_alloc(&config);
    let mut sim = Simulation::new(m);
    let mut writer = RcWriter::default();
    sim.add_trace_writer(VcdWriterDecls::new(writer.clone()));
    let fetch_decode_interface = sim.io().fetch_decode_interface;
    sim.write_clock(sim.io().cd.clk, false);
    sim.write_reset(sim.io().cd.rst, true);
    sim.write_bool(fetch_decode_interface.fetch_decode_special_op.ready, true);
    sim.write(
        fetch_decode_interface.decoded_insns[0].data,
        HdlSome(
            #[hdl]
            FetchedDecodedMOp {
                mop: MOp.AluBranch(MOp.AluBranch.AddSub(
                    #[hdl]
                    AddSubMOp {
                        alu_common: #[hdl]
                        AluCommonMOp {
                            common: CommonMOp::new(
                                0_hdl_u0,
                                #[hdl]
                                MOpDestReg {
                                    normal_regs: #[hdl]
                                    [
                                        #[hdl]
                                        MOpRegNum { value: 1u8 },
                                        MOpRegNum::const_zero(),
                                    ],
                                    flag_regs: #[hdl]
                                    [HdlSome(()), HdlNone()],
                                },
                                [2u8, 3u8, 4u8],
                                0x1234.cast_to(SInt[COMMON_MOP_3_IMM_WIDTH]),
                            ),
                            output_integer_mode: OutputIntegerMode.Full64(),
                        },
                        invert_src0: true,
                        invert_carry_in: true,
                        invert_carry_out: true,
                        add_pc: true,
                    },
                )),
                is_unrelated_pc: true,
                pc: 0x1000_hdl_u64,
            },
        ),
    );
    sim.write(
        fetch_decode_interface.decoded_insns[1].data,
        HdlSome(
            #[hdl]
            FetchedDecodedMOp {
                mop: MOp.AluBranch(MOp.AluBranch.Logical(
                    #[hdl]
                    LogicalMOp {
                        alu_common: #[hdl]
                        AluCommonMOp {
                            common: CommonMOp::new(
                                0_hdl_u0,
                                #[hdl]
                                MOpDestReg {
                                    normal_regs: [
                                        #[hdl]
                                        MOpRegNum { value: 2u8 },
                                        MOpRegNum::const_zero(),
                                    ],
                                    flag_regs: [HdlNone(), HdlSome(())],
                                },
                                [3u8, 4u8],
                                SInt[COMMON_MOP_2_IMM_WIDTH].zero(),
                            ),
                            output_integer_mode: OutputIntegerMode.Full64(),
                        },
                        lut: 0b0110_hdl_u4,
                    },
                )),
                is_unrelated_pc: false,
                pc: 0x1004_hdl_u64,
            },
        ),
    );
    for cycle in 0..20 {
        sim.advance_time(SimDuration::from_nanos(500));
        sim.write_clock(sim.io().cd.clk, true);
        sim.advance_time(SimDuration::from_nanos(500));
        sim.write_clock(sim.io().cd.clk, false);
        sim.write_reset(sim.io().cd.rst, false);
    }
    // FIXME: vcd is just whatever reg_alloc does now, which isn't known to be correct
    let vcd = String::from_utf8(writer.take()).unwrap();
    println!("####### VCD:\n{vcd}\n#######");
    if vcd != include_str!("expected/reg_alloc.vcd") {
        panic!();
    }
    #[rustfmt::skip] // work around https://github.com/rust-lang/rustfmt/issues/6161
    assert_export_firrtl! {
        m =>
        options: ExportOptions {
            simplify_enums: None,
            ..ExportOptions::default()
        },
        "/test/reg_alloc.fir": r"FIRRTL version 3.2.0
circuit reg_alloc:
    type Ty0 = {clk: Clock, rst: Reset}
    type Ty1 = {value: UInt<8>}
    type Ty2 = {}
    type Ty3 = {|HdlNone, HdlSome: Ty2|}
    type Ty4 = {normal_regs: Ty1[2], flag_regs: Ty3[2]}
    type Ty5 = {prefix_pad: UInt<0>, dest: Ty4, src: UInt<8>[3], imm_low: UInt<25>, imm_sign: SInt<1>, _phantom: Ty2}
    type Ty6 = {|Full64, DupLow32, ZeroExt32, SignExt32, ZeroExt16, SignExt16, ZeroExt8, SignExt8|}
    type Ty7 = {common: Ty5, output_integer_mode: Ty6}
    type Ty8 = {alu_common: Ty7, invert_src0: UInt<1>, invert_carry_in: UInt<1>, invert_carry_out: UInt<1>, add_pc: UInt<1>}
    type Ty9 = {alu_common: Ty7, lut: UInt<4>}
    type Ty10 = {|AddSub: Ty8, AddSubI: Ty8, Logical: Ty9|}
    type Ty11 = {prefix_pad: UInt<1>, dest: Ty4, src: UInt<8>[3], imm_low: UInt<25>, imm_sign: SInt<1>, _phantom: Ty2}
    type Ty12 = {common: Ty11}
    type Ty13 = {|ReadL2Reg: Ty12, WriteL2Reg: Ty12|}
    type Ty14 = {|Load: Ty11, Store: Ty11|}
    type Ty15 = {|AluBranch: Ty10, L2RegisterFile: Ty13, LoadStore: Ty14|}
    type Ty16 = {mop: Ty15, is_unrelated_pc: UInt<1>, pc: UInt<64>}
    type Ty17 = {|HdlNone, HdlSome: Ty16|}
    type Ty18 = {data: Ty17, flip ready: UInt<1>}
    type Ty19 = {|Trap: Ty2, ICacheFlush|}
    type Ty20 = {|HdlNone, HdlSome: Ty19|}
    type Ty21 = {data: Ty20, flip ready: UInt<1>}
    type Ty22 = {decoded_insns: Ty18[2], flip fetch_decode_special_op: Ty21}
    type Ty23 = {adj_value: UInt<2>}
    type Ty24 = {value: UInt<4>}
    type Ty25 = {unit_num: Ty23, unit_out_reg: Ty24}
    type Ty26 = {addr: UInt<8>, en: UInt<1>, clk: Clock, flip data: Ty25}
    type Ty27 = {addr: UInt<8>, en: UInt<1>, clk: Clock, flip data: UInt<2>}
    type Ty28 = {addr: UInt<8>, en: UInt<1>, clk: Clock, flip data: UInt<4>}
    type Ty29 = {addr: UInt<1>, en: UInt<1>, clk: Clock, flip data: Ty25}
    type Ty30 = {addr: UInt<1>, en: UInt<1>, clk: Clock, flip data: UInt<2>}
    type Ty31 = {addr: UInt<1>, en: UInt<1>, clk: Clock, flip data: UInt<4>}
    type Ty32 = {|HdlNone, HdlSome: UInt<2>|}
    type Ty33 = {prefix_pad: UInt<0>, dest: Ty24, src: UInt<6>[3], imm_low: UInt<25>, imm_sign: SInt<1>, _phantom: Ty2}
    type Ty34 = {common: Ty33, output_integer_mode: Ty6}
    type Ty35 = {alu_common: Ty34, invert_src0: UInt<1>, invert_carry_in: UInt<1>, invert_carry_out: UInt<1>, add_pc: UInt<1>}
    type Ty36 = {alu_common: Ty34, lut: UInt<4>}
    type Ty37 = {|AddSub: Ty35, AddSubI: Ty35, Logical: Ty36|}
    type Ty38 = {prefix_pad: UInt<1>, dest: Ty24, src: UInt<6>[3], imm_low: UInt<25>, imm_sign: SInt<1>, _phantom: Ty2}
    type Ty39 = {common: Ty38}
    type Ty40 = {|ReadL2Reg: Ty39, WriteL2Reg: Ty39|}
    type Ty41 = {|Load: Ty38, Store: Ty38|}
    type Ty42 = {|AluBranch: Ty37, L2RegisterFile: Ty40, LoadStore: Ty41|}
    type Ty43 = {|HdlNone, HdlSome: Ty42|}
    type Ty44 = {|HdlNone, HdlSome: Ty25|}
    type Ty45 = {addr: Ty1, flip data: Ty44}
    type Ty46 = {|AluBranch, L2RegisterFile, LoadStore|}
    type Ty47 = {unit_num: UInt<2>, unit_out_reg: UInt<4>}
    type Ty48 = {imm_low: UInt<25>, reversed_src: UInt<8>[0], imm_sign: SInt<1>}
    type Ty49 = {imm_low: UInt<25>, reversed_src: UInt<0>, imm_sign: UInt<1>}
    type Ty50 = {imm_low: UInt<25>, reversed_src: UInt<8>[1], imm_sign: SInt<1>}
    type Ty51 = {imm_low: UInt<25>, reversed_src: UInt<8>, imm_sign: UInt<1>}
    type Ty52 = {|HdlNone, HdlSome: Ty37|}
    type Ty53 = {data: Ty52, flip ready: UInt<1>}
    type Ty54 = {flip cd: Ty0, flip `input`: Ty53}
    type Ty55 = {|HdlNone, HdlSome: UInt<4>|}
    type Ty56 = {data: Ty55, flip ready: UInt<1>}
    type Ty57 = {flip cd: Ty0, flip free_in: Ty56[1], alloc_out: Ty56[1]}
    module reg_alloc: @[reg_alloc.rs 43:1]
        input cd: Ty0 @[reg_alloc.rs 47:29]
        input fetch_decode_interface: Ty22 @[reg_alloc.rs 50:11]
        mem rename_table_normal_mem_unit_num_adj_value: @[reg_alloc.rs 60:39]
            data-type => UInt<2>
            depth => 253
            read-latency => 0
            write-latency => 1
            read-under-write => old
            reader => r0
            reader => r1
            reader => r2
            reader => r3
            reader => r4
            reader => r5
        mem rename_table_normal_mem_unit_out_reg_value: @[reg_alloc.rs 60:39]
            data-type => UInt<4>
            depth => 253
            read-latency => 0
            write-latency => 1
            read-under-write => old
            reader => r0
            reader => r1
            reader => r2
            reader => r3
            reader => r4
            reader => r5
        mem rename_table_special_mem_unit_num_adj_value: @[reg_alloc.rs 65:40]
            data-type => UInt<2>
            depth => 2
            read-latency => 0
            write-latency => 1
            read-under-write => old
            reader => r0
            reader => r1
            reader => r2
            reader => r3
            reader => r4
            reader => r5
        mem rename_table_special_mem_unit_out_reg_value: @[reg_alloc.rs 65:40]
            data-type => UInt<4>
            depth => 2
            read-latency => 0
            write-latency => 1
            read-under-write => old
            reader => r0
            reader => r1
            reader => r2
            reader => r3
            reader => r4
            reader => r5
        wire rename_table_normal_mem_r0: Ty26 @[reg_alloc.rs 102:37]
        wire rename_table_normal_mem_r1: Ty26 @[reg_alloc.rs 102:37]
        wire rename_table_normal_mem_r2: Ty26 @[reg_alloc.rs 102:37]
        wire rename_table_normal_mem_r3: Ty26 @[reg_alloc.rs 102:37]
        wire rename_table_normal_mem_r4: Ty26 @[reg_alloc.rs 102:37]
        wire rename_table_normal_mem_r5: Ty26 @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_r0.data.unit_num.adj_value, rename_table_normal_mem_unit_num_adj_value.r0.data @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_r1.data.unit_num.adj_value, rename_table_normal_mem_unit_num_adj_value.r1.data @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_r2.data.unit_num.adj_value, rename_table_normal_mem_unit_num_adj_value.r2.data @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_r3.data.unit_num.adj_value, rename_table_normal_mem_unit_num_adj_value.r3.data @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_r4.data.unit_num.adj_value, rename_table_normal_mem_unit_num_adj_value.r4.data @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_r5.data.unit_num.adj_value, rename_table_normal_mem_unit_num_adj_value.r5.data @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r0.addr, rename_table_normal_mem_r0.addr @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r0.clk, rename_table_normal_mem_r0.clk @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r0.en, rename_table_normal_mem_r0.en @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r1.addr, rename_table_normal_mem_r1.addr @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r1.clk, rename_table_normal_mem_r1.clk @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r1.en, rename_table_normal_mem_r1.en @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r2.addr, rename_table_normal_mem_r2.addr @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r2.clk, rename_table_normal_mem_r2.clk @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r2.en, rename_table_normal_mem_r2.en @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r3.addr, rename_table_normal_mem_r3.addr @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r3.clk, rename_table_normal_mem_r3.clk @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r3.en, rename_table_normal_mem_r3.en @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r4.addr, rename_table_normal_mem_r4.addr @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r4.clk, rename_table_normal_mem_r4.clk @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r4.en, rename_table_normal_mem_r4.en @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r5.addr, rename_table_normal_mem_r5.addr @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r5.clk, rename_table_normal_mem_r5.clk @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_num_adj_value.r5.en, rename_table_normal_mem_r5.en @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_r0.data.unit_out_reg.value, rename_table_normal_mem_unit_out_reg_value.r0.data @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_r1.data.unit_out_reg.value, rename_table_normal_mem_unit_out_reg_value.r1.data @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_r2.data.unit_out_reg.value, rename_table_normal_mem_unit_out_reg_value.r2.data @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_r3.data.unit_out_reg.value, rename_table_normal_mem_unit_out_reg_value.r3.data @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_r4.data.unit_out_reg.value, rename_table_normal_mem_unit_out_reg_value.r4.data @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_r5.data.unit_out_reg.value, rename_table_normal_mem_unit_out_reg_value.r5.data @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r0.addr, rename_table_normal_mem_r0.addr @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r0.clk, rename_table_normal_mem_r0.clk @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r0.en, rename_table_normal_mem_r0.en @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r1.addr, rename_table_normal_mem_r1.addr @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r1.clk, rename_table_normal_mem_r1.clk @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r1.en, rename_table_normal_mem_r1.en @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r2.addr, rename_table_normal_mem_r2.addr @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r2.clk, rename_table_normal_mem_r2.clk @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r2.en, rename_table_normal_mem_r2.en @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r3.addr, rename_table_normal_mem_r3.addr @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r3.clk, rename_table_normal_mem_r3.clk @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r3.en, rename_table_normal_mem_r3.en @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r4.addr, rename_table_normal_mem_r4.addr @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r4.clk, rename_table_normal_mem_r4.clk @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r4.en, rename_table_normal_mem_r4.en @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r5.addr, rename_table_normal_mem_r5.addr @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r5.clk, rename_table_normal_mem_r5.clk @[reg_alloc.rs 102:37]
        connect rename_table_normal_mem_unit_out_reg_value.r5.en, rename_table_normal_mem_r5.en @[reg_alloc.rs 102:37]
        wire rename_table_special_mem_r0: Ty29 @[reg_alloc.rs 102:37]
        wire rename_table_special_mem_r1: Ty29 @[reg_alloc.rs 102:37]
        wire rename_table_special_mem_r2: Ty29 @[reg_alloc.rs 102:37]
        wire rename_table_special_mem_r3: Ty29 @[reg_alloc.rs 102:37]
        wire rename_table_special_mem_r4: Ty29 @[reg_alloc.rs 102:37]
        wire rename_table_special_mem_r5: Ty29 @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_r0.data.unit_num.adj_value, rename_table_special_mem_unit_num_adj_value.r0.data @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_r1.data.unit_num.adj_value, rename_table_special_mem_unit_num_adj_value.r1.data @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_r2.data.unit_num.adj_value, rename_table_special_mem_unit_num_adj_value.r2.data @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_r3.data.unit_num.adj_value, rename_table_special_mem_unit_num_adj_value.r3.data @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_r4.data.unit_num.adj_value, rename_table_special_mem_unit_num_adj_value.r4.data @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_r5.data.unit_num.adj_value, rename_table_special_mem_unit_num_adj_value.r5.data @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r0.addr, rename_table_special_mem_r0.addr @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r0.clk, rename_table_special_mem_r0.clk @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r0.en, rename_table_special_mem_r0.en @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r1.addr, rename_table_special_mem_r1.addr @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r1.clk, rename_table_special_mem_r1.clk @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r1.en, rename_table_special_mem_r1.en @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r2.addr, rename_table_special_mem_r2.addr @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r2.clk, rename_table_special_mem_r2.clk @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r2.en, rename_table_special_mem_r2.en @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r3.addr, rename_table_special_mem_r3.addr @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r3.clk, rename_table_special_mem_r3.clk @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r3.en, rename_table_special_mem_r3.en @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r4.addr, rename_table_special_mem_r4.addr @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r4.clk, rename_table_special_mem_r4.clk @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r4.en, rename_table_special_mem_r4.en @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r5.addr, rename_table_special_mem_r5.addr @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r5.clk, rename_table_special_mem_r5.clk @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_num_adj_value.r5.en, rename_table_special_mem_r5.en @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_r0.data.unit_out_reg.value, rename_table_special_mem_unit_out_reg_value.r0.data @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_r1.data.unit_out_reg.value, rename_table_special_mem_unit_out_reg_value.r1.data @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_r2.data.unit_out_reg.value, rename_table_special_mem_unit_out_reg_value.r2.data @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_r3.data.unit_out_reg.value, rename_table_special_mem_unit_out_reg_value.r3.data @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_r4.data.unit_out_reg.value, rename_table_special_mem_unit_out_reg_value.r4.data @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_r5.data.unit_out_reg.value, rename_table_special_mem_unit_out_reg_value.r5.data @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r0.addr, rename_table_special_mem_r0.addr @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r0.clk, rename_table_special_mem_r0.clk @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r0.en, rename_table_special_mem_r0.en @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r1.addr, rename_table_special_mem_r1.addr @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r1.clk, rename_table_special_mem_r1.clk @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r1.en, rename_table_special_mem_r1.en @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r2.addr, rename_table_special_mem_r2.addr @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r2.clk, rename_table_special_mem_r2.clk @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r2.en, rename_table_special_mem_r2.en @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r3.addr, rename_table_special_mem_r3.addr @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r3.clk, rename_table_special_mem_r3.clk @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r3.en, rename_table_special_mem_r3.en @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r4.addr, rename_table_special_mem_r4.addr @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r4.clk, rename_table_special_mem_r4.clk @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r4.en, rename_table_special_mem_r4.en @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r5.addr, rename_table_special_mem_r5.addr @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r5.clk, rename_table_special_mem_r5.clk @[reg_alloc.rs 102:37]
        connect rename_table_special_mem_unit_out_reg_value.r5.en, rename_table_special_mem_r5.en @[reg_alloc.rs 102:37]
        connect fetch_decode_interface.fetch_decode_special_op.data, {|HdlNone, HdlSome: Ty19|}(HdlNone) @[reg_alloc.rs 52:5]
        wire available_units: UInt<1>[2][2] @[reg_alloc.rs 70:9]
        wire selected_unit_indexes: Ty32[2] @[reg_alloc.rs 73:9]
        wire renamed_mops: Ty43[2] @[reg_alloc.rs 75:24]
        wire renamed_mops_out_reg: Ty44[2] @[reg_alloc.rs 77:32]
        connect fetch_decode_interface.decoded_insns[0].ready, UInt<1>(0h1) @[reg_alloc.rs 79:9]
        wire _array_literal_expr: UInt<1>[2]
        connect _array_literal_expr[0], UInt<1>(0h0)
        connect _array_literal_expr[1], UInt<1>(0h0)
        connect available_units[0], _array_literal_expr @[reg_alloc.rs 83:9]
        connect renamed_mops[0], {|HdlNone, HdlSome: Ty42|}(HdlNone) @[reg_alloc.rs 87:9]
        connect rename_table_normal_mem_r0.clk, cd.clk @[reg_alloc.rs 103:17]
        connect rename_table_normal_mem_r0.addr, UInt<8>(0h0) @[reg_alloc.rs 104:17]
        connect rename_table_normal_mem_r0.en, UInt<1>(0h0) @[reg_alloc.rs 105:17]
        wire rename_table_normal_0_src_0: Ty45 @[reg_alloc.rs 108:21]
        wire _bundle_literal_expr: Ty1
        connect _bundle_literal_expr.value, tail(UInt<32>(0h0), 24)
        connect rename_table_normal_0_src_0.addr, _bundle_literal_expr @[reg_alloc.rs 111:17]
        connect rename_table_normal_0_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlNone) @[reg_alloc.rs 112:17]
        when and(geq(rename_table_normal_0_src_0.addr.value, UInt<32>(0h1)), lt(rename_table_normal_0_src_0.addr.value, UInt<32>(0hFE))): @[reg_alloc.rs 114:17]
            ; connect different types:
            ; lhs: UInt<8>
            ; rhs: UInt<33>
            connect rename_table_normal_mem_r0.addr, sub(rename_table_normal_0_src_0.addr.value, UInt<32>(0h1)) @[reg_alloc.rs 115:21]
            connect rename_table_normal_mem_r0.en, UInt<1>(0h1) @[reg_alloc.rs 116:21]
            connect rename_table_normal_0_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, rename_table_normal_mem_r0.data) @[reg_alloc.rs 117:21]
        connect rename_table_normal_mem_r1.clk, cd.clk @[reg_alloc.rs 103:17]
        connect rename_table_normal_mem_r1.addr, UInt<8>(0h0) @[reg_alloc.rs 104:17]
        connect rename_table_normal_mem_r1.en, UInt<1>(0h0) @[reg_alloc.rs 105:17]
        wire rename_table_normal_0_src_1: Ty45 @[reg_alloc.rs 108:21]
        wire _bundle_literal_expr_1: Ty1
        connect _bundle_literal_expr_1.value, tail(UInt<32>(0h0), 24)
        connect rename_table_normal_0_src_1.addr, _bundle_literal_expr_1 @[reg_alloc.rs 111:17]
        connect rename_table_normal_0_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlNone) @[reg_alloc.rs 112:17]
        when and(geq(rename_table_normal_0_src_1.addr.value, UInt<32>(0h1)), lt(rename_table_normal_0_src_1.addr.value, UInt<32>(0hFE))): @[reg_alloc.rs 114:17]
            ; connect different types:
            ; lhs: UInt<8>
            ; rhs: UInt<33>
            connect rename_table_normal_mem_r1.addr, sub(rename_table_normal_0_src_1.addr.value, UInt<32>(0h1)) @[reg_alloc.rs 115:21]
            connect rename_table_normal_mem_r1.en, UInt<1>(0h1) @[reg_alloc.rs 116:21]
            connect rename_table_normal_0_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, rename_table_normal_mem_r1.data) @[reg_alloc.rs 117:21]
        connect rename_table_normal_mem_r2.clk, cd.clk @[reg_alloc.rs 103:17]
        connect rename_table_normal_mem_r2.addr, UInt<8>(0h0) @[reg_alloc.rs 104:17]
        connect rename_table_normal_mem_r2.en, UInt<1>(0h0) @[reg_alloc.rs 105:17]
        wire rename_table_normal_0_src_2: Ty45 @[reg_alloc.rs 108:21]
        wire _bundle_literal_expr_2: Ty1
        connect _bundle_literal_expr_2.value, tail(UInt<32>(0h0), 24)
        connect rename_table_normal_0_src_2.addr, _bundle_literal_expr_2 @[reg_alloc.rs 111:17]
        connect rename_table_normal_0_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlNone) @[reg_alloc.rs 112:17]
        when and(geq(rename_table_normal_0_src_2.addr.value, UInt<32>(0h1)), lt(rename_table_normal_0_src_2.addr.value, UInt<32>(0hFE))): @[reg_alloc.rs 114:17]
            ; connect different types:
            ; lhs: UInt<8>
            ; rhs: UInt<33>
            connect rename_table_normal_mem_r2.addr, sub(rename_table_normal_0_src_2.addr.value, UInt<32>(0h1)) @[reg_alloc.rs 115:21]
            connect rename_table_normal_mem_r2.en, UInt<1>(0h1) @[reg_alloc.rs 116:21]
            connect rename_table_normal_0_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, rename_table_normal_mem_r2.data) @[reg_alloc.rs 117:21]
        connect rename_table_special_mem_r0.clk, cd.clk @[reg_alloc.rs 103:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<8>
        connect rename_table_special_mem_r0.addr, UInt<8>(0h0) @[reg_alloc.rs 104:17]
        connect rename_table_special_mem_r0.en, UInt<1>(0h0) @[reg_alloc.rs 105:17]
        wire rename_table_special_0_src_0: Ty45 @[reg_alloc.rs 108:21]
        wire _bundle_literal_expr_3: Ty1
        connect _bundle_literal_expr_3.value, tail(UInt<32>(0h0), 24)
        connect rename_table_special_0_src_0.addr, _bundle_literal_expr_3 @[reg_alloc.rs 111:17]
        connect rename_table_special_0_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlNone) @[reg_alloc.rs 112:17]
        when and(geq(rename_table_special_0_src_0.addr.value, UInt<32>(0hFE)), lt(rename_table_special_0_src_0.addr.value, UInt<32>(0h100))): @[reg_alloc.rs 114:17]
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<33>
            connect rename_table_special_mem_r0.addr, sub(rename_table_special_0_src_0.addr.value, UInt<32>(0hFE)) @[reg_alloc.rs 115:21]
            connect rename_table_special_mem_r0.en, UInt<1>(0h1) @[reg_alloc.rs 116:21]
            connect rename_table_special_0_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, rename_table_special_mem_r0.data) @[reg_alloc.rs 117:21]
        connect rename_table_special_mem_r1.clk, cd.clk @[reg_alloc.rs 103:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<8>
        connect rename_table_special_mem_r1.addr, UInt<8>(0h0) @[reg_alloc.rs 104:17]
        connect rename_table_special_mem_r1.en, UInt<1>(0h0) @[reg_alloc.rs 105:17]
        wire rename_table_special_0_src_1: Ty45 @[reg_alloc.rs 108:21]
        wire _bundle_literal_expr_4: Ty1
        connect _bundle_literal_expr_4.value, tail(UInt<32>(0h0), 24)
        connect rename_table_special_0_src_1.addr, _bundle_literal_expr_4 @[reg_alloc.rs 111:17]
        connect rename_table_special_0_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlNone) @[reg_alloc.rs 112:17]
        when and(geq(rename_table_special_0_src_1.addr.value, UInt<32>(0hFE)), lt(rename_table_special_0_src_1.addr.value, UInt<32>(0h100))): @[reg_alloc.rs 114:17]
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<33>
            connect rename_table_special_mem_r1.addr, sub(rename_table_special_0_src_1.addr.value, UInt<32>(0hFE)) @[reg_alloc.rs 115:21]
            connect rename_table_special_mem_r1.en, UInt<1>(0h1) @[reg_alloc.rs 116:21]
            connect rename_table_special_0_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, rename_table_special_mem_r1.data) @[reg_alloc.rs 117:21]
        connect rename_table_special_mem_r2.clk, cd.clk @[reg_alloc.rs 103:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<8>
        connect rename_table_special_mem_r2.addr, UInt<8>(0h0) @[reg_alloc.rs 104:17]
        connect rename_table_special_mem_r2.en, UInt<1>(0h0) @[reg_alloc.rs 105:17]
        wire rename_table_special_0_src_2: Ty45 @[reg_alloc.rs 108:21]
        wire _bundle_literal_expr_5: Ty1
        connect _bundle_literal_expr_5.value, tail(UInt<32>(0h0), 24)
        connect rename_table_special_0_src_2.addr, _bundle_literal_expr_5 @[reg_alloc.rs 111:17]
        connect rename_table_special_0_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlNone) @[reg_alloc.rs 112:17]
        when and(geq(rename_table_special_0_src_2.addr.value, UInt<32>(0hFE)), lt(rename_table_special_0_src_2.addr.value, UInt<32>(0h100))): @[reg_alloc.rs 114:17]
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<33>
            connect rename_table_special_mem_r2.addr, sub(rename_table_special_0_src_2.addr.value, UInt<32>(0hFE)) @[reg_alloc.rs 115:21]
            connect rename_table_special_mem_r2.en, UInt<1>(0h1) @[reg_alloc.rs 116:21]
            connect rename_table_special_0_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, rename_table_special_mem_r2.data) @[reg_alloc.rs 117:21]
        match fetch_decode_interface.decoded_insns[0].data: @[reg_alloc.rs 159:9]
            HdlNone:
                skip
            HdlSome(_match_arm_value):
                wire unit_kind: Ty46 @[unit.rs 127:1]
                match _match_arm_value.mop: @[unit.rs 127:1]
                    AluBranch(_match_arm_value_1):
                        connect unit_kind, {|AluBranch, L2RegisterFile, LoadStore|}(AluBranch) @[unit.rs 127:1]
                    L2RegisterFile(_match_arm_value_2):
                        connect unit_kind, {|AluBranch, L2RegisterFile, LoadStore|}(L2RegisterFile) @[unit.rs 127:1]
                    LoadStore(_match_arm_value_3):
                        connect unit_kind, {|AluBranch, L2RegisterFile, LoadStore|}(LoadStore) @[unit.rs 127:1]
                wire available_units_for_kind: UInt<1>[2] @[unit.rs 127:1]
                match unit_kind: @[unit.rs 127:1]
                    AluBranch:
                        connect available_units_for_kind[0], UInt<1>(0h1) @[unit.rs 127:1]
                        connect available_units_for_kind[1], UInt<1>(0h1) @[unit.rs 127:1]
                    L2RegisterFile:
                        connect available_units_for_kind[0], UInt<1>(0h0) @[unit.rs 127:1]
                        connect available_units_for_kind[1], UInt<1>(0h0) @[unit.rs 127:1]
                    LoadStore:
                        connect available_units_for_kind[0], UInt<1>(0h0) @[unit.rs 127:1]
                        connect available_units_for_kind[1], UInt<1>(0h0) @[unit.rs 127:1]
                connect available_units[0], available_units_for_kind @[reg_alloc.rs 160:13]
                match renamed_mops_out_reg[0]: @[reg_alloc.rs 165:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_4):
                        wire dest_reg: Ty4 @[unit.rs 127:1]
                        match _match_arm_value.mop: @[unit.rs 127:1]
                            AluBranch(_match_arm_value_5):
                                wire dest_reg_1: Ty4 @[instruction.rs 477:1]
                                match _match_arm_value_5: @[instruction.rs 477:1]
                                    AddSub(_match_arm_value_6):
                                        connect dest_reg_1, _match_arm_value_6.alu_common.common.dest @[instruction.rs 477:1]
                                    AddSubI(_match_arm_value_7):
                                        connect dest_reg_1, _match_arm_value_7.alu_common.common.dest @[instruction.rs 477:1]
                                    Logical(_match_arm_value_8):
                                        connect dest_reg_1, _match_arm_value_8.alu_common.common.dest @[instruction.rs 477:1]
                                connect dest_reg, dest_reg_1 @[unit.rs 127:1]
                            L2RegisterFile(_match_arm_value_9):
                                wire dest_reg_2: Ty4 @[instruction.rs 504:1]
                                match _match_arm_value_9: @[instruction.rs 504:1]
                                    ReadL2Reg(_match_arm_value_10):
                                        connect dest_reg_2, _match_arm_value_10.common.dest @[instruction.rs 504:1]
                                    WriteL2Reg(_match_arm_value_11):
                                        connect dest_reg_2, _match_arm_value_11.common.dest @[instruction.rs 504:1]
                                connect dest_reg, dest_reg_2 @[unit.rs 127:1]
                            LoadStore(_match_arm_value_12):
                                wire dest_reg_3: Ty4 @[instruction.rs 539:1]
                                match _match_arm_value_12: @[instruction.rs 539:1]
                                    Load(_match_arm_value_13):
                                        connect dest_reg_3, _match_arm_value_13.dest @[instruction.rs 539:1]
                                    Store(_match_arm_value_14):
                                        connect dest_reg_3, _match_arm_value_14.dest @[instruction.rs 539:1]
                                connect dest_reg, dest_reg_3 @[unit.rs 127:1]
                        wire mapped_regs: Ty42 @[unit.rs 127:1]
                        match _match_arm_value.mop: @[unit.rs 127:1]
                            AluBranch(_match_arm_value_15):
                                wire mapped_regs_1: Ty37 @[instruction.rs 477:1]
                                match _match_arm_value_15: @[instruction.rs 477:1]
                                    AddSub(_match_arm_value_16):
                                        wire renamed_src_reg_0_0: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_6: Ty1
                                        connect _bundle_literal_expr_6.value, _match_arm_value_16.alu_common.common.src[0]
                                        connect rename_table_normal_0_src_0.addr, _bundle_literal_expr_6 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_7: Ty1
                                        connect _bundle_literal_expr_7.value, _match_arm_value_16.alu_common.common.src[0]
                                        connect rename_table_special_0_src_0.addr, _bundle_literal_expr_7 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_0_src_0.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_0_src_0.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_8: Ty25
                                                        wire _bundle_literal_expr_9: Ty23
                                                        connect _bundle_literal_expr_9.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_8.unit_num, _bundle_literal_expr_9
                                                        wire _bundle_literal_expr_10: Ty24
                                                        connect _bundle_literal_expr_10.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_8.unit_out_reg, _bundle_literal_expr_10
                                                        connect renamed_src_reg_0_0, _bundle_literal_expr_8 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_17):
                                                        connect renamed_src_reg_0_0, _match_arm_value_17 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_18):
                                                connect renamed_src_reg_0_0, _match_arm_value_18 @[reg_alloc.rs 185:33]
                                        wire renamed_src_reg_0_1: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_11: Ty1
                                        connect _bundle_literal_expr_11.value, _match_arm_value_16.alu_common.common.src[1]
                                        connect rename_table_normal_0_src_1.addr, _bundle_literal_expr_11 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_12: Ty1
                                        connect _bundle_literal_expr_12.value, _match_arm_value_16.alu_common.common.src[1]
                                        connect rename_table_special_0_src_1.addr, _bundle_literal_expr_12 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_0_src_1.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_0_src_1.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_13: Ty25
                                                        wire _bundle_literal_expr_14: Ty23
                                                        connect _bundle_literal_expr_14.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_13.unit_num, _bundle_literal_expr_14
                                                        wire _bundle_literal_expr_15: Ty24
                                                        connect _bundle_literal_expr_15.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_13.unit_out_reg, _bundle_literal_expr_15
                                                        connect renamed_src_reg_0_1, _bundle_literal_expr_13 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_19):
                                                        connect renamed_src_reg_0_1, _match_arm_value_19 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_20):
                                                connect renamed_src_reg_0_1, _match_arm_value_20 @[reg_alloc.rs 185:33]
                                        wire renamed_src_reg_0_2: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_16: Ty1
                                        connect _bundle_literal_expr_16.value, _match_arm_value_16.alu_common.common.src[2]
                                        connect rename_table_normal_0_src_2.addr, _bundle_literal_expr_16 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_17: Ty1
                                        connect _bundle_literal_expr_17.value, _match_arm_value_16.alu_common.common.src[2]
                                        connect rename_table_special_0_src_2.addr, _bundle_literal_expr_17 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_0_src_2.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_0_src_2.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_18: Ty25
                                                        wire _bundle_literal_expr_19: Ty23
                                                        connect _bundle_literal_expr_19.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_18.unit_num, _bundle_literal_expr_19
                                                        wire _bundle_literal_expr_20: Ty24
                                                        connect _bundle_literal_expr_20.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_18.unit_out_reg, _bundle_literal_expr_20
                                                        connect renamed_src_reg_0_2, _bundle_literal_expr_18 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_21):
                                                        connect renamed_src_reg_0_2, _match_arm_value_21 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_22):
                                                connect renamed_src_reg_0_2, _match_arm_value_22 @[reg_alloc.rs 185:33]
                                        wire _bundle_literal_expr_21: Ty35
                                        wire _bundle_literal_expr_22: Ty34
                                        wire _bundle_literal_expr_23: Ty33
                                        connect _bundle_literal_expr_23.prefix_pad, _match_arm_value_16.alu_common.common.prefix_pad
                                        connect _bundle_literal_expr_23.dest, _match_arm_value_4.unit_out_reg
                                        wire _array_literal_expr_1: UInt<6>[3]
                                        wire _array_literal_expr_2: UInt<6>[3]
                                        wire _cast_bundle_to_bits_expr: Ty47
                                        connect _cast_bundle_to_bits_expr.unit_num, renamed_src_reg_0_0.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr.unit_out_reg, renamed_src_reg_0_0.unit_out_reg.value
                                        wire _cast_to_bits_expr: UInt<6>
                                        connect _cast_to_bits_expr, cat(_cast_bundle_to_bits_expr.unit_out_reg, _cast_bundle_to_bits_expr.unit_num)
                                        connect _array_literal_expr_2[0], _cast_to_bits_expr
                                        wire _cast_bundle_to_bits_expr_1: Ty47
                                        connect _cast_bundle_to_bits_expr_1.unit_num, renamed_src_reg_0_1.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_1.unit_out_reg, renamed_src_reg_0_1.unit_out_reg.value
                                        wire _cast_to_bits_expr_1: UInt<6>
                                        connect _cast_to_bits_expr_1, cat(_cast_bundle_to_bits_expr_1.unit_out_reg, _cast_bundle_to_bits_expr_1.unit_num)
                                        connect _array_literal_expr_2[1], _cast_to_bits_expr_1
                                        wire _cast_bundle_to_bits_expr_2: Ty47
                                        connect _cast_bundle_to_bits_expr_2.unit_num, renamed_src_reg_0_2.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_2.unit_out_reg, renamed_src_reg_0_2.unit_out_reg.value
                                        wire _cast_to_bits_expr_2: UInt<6>
                                        connect _cast_to_bits_expr_2, cat(_cast_bundle_to_bits_expr_2.unit_out_reg, _cast_bundle_to_bits_expr_2.unit_num)
                                        connect _array_literal_expr_2[2], _cast_to_bits_expr_2
                                        connect _array_literal_expr_1[0], _array_literal_expr_2[0]
                                        wire _array_literal_expr_3: UInt<6>[3]
                                        wire _cast_bundle_to_bits_expr_3: Ty47
                                        connect _cast_bundle_to_bits_expr_3.unit_num, renamed_src_reg_0_0.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_3.unit_out_reg, renamed_src_reg_0_0.unit_out_reg.value
                                        wire _cast_to_bits_expr_3: UInt<6>
                                        connect _cast_to_bits_expr_3, cat(_cast_bundle_to_bits_expr_3.unit_out_reg, _cast_bundle_to_bits_expr_3.unit_num)
                                        connect _array_literal_expr_3[0], _cast_to_bits_expr_3
                                        wire _cast_bundle_to_bits_expr_4: Ty47
                                        connect _cast_bundle_to_bits_expr_4.unit_num, renamed_src_reg_0_1.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_4.unit_out_reg, renamed_src_reg_0_1.unit_out_reg.value
                                        wire _cast_to_bits_expr_4: UInt<6>
                                        connect _cast_to_bits_expr_4, cat(_cast_bundle_to_bits_expr_4.unit_out_reg, _cast_bundle_to_bits_expr_4.unit_num)
                                        connect _array_literal_expr_3[1], _cast_to_bits_expr_4
                                        wire _cast_bundle_to_bits_expr_5: Ty47
                                        connect _cast_bundle_to_bits_expr_5.unit_num, renamed_src_reg_0_2.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_5.unit_out_reg, renamed_src_reg_0_2.unit_out_reg.value
                                        wire _cast_to_bits_expr_5: UInt<6>
                                        connect _cast_to_bits_expr_5, cat(_cast_bundle_to_bits_expr_5.unit_out_reg, _cast_bundle_to_bits_expr_5.unit_num)
                                        connect _array_literal_expr_3[2], _cast_to_bits_expr_5
                                        connect _array_literal_expr_1[1], _array_literal_expr_3[1]
                                        wire _array_literal_expr_4: UInt<6>[3]
                                        wire _cast_bundle_to_bits_expr_6: Ty47
                                        connect _cast_bundle_to_bits_expr_6.unit_num, renamed_src_reg_0_0.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_6.unit_out_reg, renamed_src_reg_0_0.unit_out_reg.value
                                        wire _cast_to_bits_expr_6: UInt<6>
                                        connect _cast_to_bits_expr_6, cat(_cast_bundle_to_bits_expr_6.unit_out_reg, _cast_bundle_to_bits_expr_6.unit_num)
                                        connect _array_literal_expr_4[0], _cast_to_bits_expr_6
                                        wire _cast_bundle_to_bits_expr_7: Ty47
                                        connect _cast_bundle_to_bits_expr_7.unit_num, renamed_src_reg_0_1.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_7.unit_out_reg, renamed_src_reg_0_1.unit_out_reg.value
                                        wire _cast_to_bits_expr_7: UInt<6>
                                        connect _cast_to_bits_expr_7, cat(_cast_bundle_to_bits_expr_7.unit_out_reg, _cast_bundle_to_bits_expr_7.unit_num)
                                        connect _array_literal_expr_4[1], _cast_to_bits_expr_7
                                        wire _cast_bundle_to_bits_expr_8: Ty47
                                        connect _cast_bundle_to_bits_expr_8.unit_num, renamed_src_reg_0_2.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_8.unit_out_reg, renamed_src_reg_0_2.unit_out_reg.value
                                        wire _cast_to_bits_expr_8: UInt<6>
                                        connect _cast_to_bits_expr_8, cat(_cast_bundle_to_bits_expr_8.unit_out_reg, _cast_bundle_to_bits_expr_8.unit_num)
                                        connect _array_literal_expr_4[2], _cast_to_bits_expr_8
                                        connect _array_literal_expr_1[2], _array_literal_expr_4[2]
                                        connect _bundle_literal_expr_23.src, _array_literal_expr_1
                                        wire _bundle_literal_expr_24: Ty48
                                        connect _bundle_literal_expr_24.imm_low, _match_arm_value_16.alu_common.common.imm_low
                                        wire _array_literal_expr_5: UInt<8>[0]
                                        invalidate _array_literal_expr_5
                                        connect _bundle_literal_expr_24.reversed_src, _array_literal_expr_5
                                        connect _bundle_literal_expr_24.imm_sign, _match_arm_value_16.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_9: Ty49
                                        connect _cast_bundle_to_bits_expr_9.imm_low, _bundle_literal_expr_24.imm_low
                                        connect _cast_bundle_to_bits_expr_9.reversed_src, UInt<0>(0)
                                        connect _cast_bundle_to_bits_expr_9.imm_sign, asUInt(_bundle_literal_expr_24.imm_sign)
                                        wire _cast_to_bits_expr_9: UInt<26>
                                        connect _cast_to_bits_expr_9, cat(_cast_bundle_to_bits_expr_9.imm_sign, cat(_cast_bundle_to_bits_expr_9.reversed_src, _cast_bundle_to_bits_expr_9.imm_low))
                                        connect _bundle_literal_expr_23.imm_low, bits(asSInt(_cast_to_bits_expr_9), 24, 0)
                                        wire _bundle_literal_expr_25: Ty48
                                        connect _bundle_literal_expr_25.imm_low, _match_arm_value_16.alu_common.common.imm_low
                                        wire _array_literal_expr_6: UInt<8>[0]
                                        invalidate _array_literal_expr_6
                                        connect _bundle_literal_expr_25.reversed_src, _array_literal_expr_6
                                        connect _bundle_literal_expr_25.imm_sign, _match_arm_value_16.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_10: Ty49
                                        connect _cast_bundle_to_bits_expr_10.imm_low, _bundle_literal_expr_25.imm_low
                                        connect _cast_bundle_to_bits_expr_10.reversed_src, UInt<0>(0)
                                        connect _cast_bundle_to_bits_expr_10.imm_sign, asUInt(_bundle_literal_expr_25.imm_sign)
                                        wire _cast_to_bits_expr_10: UInt<26>
                                        connect _cast_to_bits_expr_10, cat(_cast_bundle_to_bits_expr_10.imm_sign, cat(_cast_bundle_to_bits_expr_10.reversed_src, _cast_bundle_to_bits_expr_10.imm_low))
                                        connect _bundle_literal_expr_23.imm_sign, shr(asSInt(_cast_to_bits_expr_10), 25)
                                        wire _bundle_literal_expr_26: Ty2
                                        invalidate _bundle_literal_expr_26
                                        connect _bundle_literal_expr_23._phantom, _bundle_literal_expr_26
                                        connect _bundle_literal_expr_22.common, _bundle_literal_expr_23
                                        connect _bundle_literal_expr_22.output_integer_mode, _match_arm_value_16.alu_common.output_integer_mode
                                        connect _bundle_literal_expr_21.alu_common, _bundle_literal_expr_22
                                        connect _bundle_literal_expr_21.invert_src0, _match_arm_value_16.invert_src0
                                        connect _bundle_literal_expr_21.invert_carry_in, _match_arm_value_16.invert_carry_in
                                        connect _bundle_literal_expr_21.invert_carry_out, _match_arm_value_16.invert_carry_out
                                        connect _bundle_literal_expr_21.add_pc, _match_arm_value_16.add_pc
                                        connect mapped_regs_1, {|AddSub: Ty35, AddSubI: Ty35, Logical: Ty36|}(AddSub, _bundle_literal_expr_21) @[instruction.rs 477:1]
                                    AddSubI(_match_arm_value_23):
                                        wire renamed_src_reg_0_0_1: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_27: Ty1
                                        connect _bundle_literal_expr_27.value, _match_arm_value_23.alu_common.common.src[0]
                                        connect rename_table_normal_0_src_0.addr, _bundle_literal_expr_27 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_28: Ty1
                                        connect _bundle_literal_expr_28.value, _match_arm_value_23.alu_common.common.src[0]
                                        connect rename_table_special_0_src_0.addr, _bundle_literal_expr_28 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_0_src_0.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_0_src_0.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_29: Ty25
                                                        wire _bundle_literal_expr_30: Ty23
                                                        connect _bundle_literal_expr_30.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_29.unit_num, _bundle_literal_expr_30
                                                        wire _bundle_literal_expr_31: Ty24
                                                        connect _bundle_literal_expr_31.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_29.unit_out_reg, _bundle_literal_expr_31
                                                        connect renamed_src_reg_0_0_1, _bundle_literal_expr_29 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_24):
                                                        connect renamed_src_reg_0_0_1, _match_arm_value_24 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_25):
                                                connect renamed_src_reg_0_0_1, _match_arm_value_25 @[reg_alloc.rs 185:33]
                                        wire renamed_src_reg_0_1_1: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_32: Ty1
                                        connect _bundle_literal_expr_32.value, _match_arm_value_23.alu_common.common.src[1]
                                        connect rename_table_normal_0_src_1.addr, _bundle_literal_expr_32 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_33: Ty1
                                        connect _bundle_literal_expr_33.value, _match_arm_value_23.alu_common.common.src[1]
                                        connect rename_table_special_0_src_1.addr, _bundle_literal_expr_33 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_0_src_1.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_0_src_1.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_34: Ty25
                                                        wire _bundle_literal_expr_35: Ty23
                                                        connect _bundle_literal_expr_35.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_34.unit_num, _bundle_literal_expr_35
                                                        wire _bundle_literal_expr_36: Ty24
                                                        connect _bundle_literal_expr_36.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_34.unit_out_reg, _bundle_literal_expr_36
                                                        connect renamed_src_reg_0_1_1, _bundle_literal_expr_34 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_26):
                                                        connect renamed_src_reg_0_1_1, _match_arm_value_26 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_27):
                                                connect renamed_src_reg_0_1_1, _match_arm_value_27 @[reg_alloc.rs 185:33]
                                        wire _bundle_literal_expr_37: Ty35
                                        wire _bundle_literal_expr_38: Ty34
                                        wire _bundle_literal_expr_39: Ty33
                                        connect _bundle_literal_expr_39.prefix_pad, _match_arm_value_23.alu_common.common.prefix_pad
                                        connect _bundle_literal_expr_39.dest, _match_arm_value_4.unit_out_reg
                                        wire _array_literal_expr_7: UInt<6>[3]
                                        wire _array_literal_expr_8: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_11: Ty47
                                        connect _cast_bundle_to_bits_expr_11.unit_num, renamed_src_reg_0_0_1.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_11.unit_out_reg, renamed_src_reg_0_0_1.unit_out_reg.value
                                        wire _cast_to_bits_expr_11: UInt<6>
                                        connect _cast_to_bits_expr_11, cat(_cast_bundle_to_bits_expr_11.unit_out_reg, _cast_bundle_to_bits_expr_11.unit_num)
                                        connect _array_literal_expr_8[0], _cast_to_bits_expr_11
                                        wire _cast_bundle_to_bits_expr_12: Ty47
                                        connect _cast_bundle_to_bits_expr_12.unit_num, renamed_src_reg_0_1_1.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_12.unit_out_reg, renamed_src_reg_0_1_1.unit_out_reg.value
                                        wire _cast_to_bits_expr_12: UInt<6>
                                        connect _cast_to_bits_expr_12, cat(_cast_bundle_to_bits_expr_12.unit_out_reg, _cast_bundle_to_bits_expr_12.unit_num)
                                        connect _array_literal_expr_8[1], _cast_to_bits_expr_12
                                        connect _array_literal_expr_7[0], _array_literal_expr_8[0]
                                        wire _array_literal_expr_9: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_13: Ty47
                                        connect _cast_bundle_to_bits_expr_13.unit_num, renamed_src_reg_0_0_1.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_13.unit_out_reg, renamed_src_reg_0_0_1.unit_out_reg.value
                                        wire _cast_to_bits_expr_13: UInt<6>
                                        connect _cast_to_bits_expr_13, cat(_cast_bundle_to_bits_expr_13.unit_out_reg, _cast_bundle_to_bits_expr_13.unit_num)
                                        connect _array_literal_expr_9[0], _cast_to_bits_expr_13
                                        wire _cast_bundle_to_bits_expr_14: Ty47
                                        connect _cast_bundle_to_bits_expr_14.unit_num, renamed_src_reg_0_1_1.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_14.unit_out_reg, renamed_src_reg_0_1_1.unit_out_reg.value
                                        wire _cast_to_bits_expr_14: UInt<6>
                                        connect _cast_to_bits_expr_14, cat(_cast_bundle_to_bits_expr_14.unit_out_reg, _cast_bundle_to_bits_expr_14.unit_num)
                                        connect _array_literal_expr_9[1], _cast_to_bits_expr_14
                                        connect _array_literal_expr_7[1], _array_literal_expr_9[1]
                                        wire _bundle_literal_expr_40: Ty50
                                        connect _bundle_literal_expr_40.imm_low, _match_arm_value_23.alu_common.common.imm_low
                                        wire _array_literal_expr_10: UInt<8>[1]
                                        connect _array_literal_expr_10[0], _match_arm_value_23.alu_common.common.src[2]
                                        connect _bundle_literal_expr_40.reversed_src, _array_literal_expr_10
                                        connect _bundle_literal_expr_40.imm_sign, _match_arm_value_23.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_15: Ty51
                                        connect _cast_bundle_to_bits_expr_15.imm_low, _bundle_literal_expr_40.imm_low
                                        connect _cast_bundle_to_bits_expr_15.reversed_src, _bundle_literal_expr_40.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_15.imm_sign, asUInt(_bundle_literal_expr_40.imm_sign)
                                        wire _cast_to_bits_expr_15: UInt<34>
                                        connect _cast_to_bits_expr_15, cat(_cast_bundle_to_bits_expr_15.imm_sign, cat(_cast_bundle_to_bits_expr_15.reversed_src, _cast_bundle_to_bits_expr_15.imm_low))
                                        wire _cast_bits_to_bundle_expr: Ty50
                                        wire _cast_bits_to_bundle_expr_flattened: Ty51
                                        connect _cast_bits_to_bundle_expr_flattened.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_15)), 24, 0)
                                        connect _cast_bits_to_bundle_expr.imm_low, _cast_bits_to_bundle_expr_flattened.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_15)), 32, 25)
                                        wire _cast_bits_to_array_expr: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened[0], bits(_cast_bits_to_bundle_expr_flattened.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr[0], _cast_bits_to_array_expr_flattened[0]
                                        connect _cast_bits_to_bundle_expr.reversed_src, _cast_bits_to_array_expr
                                        connect _cast_bits_to_bundle_expr_flattened.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_15)), 33, 33)
                                        connect _cast_bits_to_bundle_expr.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened.imm_sign)
                                        connect _array_literal_expr_7[2], tail(_cast_bits_to_bundle_expr.reversed_src[0], 2)
                                        connect _bundle_literal_expr_39.src, _array_literal_expr_7
                                        wire _bundle_literal_expr_41: Ty50
                                        connect _bundle_literal_expr_41.imm_low, _match_arm_value_23.alu_common.common.imm_low
                                        wire _array_literal_expr_11: UInt<8>[1]
                                        connect _array_literal_expr_11[0], _match_arm_value_23.alu_common.common.src[2]
                                        connect _bundle_literal_expr_41.reversed_src, _array_literal_expr_11
                                        connect _bundle_literal_expr_41.imm_sign, _match_arm_value_23.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_16: Ty51
                                        connect _cast_bundle_to_bits_expr_16.imm_low, _bundle_literal_expr_41.imm_low
                                        connect _cast_bundle_to_bits_expr_16.reversed_src, _bundle_literal_expr_41.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_16.imm_sign, asUInt(_bundle_literal_expr_41.imm_sign)
                                        wire _cast_to_bits_expr_16: UInt<34>
                                        connect _cast_to_bits_expr_16, cat(_cast_bundle_to_bits_expr_16.imm_sign, cat(_cast_bundle_to_bits_expr_16.reversed_src, _cast_bundle_to_bits_expr_16.imm_low))
                                        connect _bundle_literal_expr_39.imm_low, bits(asSInt(_cast_to_bits_expr_16), 24, 0)
                                        wire _bundle_literal_expr_42: Ty50
                                        connect _bundle_literal_expr_42.imm_low, _match_arm_value_23.alu_common.common.imm_low
                                        wire _array_literal_expr_12: UInt<8>[1]
                                        connect _array_literal_expr_12[0], _match_arm_value_23.alu_common.common.src[2]
                                        connect _bundle_literal_expr_42.reversed_src, _array_literal_expr_12
                                        connect _bundle_literal_expr_42.imm_sign, _match_arm_value_23.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_17: Ty51
                                        connect _cast_bundle_to_bits_expr_17.imm_low, _bundle_literal_expr_42.imm_low
                                        connect _cast_bundle_to_bits_expr_17.reversed_src, _bundle_literal_expr_42.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_17.imm_sign, asUInt(_bundle_literal_expr_42.imm_sign)
                                        wire _cast_to_bits_expr_17: UInt<34>
                                        connect _cast_to_bits_expr_17, cat(_cast_bundle_to_bits_expr_17.imm_sign, cat(_cast_bundle_to_bits_expr_17.reversed_src, _cast_bundle_to_bits_expr_17.imm_low))
                                        connect _bundle_literal_expr_39.imm_sign, shr(asSInt(_cast_to_bits_expr_17), 33)
                                        wire _bundle_literal_expr_43: Ty2
                                        invalidate _bundle_literal_expr_43
                                        connect _bundle_literal_expr_39._phantom, _bundle_literal_expr_43
                                        connect _bundle_literal_expr_38.common, _bundle_literal_expr_39
                                        connect _bundle_literal_expr_38.output_integer_mode, _match_arm_value_23.alu_common.output_integer_mode
                                        connect _bundle_literal_expr_37.alu_common, _bundle_literal_expr_38
                                        connect _bundle_literal_expr_37.invert_src0, _match_arm_value_23.invert_src0
                                        connect _bundle_literal_expr_37.invert_carry_in, _match_arm_value_23.invert_carry_in
                                        connect _bundle_literal_expr_37.invert_carry_out, _match_arm_value_23.invert_carry_out
                                        connect _bundle_literal_expr_37.add_pc, _match_arm_value_23.add_pc
                                        connect mapped_regs_1, {|AddSub: Ty35, AddSubI: Ty35, Logical: Ty36|}(AddSubI, _bundle_literal_expr_37) @[instruction.rs 477:1]
                                    Logical(_match_arm_value_28):
                                        wire renamed_src_reg_0_0_2: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_44: Ty1
                                        connect _bundle_literal_expr_44.value, _match_arm_value_28.alu_common.common.src[0]
                                        connect rename_table_normal_0_src_0.addr, _bundle_literal_expr_44 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_45: Ty1
                                        connect _bundle_literal_expr_45.value, _match_arm_value_28.alu_common.common.src[0]
                                        connect rename_table_special_0_src_0.addr, _bundle_literal_expr_45 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_0_src_0.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_0_src_0.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_46: Ty25
                                                        wire _bundle_literal_expr_47: Ty23
                                                        connect _bundle_literal_expr_47.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_46.unit_num, _bundle_literal_expr_47
                                                        wire _bundle_literal_expr_48: Ty24
                                                        connect _bundle_literal_expr_48.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_46.unit_out_reg, _bundle_literal_expr_48
                                                        connect renamed_src_reg_0_0_2, _bundle_literal_expr_46 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_29):
                                                        connect renamed_src_reg_0_0_2, _match_arm_value_29 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_30):
                                                connect renamed_src_reg_0_0_2, _match_arm_value_30 @[reg_alloc.rs 185:33]
                                        wire renamed_src_reg_0_1_2: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_49: Ty1
                                        connect _bundle_literal_expr_49.value, _match_arm_value_28.alu_common.common.src[1]
                                        connect rename_table_normal_0_src_1.addr, _bundle_literal_expr_49 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_50: Ty1
                                        connect _bundle_literal_expr_50.value, _match_arm_value_28.alu_common.common.src[1]
                                        connect rename_table_special_0_src_1.addr, _bundle_literal_expr_50 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_0_src_1.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_0_src_1.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_51: Ty25
                                                        wire _bundle_literal_expr_52: Ty23
                                                        connect _bundle_literal_expr_52.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_51.unit_num, _bundle_literal_expr_52
                                                        wire _bundle_literal_expr_53: Ty24
                                                        connect _bundle_literal_expr_53.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_51.unit_out_reg, _bundle_literal_expr_53
                                                        connect renamed_src_reg_0_1_2, _bundle_literal_expr_51 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_31):
                                                        connect renamed_src_reg_0_1_2, _match_arm_value_31 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_32):
                                                connect renamed_src_reg_0_1_2, _match_arm_value_32 @[reg_alloc.rs 185:33]
                                        wire _bundle_literal_expr_54: Ty36
                                        wire _bundle_literal_expr_55: Ty34
                                        wire _bundle_literal_expr_56: Ty33
                                        connect _bundle_literal_expr_56.prefix_pad, _match_arm_value_28.alu_common.common.prefix_pad
                                        connect _bundle_literal_expr_56.dest, _match_arm_value_4.unit_out_reg
                                        wire _array_literal_expr_13: UInt<6>[3]
                                        wire _array_literal_expr_14: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_18: Ty47
                                        connect _cast_bundle_to_bits_expr_18.unit_num, renamed_src_reg_0_0_2.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_18.unit_out_reg, renamed_src_reg_0_0_2.unit_out_reg.value
                                        wire _cast_to_bits_expr_18: UInt<6>
                                        connect _cast_to_bits_expr_18, cat(_cast_bundle_to_bits_expr_18.unit_out_reg, _cast_bundle_to_bits_expr_18.unit_num)
                                        connect _array_literal_expr_14[0], _cast_to_bits_expr_18
                                        wire _cast_bundle_to_bits_expr_19: Ty47
                                        connect _cast_bundle_to_bits_expr_19.unit_num, renamed_src_reg_0_1_2.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_19.unit_out_reg, renamed_src_reg_0_1_2.unit_out_reg.value
                                        wire _cast_to_bits_expr_19: UInt<6>
                                        connect _cast_to_bits_expr_19, cat(_cast_bundle_to_bits_expr_19.unit_out_reg, _cast_bundle_to_bits_expr_19.unit_num)
                                        connect _array_literal_expr_14[1], _cast_to_bits_expr_19
                                        connect _array_literal_expr_13[0], _array_literal_expr_14[0]
                                        wire _array_literal_expr_15: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_20: Ty47
                                        connect _cast_bundle_to_bits_expr_20.unit_num, renamed_src_reg_0_0_2.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_20.unit_out_reg, renamed_src_reg_0_0_2.unit_out_reg.value
                                        wire _cast_to_bits_expr_20: UInt<6>
                                        connect _cast_to_bits_expr_20, cat(_cast_bundle_to_bits_expr_20.unit_out_reg, _cast_bundle_to_bits_expr_20.unit_num)
                                        connect _array_literal_expr_15[0], _cast_to_bits_expr_20
                                        wire _cast_bundle_to_bits_expr_21: Ty47
                                        connect _cast_bundle_to_bits_expr_21.unit_num, renamed_src_reg_0_1_2.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_21.unit_out_reg, renamed_src_reg_0_1_2.unit_out_reg.value
                                        wire _cast_to_bits_expr_21: UInt<6>
                                        connect _cast_to_bits_expr_21, cat(_cast_bundle_to_bits_expr_21.unit_out_reg, _cast_bundle_to_bits_expr_21.unit_num)
                                        connect _array_literal_expr_15[1], _cast_to_bits_expr_21
                                        connect _array_literal_expr_13[1], _array_literal_expr_15[1]
                                        wire _bundle_literal_expr_57: Ty50
                                        connect _bundle_literal_expr_57.imm_low, _match_arm_value_28.alu_common.common.imm_low
                                        wire _array_literal_expr_16: UInt<8>[1]
                                        connect _array_literal_expr_16[0], _match_arm_value_28.alu_common.common.src[2]
                                        connect _bundle_literal_expr_57.reversed_src, _array_literal_expr_16
                                        connect _bundle_literal_expr_57.imm_sign, _match_arm_value_28.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_22: Ty51
                                        connect _cast_bundle_to_bits_expr_22.imm_low, _bundle_literal_expr_57.imm_low
                                        connect _cast_bundle_to_bits_expr_22.reversed_src, _bundle_literal_expr_57.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_22.imm_sign, asUInt(_bundle_literal_expr_57.imm_sign)
                                        wire _cast_to_bits_expr_22: UInt<34>
                                        connect _cast_to_bits_expr_22, cat(_cast_bundle_to_bits_expr_22.imm_sign, cat(_cast_bundle_to_bits_expr_22.reversed_src, _cast_bundle_to_bits_expr_22.imm_low))
                                        wire _cast_bits_to_bundle_expr_1: Ty50
                                        wire _cast_bits_to_bundle_expr_flattened_1: Ty51
                                        connect _cast_bits_to_bundle_expr_flattened_1.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_22)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_1.imm_low, _cast_bits_to_bundle_expr_flattened_1.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_1.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_22)), 32, 25)
                                        wire _cast_bits_to_array_expr_1: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_1: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_1[0], bits(_cast_bits_to_bundle_expr_flattened_1.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_1[0], _cast_bits_to_array_expr_flattened_1[0]
                                        connect _cast_bits_to_bundle_expr_1.reversed_src, _cast_bits_to_array_expr_1
                                        connect _cast_bits_to_bundle_expr_flattened_1.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_22)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_1.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_1.imm_sign)
                                        connect _array_literal_expr_13[2], tail(_cast_bits_to_bundle_expr_1.reversed_src[0], 2)
                                        connect _bundle_literal_expr_56.src, _array_literal_expr_13
                                        wire _bundle_literal_expr_58: Ty50
                                        connect _bundle_literal_expr_58.imm_low, _match_arm_value_28.alu_common.common.imm_low
                                        wire _array_literal_expr_17: UInt<8>[1]
                                        connect _array_literal_expr_17[0], _match_arm_value_28.alu_common.common.src[2]
                                        connect _bundle_literal_expr_58.reversed_src, _array_literal_expr_17
                                        connect _bundle_literal_expr_58.imm_sign, _match_arm_value_28.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_23: Ty51
                                        connect _cast_bundle_to_bits_expr_23.imm_low, _bundle_literal_expr_58.imm_low
                                        connect _cast_bundle_to_bits_expr_23.reversed_src, _bundle_literal_expr_58.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_23.imm_sign, asUInt(_bundle_literal_expr_58.imm_sign)
                                        wire _cast_to_bits_expr_23: UInt<34>
                                        connect _cast_to_bits_expr_23, cat(_cast_bundle_to_bits_expr_23.imm_sign, cat(_cast_bundle_to_bits_expr_23.reversed_src, _cast_bundle_to_bits_expr_23.imm_low))
                                        connect _bundle_literal_expr_56.imm_low, bits(asSInt(_cast_to_bits_expr_23), 24, 0)
                                        wire _bundle_literal_expr_59: Ty50
                                        connect _bundle_literal_expr_59.imm_low, _match_arm_value_28.alu_common.common.imm_low
                                        wire _array_literal_expr_18: UInt<8>[1]
                                        connect _array_literal_expr_18[0], _match_arm_value_28.alu_common.common.src[2]
                                        connect _bundle_literal_expr_59.reversed_src, _array_literal_expr_18
                                        connect _bundle_literal_expr_59.imm_sign, _match_arm_value_28.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_24: Ty51
                                        connect _cast_bundle_to_bits_expr_24.imm_low, _bundle_literal_expr_59.imm_low
                                        connect _cast_bundle_to_bits_expr_24.reversed_src, _bundle_literal_expr_59.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_24.imm_sign, asUInt(_bundle_literal_expr_59.imm_sign)
                                        wire _cast_to_bits_expr_24: UInt<34>
                                        connect _cast_to_bits_expr_24, cat(_cast_bundle_to_bits_expr_24.imm_sign, cat(_cast_bundle_to_bits_expr_24.reversed_src, _cast_bundle_to_bits_expr_24.imm_low))
                                        connect _bundle_literal_expr_56.imm_sign, shr(asSInt(_cast_to_bits_expr_24), 33)
                                        wire _bundle_literal_expr_60: Ty2
                                        invalidate _bundle_literal_expr_60
                                        connect _bundle_literal_expr_56._phantom, _bundle_literal_expr_60
                                        connect _bundle_literal_expr_55.common, _bundle_literal_expr_56
                                        connect _bundle_literal_expr_55.output_integer_mode, _match_arm_value_28.alu_common.output_integer_mode
                                        connect _bundle_literal_expr_54.alu_common, _bundle_literal_expr_55
                                        connect _bundle_literal_expr_54.lut, _match_arm_value_28.lut
                                        connect mapped_regs_1, {|AddSub: Ty35, AddSubI: Ty35, Logical: Ty36|}(Logical, _bundle_literal_expr_54) @[instruction.rs 477:1]
                                connect mapped_regs, {|AluBranch: Ty37, L2RegisterFile: Ty40, LoadStore: Ty41|}(AluBranch, mapped_regs_1) @[unit.rs 127:1]
                            L2RegisterFile(_match_arm_value_33):
                                wire mapped_regs_2: Ty40 @[instruction.rs 504:1]
                                match _match_arm_value_33: @[instruction.rs 504:1]
                                    ReadL2Reg(_match_arm_value_34):
                                        wire _bundle_literal_expr_61: Ty39
                                        wire _bundle_literal_expr_62: Ty38
                                        connect _bundle_literal_expr_62.prefix_pad, _match_arm_value_34.common.prefix_pad
                                        connect _bundle_literal_expr_62.dest, _match_arm_value_4.unit_out_reg
                                        wire _array_literal_expr_19: UInt<6>[3]
                                        connect _array_literal_expr_19[0], pad(UInt<0>(0h0), 6)
                                        connect _array_literal_expr_19[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_63: Ty50
                                        connect _bundle_literal_expr_63.imm_low, _match_arm_value_34.common.imm_low
                                        wire _array_literal_expr_20: UInt<8>[1]
                                        connect _array_literal_expr_20[0], _match_arm_value_34.common.src[2]
                                        connect _bundle_literal_expr_63.reversed_src, _array_literal_expr_20
                                        connect _bundle_literal_expr_63.imm_sign, _match_arm_value_34.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_25: Ty51
                                        connect _cast_bundle_to_bits_expr_25.imm_low, _bundle_literal_expr_63.imm_low
                                        connect _cast_bundle_to_bits_expr_25.reversed_src, _bundle_literal_expr_63.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_25.imm_sign, asUInt(_bundle_literal_expr_63.imm_sign)
                                        wire _cast_to_bits_expr_25: UInt<34>
                                        connect _cast_to_bits_expr_25, cat(_cast_bundle_to_bits_expr_25.imm_sign, cat(_cast_bundle_to_bits_expr_25.reversed_src, _cast_bundle_to_bits_expr_25.imm_low))
                                        wire _cast_bits_to_bundle_expr_2: Ty50
                                        wire _cast_bits_to_bundle_expr_flattened_2: Ty51
                                        connect _cast_bits_to_bundle_expr_flattened_2.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_25)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_2.imm_low, _cast_bits_to_bundle_expr_flattened_2.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_2.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_25)), 32, 25)
                                        wire _cast_bits_to_array_expr_2: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_2: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_2[0], bits(_cast_bits_to_bundle_expr_flattened_2.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_2[0], _cast_bits_to_array_expr_flattened_2[0]
                                        connect _cast_bits_to_bundle_expr_2.reversed_src, _cast_bits_to_array_expr_2
                                        connect _cast_bits_to_bundle_expr_flattened_2.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_25)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_2.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_2.imm_sign)
                                        connect _array_literal_expr_19[2], tail(_cast_bits_to_bundle_expr_2.reversed_src[0], 2)
                                        connect _bundle_literal_expr_62.src, _array_literal_expr_19
                                        wire _bundle_literal_expr_64: Ty50
                                        connect _bundle_literal_expr_64.imm_low, _match_arm_value_34.common.imm_low
                                        wire _array_literal_expr_21: UInt<8>[1]
                                        connect _array_literal_expr_21[0], _match_arm_value_34.common.src[2]
                                        connect _bundle_literal_expr_64.reversed_src, _array_literal_expr_21
                                        connect _bundle_literal_expr_64.imm_sign, _match_arm_value_34.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_26: Ty51
                                        connect _cast_bundle_to_bits_expr_26.imm_low, _bundle_literal_expr_64.imm_low
                                        connect _cast_bundle_to_bits_expr_26.reversed_src, _bundle_literal_expr_64.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_26.imm_sign, asUInt(_bundle_literal_expr_64.imm_sign)
                                        wire _cast_to_bits_expr_26: UInt<34>
                                        connect _cast_to_bits_expr_26, cat(_cast_bundle_to_bits_expr_26.imm_sign, cat(_cast_bundle_to_bits_expr_26.reversed_src, _cast_bundle_to_bits_expr_26.imm_low))
                                        connect _bundle_literal_expr_62.imm_low, bits(asSInt(_cast_to_bits_expr_26), 24, 0)
                                        wire _bundle_literal_expr_65: Ty50
                                        connect _bundle_literal_expr_65.imm_low, _match_arm_value_34.common.imm_low
                                        wire _array_literal_expr_22: UInt<8>[1]
                                        connect _array_literal_expr_22[0], _match_arm_value_34.common.src[2]
                                        connect _bundle_literal_expr_65.reversed_src, _array_literal_expr_22
                                        connect _bundle_literal_expr_65.imm_sign, _match_arm_value_34.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_27: Ty51
                                        connect _cast_bundle_to_bits_expr_27.imm_low, _bundle_literal_expr_65.imm_low
                                        connect _cast_bundle_to_bits_expr_27.reversed_src, _bundle_literal_expr_65.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_27.imm_sign, asUInt(_bundle_literal_expr_65.imm_sign)
                                        wire _cast_to_bits_expr_27: UInt<34>
                                        connect _cast_to_bits_expr_27, cat(_cast_bundle_to_bits_expr_27.imm_sign, cat(_cast_bundle_to_bits_expr_27.reversed_src, _cast_bundle_to_bits_expr_27.imm_low))
                                        connect _bundle_literal_expr_62.imm_sign, shr(asSInt(_cast_to_bits_expr_27), 33)
                                        wire _bundle_literal_expr_66: Ty2
                                        invalidate _bundle_literal_expr_66
                                        connect _bundle_literal_expr_62._phantom, _bundle_literal_expr_66
                                        connect _bundle_literal_expr_61.common, _bundle_literal_expr_62
                                        connect mapped_regs_2, {|ReadL2Reg: Ty39, WriteL2Reg: Ty39|}(ReadL2Reg, _bundle_literal_expr_61) @[instruction.rs 504:1]
                                    WriteL2Reg(_match_arm_value_35):
                                        wire renamed_src_reg_0_0_3: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_67: Ty1
                                        connect _bundle_literal_expr_67.value, _match_arm_value_35.common.src[0]
                                        connect rename_table_normal_0_src_0.addr, _bundle_literal_expr_67 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_68: Ty1
                                        connect _bundle_literal_expr_68.value, _match_arm_value_35.common.src[0]
                                        connect rename_table_special_0_src_0.addr, _bundle_literal_expr_68 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_0_src_0.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_0_src_0.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_69: Ty25
                                                        wire _bundle_literal_expr_70: Ty23
                                                        connect _bundle_literal_expr_70.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_69.unit_num, _bundle_literal_expr_70
                                                        wire _bundle_literal_expr_71: Ty24
                                                        connect _bundle_literal_expr_71.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_69.unit_out_reg, _bundle_literal_expr_71
                                                        connect renamed_src_reg_0_0_3, _bundle_literal_expr_69 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_36):
                                                        connect renamed_src_reg_0_0_3, _match_arm_value_36 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_37):
                                                connect renamed_src_reg_0_0_3, _match_arm_value_37 @[reg_alloc.rs 185:33]
                                        wire _bundle_literal_expr_72: Ty39
                                        wire _bundle_literal_expr_73: Ty38
                                        connect _bundle_literal_expr_73.prefix_pad, _match_arm_value_35.common.prefix_pad
                                        connect _bundle_literal_expr_73.dest, _match_arm_value_4.unit_out_reg
                                        wire _array_literal_expr_23: UInt<6>[3]
                                        wire _array_literal_expr_24: UInt<6>[1]
                                        wire _cast_bundle_to_bits_expr_28: Ty47
                                        connect _cast_bundle_to_bits_expr_28.unit_num, renamed_src_reg_0_0_3.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_28.unit_out_reg, renamed_src_reg_0_0_3.unit_out_reg.value
                                        wire _cast_to_bits_expr_28: UInt<6>
                                        connect _cast_to_bits_expr_28, cat(_cast_bundle_to_bits_expr_28.unit_out_reg, _cast_bundle_to_bits_expr_28.unit_num)
                                        connect _array_literal_expr_24[0], _cast_to_bits_expr_28
                                        connect _array_literal_expr_23[0], _array_literal_expr_24[0]
                                        connect _array_literal_expr_23[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_74: Ty50
                                        connect _bundle_literal_expr_74.imm_low, _match_arm_value_35.common.imm_low
                                        wire _array_literal_expr_25: UInt<8>[1]
                                        connect _array_literal_expr_25[0], _match_arm_value_35.common.src[2]
                                        connect _bundle_literal_expr_74.reversed_src, _array_literal_expr_25
                                        connect _bundle_literal_expr_74.imm_sign, _match_arm_value_35.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_29: Ty51
                                        connect _cast_bundle_to_bits_expr_29.imm_low, _bundle_literal_expr_74.imm_low
                                        connect _cast_bundle_to_bits_expr_29.reversed_src, _bundle_literal_expr_74.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_29.imm_sign, asUInt(_bundle_literal_expr_74.imm_sign)
                                        wire _cast_to_bits_expr_29: UInt<34>
                                        connect _cast_to_bits_expr_29, cat(_cast_bundle_to_bits_expr_29.imm_sign, cat(_cast_bundle_to_bits_expr_29.reversed_src, _cast_bundle_to_bits_expr_29.imm_low))
                                        wire _cast_bits_to_bundle_expr_3: Ty50
                                        wire _cast_bits_to_bundle_expr_flattened_3: Ty51
                                        connect _cast_bits_to_bundle_expr_flattened_3.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_29)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_3.imm_low, _cast_bits_to_bundle_expr_flattened_3.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_3.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_29)), 32, 25)
                                        wire _cast_bits_to_array_expr_3: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_3: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_3[0], bits(_cast_bits_to_bundle_expr_flattened_3.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_3[0], _cast_bits_to_array_expr_flattened_3[0]
                                        connect _cast_bits_to_bundle_expr_3.reversed_src, _cast_bits_to_array_expr_3
                                        connect _cast_bits_to_bundle_expr_flattened_3.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_29)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_3.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_3.imm_sign)
                                        connect _array_literal_expr_23[2], tail(_cast_bits_to_bundle_expr_3.reversed_src[0], 2)
                                        connect _bundle_literal_expr_73.src, _array_literal_expr_23
                                        wire _bundle_literal_expr_75: Ty50
                                        connect _bundle_literal_expr_75.imm_low, _match_arm_value_35.common.imm_low
                                        wire _array_literal_expr_26: UInt<8>[1]
                                        connect _array_literal_expr_26[0], _match_arm_value_35.common.src[2]
                                        connect _bundle_literal_expr_75.reversed_src, _array_literal_expr_26
                                        connect _bundle_literal_expr_75.imm_sign, _match_arm_value_35.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_30: Ty51
                                        connect _cast_bundle_to_bits_expr_30.imm_low, _bundle_literal_expr_75.imm_low
                                        connect _cast_bundle_to_bits_expr_30.reversed_src, _bundle_literal_expr_75.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_30.imm_sign, asUInt(_bundle_literal_expr_75.imm_sign)
                                        wire _cast_to_bits_expr_30: UInt<34>
                                        connect _cast_to_bits_expr_30, cat(_cast_bundle_to_bits_expr_30.imm_sign, cat(_cast_bundle_to_bits_expr_30.reversed_src, _cast_bundle_to_bits_expr_30.imm_low))
                                        connect _bundle_literal_expr_73.imm_low, bits(asSInt(_cast_to_bits_expr_30), 24, 0)
                                        wire _bundle_literal_expr_76: Ty50
                                        connect _bundle_literal_expr_76.imm_low, _match_arm_value_35.common.imm_low
                                        wire _array_literal_expr_27: UInt<8>[1]
                                        connect _array_literal_expr_27[0], _match_arm_value_35.common.src[2]
                                        connect _bundle_literal_expr_76.reversed_src, _array_literal_expr_27
                                        connect _bundle_literal_expr_76.imm_sign, _match_arm_value_35.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_31: Ty51
                                        connect _cast_bundle_to_bits_expr_31.imm_low, _bundle_literal_expr_76.imm_low
                                        connect _cast_bundle_to_bits_expr_31.reversed_src, _bundle_literal_expr_76.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_31.imm_sign, asUInt(_bundle_literal_expr_76.imm_sign)
                                        wire _cast_to_bits_expr_31: UInt<34>
                                        connect _cast_to_bits_expr_31, cat(_cast_bundle_to_bits_expr_31.imm_sign, cat(_cast_bundle_to_bits_expr_31.reversed_src, _cast_bundle_to_bits_expr_31.imm_low))
                                        connect _bundle_literal_expr_73.imm_sign, shr(asSInt(_cast_to_bits_expr_31), 33)
                                        wire _bundle_literal_expr_77: Ty2
                                        invalidate _bundle_literal_expr_77
                                        connect _bundle_literal_expr_73._phantom, _bundle_literal_expr_77
                                        connect _bundle_literal_expr_72.common, _bundle_literal_expr_73
                                        connect mapped_regs_2, {|ReadL2Reg: Ty39, WriteL2Reg: Ty39|}(WriteL2Reg, _bundle_literal_expr_72) @[instruction.rs 504:1]
                                connect mapped_regs, {|AluBranch: Ty37, L2RegisterFile: Ty40, LoadStore: Ty41|}(L2RegisterFile, mapped_regs_2) @[unit.rs 127:1]
                            LoadStore(_match_arm_value_38):
                                wire mapped_regs_3: Ty41 @[instruction.rs 539:1]
                                match _match_arm_value_38: @[instruction.rs 539:1]
                                    Load(_match_arm_value_39):
                                        wire _bundle_literal_expr_78: Ty38
                                        connect _bundle_literal_expr_78.prefix_pad, _match_arm_value_39.prefix_pad
                                        connect _bundle_literal_expr_78.dest, _match_arm_value_4.unit_out_reg
                                        wire _array_literal_expr_28: UInt<6>[3]
                                        connect _array_literal_expr_28[0], pad(UInt<0>(0h0), 6)
                                        connect _array_literal_expr_28[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_79: Ty50
                                        connect _bundle_literal_expr_79.imm_low, _match_arm_value_39.imm_low
                                        wire _array_literal_expr_29: UInt<8>[1]
                                        connect _array_literal_expr_29[0], _match_arm_value_39.src[2]
                                        connect _bundle_literal_expr_79.reversed_src, _array_literal_expr_29
                                        connect _bundle_literal_expr_79.imm_sign, _match_arm_value_39.imm_sign
                                        wire _cast_bundle_to_bits_expr_32: Ty51
                                        connect _cast_bundle_to_bits_expr_32.imm_low, _bundle_literal_expr_79.imm_low
                                        connect _cast_bundle_to_bits_expr_32.reversed_src, _bundle_literal_expr_79.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_32.imm_sign, asUInt(_bundle_literal_expr_79.imm_sign)
                                        wire _cast_to_bits_expr_32: UInt<34>
                                        connect _cast_to_bits_expr_32, cat(_cast_bundle_to_bits_expr_32.imm_sign, cat(_cast_bundle_to_bits_expr_32.reversed_src, _cast_bundle_to_bits_expr_32.imm_low))
                                        wire _cast_bits_to_bundle_expr_4: Ty50
                                        wire _cast_bits_to_bundle_expr_flattened_4: Ty51
                                        connect _cast_bits_to_bundle_expr_flattened_4.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_32)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_4.imm_low, _cast_bits_to_bundle_expr_flattened_4.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_4.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_32)), 32, 25)
                                        wire _cast_bits_to_array_expr_4: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_4: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_4[0], bits(_cast_bits_to_bundle_expr_flattened_4.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_4[0], _cast_bits_to_array_expr_flattened_4[0]
                                        connect _cast_bits_to_bundle_expr_4.reversed_src, _cast_bits_to_array_expr_4
                                        connect _cast_bits_to_bundle_expr_flattened_4.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_32)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_4.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_4.imm_sign)
                                        connect _array_literal_expr_28[2], tail(_cast_bits_to_bundle_expr_4.reversed_src[0], 2)
                                        connect _bundle_literal_expr_78.src, _array_literal_expr_28
                                        wire _bundle_literal_expr_80: Ty50
                                        connect _bundle_literal_expr_80.imm_low, _match_arm_value_39.imm_low
                                        wire _array_literal_expr_30: UInt<8>[1]
                                        connect _array_literal_expr_30[0], _match_arm_value_39.src[2]
                                        connect _bundle_literal_expr_80.reversed_src, _array_literal_expr_30
                                        connect _bundle_literal_expr_80.imm_sign, _match_arm_value_39.imm_sign
                                        wire _cast_bundle_to_bits_expr_33: Ty51
                                        connect _cast_bundle_to_bits_expr_33.imm_low, _bundle_literal_expr_80.imm_low
                                        connect _cast_bundle_to_bits_expr_33.reversed_src, _bundle_literal_expr_80.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_33.imm_sign, asUInt(_bundle_literal_expr_80.imm_sign)
                                        wire _cast_to_bits_expr_33: UInt<34>
                                        connect _cast_to_bits_expr_33, cat(_cast_bundle_to_bits_expr_33.imm_sign, cat(_cast_bundle_to_bits_expr_33.reversed_src, _cast_bundle_to_bits_expr_33.imm_low))
                                        connect _bundle_literal_expr_78.imm_low, bits(asSInt(_cast_to_bits_expr_33), 24, 0)
                                        wire _bundle_literal_expr_81: Ty50
                                        connect _bundle_literal_expr_81.imm_low, _match_arm_value_39.imm_low
                                        wire _array_literal_expr_31: UInt<8>[1]
                                        connect _array_literal_expr_31[0], _match_arm_value_39.src[2]
                                        connect _bundle_literal_expr_81.reversed_src, _array_literal_expr_31
                                        connect _bundle_literal_expr_81.imm_sign, _match_arm_value_39.imm_sign
                                        wire _cast_bundle_to_bits_expr_34: Ty51
                                        connect _cast_bundle_to_bits_expr_34.imm_low, _bundle_literal_expr_81.imm_low
                                        connect _cast_bundle_to_bits_expr_34.reversed_src, _bundle_literal_expr_81.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_34.imm_sign, asUInt(_bundle_literal_expr_81.imm_sign)
                                        wire _cast_to_bits_expr_34: UInt<34>
                                        connect _cast_to_bits_expr_34, cat(_cast_bundle_to_bits_expr_34.imm_sign, cat(_cast_bundle_to_bits_expr_34.reversed_src, _cast_bundle_to_bits_expr_34.imm_low))
                                        connect _bundle_literal_expr_78.imm_sign, shr(asSInt(_cast_to_bits_expr_34), 33)
                                        wire _bundle_literal_expr_82: Ty2
                                        invalidate _bundle_literal_expr_82
                                        connect _bundle_literal_expr_78._phantom, _bundle_literal_expr_82
                                        connect mapped_regs_3, {|Load: Ty38, Store: Ty38|}(Load, _bundle_literal_expr_78) @[instruction.rs 539:1]
                                    Store(_match_arm_value_40):
                                        wire renamed_src_reg_0_0_4: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_83: Ty1
                                        connect _bundle_literal_expr_83.value, _match_arm_value_40.src[0]
                                        connect rename_table_normal_0_src_0.addr, _bundle_literal_expr_83 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_84: Ty1
                                        connect _bundle_literal_expr_84.value, _match_arm_value_40.src[0]
                                        connect rename_table_special_0_src_0.addr, _bundle_literal_expr_84 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_0_src_0.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_0_src_0.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_85: Ty25
                                                        wire _bundle_literal_expr_86: Ty23
                                                        connect _bundle_literal_expr_86.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_85.unit_num, _bundle_literal_expr_86
                                                        wire _bundle_literal_expr_87: Ty24
                                                        connect _bundle_literal_expr_87.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_85.unit_out_reg, _bundle_literal_expr_87
                                                        connect renamed_src_reg_0_0_4, _bundle_literal_expr_85 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_41):
                                                        connect renamed_src_reg_0_0_4, _match_arm_value_41 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_42):
                                                connect renamed_src_reg_0_0_4, _match_arm_value_42 @[reg_alloc.rs 185:33]
                                        wire _bundle_literal_expr_88: Ty38
                                        connect _bundle_literal_expr_88.prefix_pad, _match_arm_value_40.prefix_pad
                                        connect _bundle_literal_expr_88.dest, _match_arm_value_4.unit_out_reg
                                        wire _array_literal_expr_32: UInt<6>[3]
                                        wire _array_literal_expr_33: UInt<6>[1]
                                        wire _cast_bundle_to_bits_expr_35: Ty47
                                        connect _cast_bundle_to_bits_expr_35.unit_num, renamed_src_reg_0_0_4.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_35.unit_out_reg, renamed_src_reg_0_0_4.unit_out_reg.value
                                        wire _cast_to_bits_expr_35: UInt<6>
                                        connect _cast_to_bits_expr_35, cat(_cast_bundle_to_bits_expr_35.unit_out_reg, _cast_bundle_to_bits_expr_35.unit_num)
                                        connect _array_literal_expr_33[0], _cast_to_bits_expr_35
                                        connect _array_literal_expr_32[0], _array_literal_expr_33[0]
                                        connect _array_literal_expr_32[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_89: Ty50
                                        connect _bundle_literal_expr_89.imm_low, _match_arm_value_40.imm_low
                                        wire _array_literal_expr_34: UInt<8>[1]
                                        connect _array_literal_expr_34[0], _match_arm_value_40.src[2]
                                        connect _bundle_literal_expr_89.reversed_src, _array_literal_expr_34
                                        connect _bundle_literal_expr_89.imm_sign, _match_arm_value_40.imm_sign
                                        wire _cast_bundle_to_bits_expr_36: Ty51
                                        connect _cast_bundle_to_bits_expr_36.imm_low, _bundle_literal_expr_89.imm_low
                                        connect _cast_bundle_to_bits_expr_36.reversed_src, _bundle_literal_expr_89.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_36.imm_sign, asUInt(_bundle_literal_expr_89.imm_sign)
                                        wire _cast_to_bits_expr_36: UInt<34>
                                        connect _cast_to_bits_expr_36, cat(_cast_bundle_to_bits_expr_36.imm_sign, cat(_cast_bundle_to_bits_expr_36.reversed_src, _cast_bundle_to_bits_expr_36.imm_low))
                                        wire _cast_bits_to_bundle_expr_5: Ty50
                                        wire _cast_bits_to_bundle_expr_flattened_5: Ty51
                                        connect _cast_bits_to_bundle_expr_flattened_5.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_36)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_5.imm_low, _cast_bits_to_bundle_expr_flattened_5.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_5.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_36)), 32, 25)
                                        wire _cast_bits_to_array_expr_5: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_5: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_5[0], bits(_cast_bits_to_bundle_expr_flattened_5.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_5[0], _cast_bits_to_array_expr_flattened_5[0]
                                        connect _cast_bits_to_bundle_expr_5.reversed_src, _cast_bits_to_array_expr_5
                                        connect _cast_bits_to_bundle_expr_flattened_5.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_36)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_5.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_5.imm_sign)
                                        connect _array_literal_expr_32[2], tail(_cast_bits_to_bundle_expr_5.reversed_src[0], 2)
                                        connect _bundle_literal_expr_88.src, _array_literal_expr_32
                                        wire _bundle_literal_expr_90: Ty50
                                        connect _bundle_literal_expr_90.imm_low, _match_arm_value_40.imm_low
                                        wire _array_literal_expr_35: UInt<8>[1]
                                        connect _array_literal_expr_35[0], _match_arm_value_40.src[2]
                                        connect _bundle_literal_expr_90.reversed_src, _array_literal_expr_35
                                        connect _bundle_literal_expr_90.imm_sign, _match_arm_value_40.imm_sign
                                        wire _cast_bundle_to_bits_expr_37: Ty51
                                        connect _cast_bundle_to_bits_expr_37.imm_low, _bundle_literal_expr_90.imm_low
                                        connect _cast_bundle_to_bits_expr_37.reversed_src, _bundle_literal_expr_90.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_37.imm_sign, asUInt(_bundle_literal_expr_90.imm_sign)
                                        wire _cast_to_bits_expr_37: UInt<34>
                                        connect _cast_to_bits_expr_37, cat(_cast_bundle_to_bits_expr_37.imm_sign, cat(_cast_bundle_to_bits_expr_37.reversed_src, _cast_bundle_to_bits_expr_37.imm_low))
                                        connect _bundle_literal_expr_88.imm_low, bits(asSInt(_cast_to_bits_expr_37), 24, 0)
                                        wire _bundle_literal_expr_91: Ty50
                                        connect _bundle_literal_expr_91.imm_low, _match_arm_value_40.imm_low
                                        wire _array_literal_expr_36: UInt<8>[1]
                                        connect _array_literal_expr_36[0], _match_arm_value_40.src[2]
                                        connect _bundle_literal_expr_91.reversed_src, _array_literal_expr_36
                                        connect _bundle_literal_expr_91.imm_sign, _match_arm_value_40.imm_sign
                                        wire _cast_bundle_to_bits_expr_38: Ty51
                                        connect _cast_bundle_to_bits_expr_38.imm_low, _bundle_literal_expr_91.imm_low
                                        connect _cast_bundle_to_bits_expr_38.reversed_src, _bundle_literal_expr_91.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_38.imm_sign, asUInt(_bundle_literal_expr_91.imm_sign)
                                        wire _cast_to_bits_expr_38: UInt<34>
                                        connect _cast_to_bits_expr_38, cat(_cast_bundle_to_bits_expr_38.imm_sign, cat(_cast_bundle_to_bits_expr_38.reversed_src, _cast_bundle_to_bits_expr_38.imm_low))
                                        connect _bundle_literal_expr_88.imm_sign, shr(asSInt(_cast_to_bits_expr_38), 33)
                                        wire _bundle_literal_expr_92: Ty2
                                        invalidate _bundle_literal_expr_92
                                        connect _bundle_literal_expr_88._phantom, _bundle_literal_expr_92
                                        connect mapped_regs_3, {|Load: Ty38, Store: Ty38|}(Store, _bundle_literal_expr_88) @[instruction.rs 539:1]
                                connect mapped_regs, {|AluBranch: Ty37, L2RegisterFile: Ty40, LoadStore: Ty41|}(LoadStore, mapped_regs_3) @[unit.rs 127:1]
                        connect renamed_mops[0], {|HdlNone, HdlSome: Ty42|}(HdlSome, mapped_regs) @[reg_alloc.rs 167:17]
        wire selected_unit_index_leaf_0_0: Ty32 @[reg_alloc.rs 209:25]
        connect selected_unit_index_leaf_0_0, {|HdlNone, HdlSome: UInt<2>|}(HdlNone) @[reg_alloc.rs 212:21]
        wire unit_index_0_0: UInt<2> @[reg_alloc.rs 218:25]
        ; connect different types:
        ; lhs: UInt<2>
        ; rhs: UInt<64>
        connect unit_index_0_0, UInt<64>(0h0) @[reg_alloc.rs 221:21]
        when available_units[0][0]: @[reg_alloc.rs 223:21]
            connect selected_unit_index_leaf_0_0, {|HdlNone, HdlSome: UInt<2>|}(HdlSome, unit_index_0_0) @[reg_alloc.rs 224:25]
        wire selected_unit_index_leaf_0_1: Ty32 @[reg_alloc.rs 209:25]
        connect selected_unit_index_leaf_0_1, {|HdlNone, HdlSome: UInt<2>|}(HdlNone) @[reg_alloc.rs 212:21]
        wire unit_index_0_1: UInt<2> @[reg_alloc.rs 218:25]
        ; connect different types:
        ; lhs: UInt<2>
        ; rhs: UInt<64>
        connect unit_index_0_1, UInt<64>(0h1) @[reg_alloc.rs 221:21]
        when available_units[0][1]: @[reg_alloc.rs 223:21]
            connect selected_unit_index_leaf_0_1, {|HdlNone, HdlSome: UInt<2>|}(HdlSome, unit_index_0_1) @[reg_alloc.rs 224:25]
        wire selected_unit_index_node_0_0: Ty32 @[reg_alloc.rs 231:25]
        connect selected_unit_index_node_0_0, selected_unit_index_leaf_0_0 @[reg_alloc.rs 235:21]
        match selected_unit_index_leaf_0_0: @[reg_alloc.rs 237:21]
            HdlNone:
                connect selected_unit_index_node_0_0, selected_unit_index_leaf_0_1 @[reg_alloc.rs 238:25]
            HdlSome(_match_arm_value_43):
                skip
        connect selected_unit_indexes[0], selected_unit_index_node_0_0 @[reg_alloc.rs 201:9]
        connect fetch_decode_interface.decoded_insns[1].ready, UInt<1>(0h1) @[reg_alloc.rs 79:9]
        wire _array_literal_expr_37: UInt<1>[2]
        connect _array_literal_expr_37[0], UInt<1>(0h0)
        connect _array_literal_expr_37[1], UInt<1>(0h0)
        connect available_units[1], _array_literal_expr_37 @[reg_alloc.rs 83:9]
        connect renamed_mops[1], {|HdlNone, HdlSome: Ty42|}(HdlNone) @[reg_alloc.rs 87:9]
        connect rename_table_normal_mem_r3.clk, cd.clk @[reg_alloc.rs 103:17]
        connect rename_table_normal_mem_r3.addr, UInt<8>(0h0) @[reg_alloc.rs 104:17]
        connect rename_table_normal_mem_r3.en, UInt<1>(0h0) @[reg_alloc.rs 105:17]
        wire rename_table_normal_1_src_0: Ty45 @[reg_alloc.rs 108:21]
        wire _bundle_literal_expr_93: Ty1
        connect _bundle_literal_expr_93.value, tail(UInt<32>(0h0), 24)
        connect rename_table_normal_1_src_0.addr, _bundle_literal_expr_93 @[reg_alloc.rs 111:17]
        connect rename_table_normal_1_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlNone) @[reg_alloc.rs 112:17]
        when and(geq(rename_table_normal_1_src_0.addr.value, UInt<32>(0h1)), lt(rename_table_normal_1_src_0.addr.value, UInt<32>(0hFE))): @[reg_alloc.rs 114:17]
            ; connect different types:
            ; lhs: UInt<8>
            ; rhs: UInt<33>
            connect rename_table_normal_mem_r3.addr, sub(rename_table_normal_1_src_0.addr.value, UInt<32>(0h1)) @[reg_alloc.rs 115:21]
            connect rename_table_normal_mem_r3.en, UInt<1>(0h1) @[reg_alloc.rs 116:21]
            connect rename_table_normal_1_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, rename_table_normal_mem_r3.data) @[reg_alloc.rs 117:21]
            match fetch_decode_interface.decoded_insns[0].data: @[reg_alloc.rs 120:25]
                HdlNone:
                    skip
                HdlSome(_match_arm_value_44):
                    match renamed_mops_out_reg[0]: @[reg_alloc.rs 124:29]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_45):
                            wire dest_reg_4: Ty4 @[unit.rs 127:1]
                            match _match_arm_value_44.mop: @[unit.rs 127:1]
                                AluBranch(_match_arm_value_46):
                                    wire dest_reg_5: Ty4 @[instruction.rs 477:1]
                                    match _match_arm_value_46: @[instruction.rs 477:1]
                                        AddSub(_match_arm_value_47):
                                            connect dest_reg_5, _match_arm_value_47.alu_common.common.dest @[instruction.rs 477:1]
                                        AddSubI(_match_arm_value_48):
                                            connect dest_reg_5, _match_arm_value_48.alu_common.common.dest @[instruction.rs 477:1]
                                        Logical(_match_arm_value_49):
                                            connect dest_reg_5, _match_arm_value_49.alu_common.common.dest @[instruction.rs 477:1]
                                    connect dest_reg_4, dest_reg_5 @[unit.rs 127:1]
                                L2RegisterFile(_match_arm_value_50):
                                    wire dest_reg_6: Ty4 @[instruction.rs 504:1]
                                    match _match_arm_value_50: @[instruction.rs 504:1]
                                        ReadL2Reg(_match_arm_value_51):
                                            connect dest_reg_6, _match_arm_value_51.common.dest @[instruction.rs 504:1]
                                        WriteL2Reg(_match_arm_value_52):
                                            connect dest_reg_6, _match_arm_value_52.common.dest @[instruction.rs 504:1]
                                    connect dest_reg_4, dest_reg_6 @[unit.rs 127:1]
                                LoadStore(_match_arm_value_53):
                                    wire dest_reg_7: Ty4 @[instruction.rs 539:1]
                                    match _match_arm_value_53: @[instruction.rs 539:1]
                                        Load(_match_arm_value_54):
                                            connect dest_reg_7, _match_arm_value_54.dest @[instruction.rs 539:1]
                                        Store(_match_arm_value_55):
                                            connect dest_reg_7, _match_arm_value_55.dest @[instruction.rs 539:1]
                                    connect dest_reg_4, dest_reg_7 @[unit.rs 127:1]
                            wire flag_reg: Ty1 @[instruction.rs 720:32]
                            wire _bundle_literal_expr_94: Ty1
                            connect _bundle_literal_expr_94.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg, _bundle_literal_expr_94 @[instruction.rs 721:17]
                            match dest_reg_4.flag_regs[0]: @[instruction.rs 723:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_56):
                                    wire _bundle_literal_expr_95: Ty1
                                    connect _bundle_literal_expr_95.value, tail(UInt<32>(0hFE), 24)
                                    connect flag_reg, _bundle_literal_expr_95 @[instruction.rs 725:21]
                            wire flag_reg_1: Ty1 @[instruction.rs 720:32]
                            wire _bundle_literal_expr_96: Ty1
                            connect _bundle_literal_expr_96.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_1, _bundle_literal_expr_96 @[instruction.rs 721:17]
                            match dest_reg_4.flag_regs[1]: @[instruction.rs 723:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_57):
                                    wire _bundle_literal_expr_97: Ty1
                                    connect _bundle_literal_expr_97.value, tail(UInt<32>(0hFF), 24)
                                    connect flag_reg_1, _bundle_literal_expr_97 @[instruction.rs 725:21]
                            when eq(dest_reg_4.normal_regs[0].value, rename_table_normal_1_src_0.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_normal_1_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_45) @[reg_alloc.rs 131:41]
                            when eq(dest_reg_4.normal_regs[1].value, rename_table_normal_1_src_0.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_normal_1_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_45) @[reg_alloc.rs 131:41]
                            when eq(flag_reg.value, rename_table_normal_1_src_0.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_normal_1_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_45) @[reg_alloc.rs 131:41]
                            when eq(flag_reg_1.value, rename_table_normal_1_src_0.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_normal_1_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_45) @[reg_alloc.rs 131:41]
        connect rename_table_normal_mem_r4.clk, cd.clk @[reg_alloc.rs 103:17]
        connect rename_table_normal_mem_r4.addr, UInt<8>(0h0) @[reg_alloc.rs 104:17]
        connect rename_table_normal_mem_r4.en, UInt<1>(0h0) @[reg_alloc.rs 105:17]
        wire rename_table_normal_1_src_1: Ty45 @[reg_alloc.rs 108:21]
        wire _bundle_literal_expr_98: Ty1
        connect _bundle_literal_expr_98.value, tail(UInt<32>(0h0), 24)
        connect rename_table_normal_1_src_1.addr, _bundle_literal_expr_98 @[reg_alloc.rs 111:17]
        connect rename_table_normal_1_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlNone) @[reg_alloc.rs 112:17]
        when and(geq(rename_table_normal_1_src_1.addr.value, UInt<32>(0h1)), lt(rename_table_normal_1_src_1.addr.value, UInt<32>(0hFE))): @[reg_alloc.rs 114:17]
            ; connect different types:
            ; lhs: UInt<8>
            ; rhs: UInt<33>
            connect rename_table_normal_mem_r4.addr, sub(rename_table_normal_1_src_1.addr.value, UInt<32>(0h1)) @[reg_alloc.rs 115:21]
            connect rename_table_normal_mem_r4.en, UInt<1>(0h1) @[reg_alloc.rs 116:21]
            connect rename_table_normal_1_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, rename_table_normal_mem_r4.data) @[reg_alloc.rs 117:21]
            match fetch_decode_interface.decoded_insns[0].data: @[reg_alloc.rs 120:25]
                HdlNone:
                    skip
                HdlSome(_match_arm_value_58):
                    match renamed_mops_out_reg[0]: @[reg_alloc.rs 124:29]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_59):
                            wire dest_reg_8: Ty4 @[unit.rs 127:1]
                            match _match_arm_value_58.mop: @[unit.rs 127:1]
                                AluBranch(_match_arm_value_60):
                                    wire dest_reg_9: Ty4 @[instruction.rs 477:1]
                                    match _match_arm_value_60: @[instruction.rs 477:1]
                                        AddSub(_match_arm_value_61):
                                            connect dest_reg_9, _match_arm_value_61.alu_common.common.dest @[instruction.rs 477:1]
                                        AddSubI(_match_arm_value_62):
                                            connect dest_reg_9, _match_arm_value_62.alu_common.common.dest @[instruction.rs 477:1]
                                        Logical(_match_arm_value_63):
                                            connect dest_reg_9, _match_arm_value_63.alu_common.common.dest @[instruction.rs 477:1]
                                    connect dest_reg_8, dest_reg_9 @[unit.rs 127:1]
                                L2RegisterFile(_match_arm_value_64):
                                    wire dest_reg_10: Ty4 @[instruction.rs 504:1]
                                    match _match_arm_value_64: @[instruction.rs 504:1]
                                        ReadL2Reg(_match_arm_value_65):
                                            connect dest_reg_10, _match_arm_value_65.common.dest @[instruction.rs 504:1]
                                        WriteL2Reg(_match_arm_value_66):
                                            connect dest_reg_10, _match_arm_value_66.common.dest @[instruction.rs 504:1]
                                    connect dest_reg_8, dest_reg_10 @[unit.rs 127:1]
                                LoadStore(_match_arm_value_67):
                                    wire dest_reg_11: Ty4 @[instruction.rs 539:1]
                                    match _match_arm_value_67: @[instruction.rs 539:1]
                                        Load(_match_arm_value_68):
                                            connect dest_reg_11, _match_arm_value_68.dest @[instruction.rs 539:1]
                                        Store(_match_arm_value_69):
                                            connect dest_reg_11, _match_arm_value_69.dest @[instruction.rs 539:1]
                                    connect dest_reg_8, dest_reg_11 @[unit.rs 127:1]
                            wire flag_reg_2: Ty1 @[instruction.rs 720:32]
                            wire _bundle_literal_expr_99: Ty1
                            connect _bundle_literal_expr_99.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_2, _bundle_literal_expr_99 @[instruction.rs 721:17]
                            match dest_reg_8.flag_regs[0]: @[instruction.rs 723:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_70):
                                    wire _bundle_literal_expr_100: Ty1
                                    connect _bundle_literal_expr_100.value, tail(UInt<32>(0hFE), 24)
                                    connect flag_reg_2, _bundle_literal_expr_100 @[instruction.rs 725:21]
                            wire flag_reg_3: Ty1 @[instruction.rs 720:32]
                            wire _bundle_literal_expr_101: Ty1
                            connect _bundle_literal_expr_101.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_3, _bundle_literal_expr_101 @[instruction.rs 721:17]
                            match dest_reg_8.flag_regs[1]: @[instruction.rs 723:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_71):
                                    wire _bundle_literal_expr_102: Ty1
                                    connect _bundle_literal_expr_102.value, tail(UInt<32>(0hFF), 24)
                                    connect flag_reg_3, _bundle_literal_expr_102 @[instruction.rs 725:21]
                            when eq(dest_reg_8.normal_regs[0].value, rename_table_normal_1_src_1.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_normal_1_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_59) @[reg_alloc.rs 131:41]
                            when eq(dest_reg_8.normal_regs[1].value, rename_table_normal_1_src_1.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_normal_1_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_59) @[reg_alloc.rs 131:41]
                            when eq(flag_reg_2.value, rename_table_normal_1_src_1.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_normal_1_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_59) @[reg_alloc.rs 131:41]
                            when eq(flag_reg_3.value, rename_table_normal_1_src_1.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_normal_1_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_59) @[reg_alloc.rs 131:41]
        connect rename_table_normal_mem_r5.clk, cd.clk @[reg_alloc.rs 103:17]
        connect rename_table_normal_mem_r5.addr, UInt<8>(0h0) @[reg_alloc.rs 104:17]
        connect rename_table_normal_mem_r5.en, UInt<1>(0h0) @[reg_alloc.rs 105:17]
        wire rename_table_normal_1_src_2: Ty45 @[reg_alloc.rs 108:21]
        wire _bundle_literal_expr_103: Ty1
        connect _bundle_literal_expr_103.value, tail(UInt<32>(0h0), 24)
        connect rename_table_normal_1_src_2.addr, _bundle_literal_expr_103 @[reg_alloc.rs 111:17]
        connect rename_table_normal_1_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlNone) @[reg_alloc.rs 112:17]
        when and(geq(rename_table_normal_1_src_2.addr.value, UInt<32>(0h1)), lt(rename_table_normal_1_src_2.addr.value, UInt<32>(0hFE))): @[reg_alloc.rs 114:17]
            ; connect different types:
            ; lhs: UInt<8>
            ; rhs: UInt<33>
            connect rename_table_normal_mem_r5.addr, sub(rename_table_normal_1_src_2.addr.value, UInt<32>(0h1)) @[reg_alloc.rs 115:21]
            connect rename_table_normal_mem_r5.en, UInt<1>(0h1) @[reg_alloc.rs 116:21]
            connect rename_table_normal_1_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, rename_table_normal_mem_r5.data) @[reg_alloc.rs 117:21]
            match fetch_decode_interface.decoded_insns[0].data: @[reg_alloc.rs 120:25]
                HdlNone:
                    skip
                HdlSome(_match_arm_value_72):
                    match renamed_mops_out_reg[0]: @[reg_alloc.rs 124:29]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_73):
                            wire dest_reg_12: Ty4 @[unit.rs 127:1]
                            match _match_arm_value_72.mop: @[unit.rs 127:1]
                                AluBranch(_match_arm_value_74):
                                    wire dest_reg_13: Ty4 @[instruction.rs 477:1]
                                    match _match_arm_value_74: @[instruction.rs 477:1]
                                        AddSub(_match_arm_value_75):
                                            connect dest_reg_13, _match_arm_value_75.alu_common.common.dest @[instruction.rs 477:1]
                                        AddSubI(_match_arm_value_76):
                                            connect dest_reg_13, _match_arm_value_76.alu_common.common.dest @[instruction.rs 477:1]
                                        Logical(_match_arm_value_77):
                                            connect dest_reg_13, _match_arm_value_77.alu_common.common.dest @[instruction.rs 477:1]
                                    connect dest_reg_12, dest_reg_13 @[unit.rs 127:1]
                                L2RegisterFile(_match_arm_value_78):
                                    wire dest_reg_14: Ty4 @[instruction.rs 504:1]
                                    match _match_arm_value_78: @[instruction.rs 504:1]
                                        ReadL2Reg(_match_arm_value_79):
                                            connect dest_reg_14, _match_arm_value_79.common.dest @[instruction.rs 504:1]
                                        WriteL2Reg(_match_arm_value_80):
                                            connect dest_reg_14, _match_arm_value_80.common.dest @[instruction.rs 504:1]
                                    connect dest_reg_12, dest_reg_14 @[unit.rs 127:1]
                                LoadStore(_match_arm_value_81):
                                    wire dest_reg_15: Ty4 @[instruction.rs 539:1]
                                    match _match_arm_value_81: @[instruction.rs 539:1]
                                        Load(_match_arm_value_82):
                                            connect dest_reg_15, _match_arm_value_82.dest @[instruction.rs 539:1]
                                        Store(_match_arm_value_83):
                                            connect dest_reg_15, _match_arm_value_83.dest @[instruction.rs 539:1]
                                    connect dest_reg_12, dest_reg_15 @[unit.rs 127:1]
                            wire flag_reg_4: Ty1 @[instruction.rs 720:32]
                            wire _bundle_literal_expr_104: Ty1
                            connect _bundle_literal_expr_104.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_4, _bundle_literal_expr_104 @[instruction.rs 721:17]
                            match dest_reg_12.flag_regs[0]: @[instruction.rs 723:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_84):
                                    wire _bundle_literal_expr_105: Ty1
                                    connect _bundle_literal_expr_105.value, tail(UInt<32>(0hFE), 24)
                                    connect flag_reg_4, _bundle_literal_expr_105 @[instruction.rs 725:21]
                            wire flag_reg_5: Ty1 @[instruction.rs 720:32]
                            wire _bundle_literal_expr_106: Ty1
                            connect _bundle_literal_expr_106.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_5, _bundle_literal_expr_106 @[instruction.rs 721:17]
                            match dest_reg_12.flag_regs[1]: @[instruction.rs 723:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_85):
                                    wire _bundle_literal_expr_107: Ty1
                                    connect _bundle_literal_expr_107.value, tail(UInt<32>(0hFF), 24)
                                    connect flag_reg_5, _bundle_literal_expr_107 @[instruction.rs 725:21]
                            when eq(dest_reg_12.normal_regs[0].value, rename_table_normal_1_src_2.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_normal_1_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_73) @[reg_alloc.rs 131:41]
                            when eq(dest_reg_12.normal_regs[1].value, rename_table_normal_1_src_2.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_normal_1_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_73) @[reg_alloc.rs 131:41]
                            when eq(flag_reg_4.value, rename_table_normal_1_src_2.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_normal_1_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_73) @[reg_alloc.rs 131:41]
                            when eq(flag_reg_5.value, rename_table_normal_1_src_2.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_normal_1_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_73) @[reg_alloc.rs 131:41]
        connect rename_table_special_mem_r3.clk, cd.clk @[reg_alloc.rs 103:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<8>
        connect rename_table_special_mem_r3.addr, UInt<8>(0h0) @[reg_alloc.rs 104:17]
        connect rename_table_special_mem_r3.en, UInt<1>(0h0) @[reg_alloc.rs 105:17]
        wire rename_table_special_1_src_0: Ty45 @[reg_alloc.rs 108:21]
        wire _bundle_literal_expr_108: Ty1
        connect _bundle_literal_expr_108.value, tail(UInt<32>(0h0), 24)
        connect rename_table_special_1_src_0.addr, _bundle_literal_expr_108 @[reg_alloc.rs 111:17]
        connect rename_table_special_1_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlNone) @[reg_alloc.rs 112:17]
        when and(geq(rename_table_special_1_src_0.addr.value, UInt<32>(0hFE)), lt(rename_table_special_1_src_0.addr.value, UInt<32>(0h100))): @[reg_alloc.rs 114:17]
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<33>
            connect rename_table_special_mem_r3.addr, sub(rename_table_special_1_src_0.addr.value, UInt<32>(0hFE)) @[reg_alloc.rs 115:21]
            connect rename_table_special_mem_r3.en, UInt<1>(0h1) @[reg_alloc.rs 116:21]
            connect rename_table_special_1_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, rename_table_special_mem_r3.data) @[reg_alloc.rs 117:21]
            match fetch_decode_interface.decoded_insns[0].data: @[reg_alloc.rs 120:25]
                HdlNone:
                    skip
                HdlSome(_match_arm_value_86):
                    match renamed_mops_out_reg[0]: @[reg_alloc.rs 124:29]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_87):
                            wire dest_reg_16: Ty4 @[unit.rs 127:1]
                            match _match_arm_value_86.mop: @[unit.rs 127:1]
                                AluBranch(_match_arm_value_88):
                                    wire dest_reg_17: Ty4 @[instruction.rs 477:1]
                                    match _match_arm_value_88: @[instruction.rs 477:1]
                                        AddSub(_match_arm_value_89):
                                            connect dest_reg_17, _match_arm_value_89.alu_common.common.dest @[instruction.rs 477:1]
                                        AddSubI(_match_arm_value_90):
                                            connect dest_reg_17, _match_arm_value_90.alu_common.common.dest @[instruction.rs 477:1]
                                        Logical(_match_arm_value_91):
                                            connect dest_reg_17, _match_arm_value_91.alu_common.common.dest @[instruction.rs 477:1]
                                    connect dest_reg_16, dest_reg_17 @[unit.rs 127:1]
                                L2RegisterFile(_match_arm_value_92):
                                    wire dest_reg_18: Ty4 @[instruction.rs 504:1]
                                    match _match_arm_value_92: @[instruction.rs 504:1]
                                        ReadL2Reg(_match_arm_value_93):
                                            connect dest_reg_18, _match_arm_value_93.common.dest @[instruction.rs 504:1]
                                        WriteL2Reg(_match_arm_value_94):
                                            connect dest_reg_18, _match_arm_value_94.common.dest @[instruction.rs 504:1]
                                    connect dest_reg_16, dest_reg_18 @[unit.rs 127:1]
                                LoadStore(_match_arm_value_95):
                                    wire dest_reg_19: Ty4 @[instruction.rs 539:1]
                                    match _match_arm_value_95: @[instruction.rs 539:1]
                                        Load(_match_arm_value_96):
                                            connect dest_reg_19, _match_arm_value_96.dest @[instruction.rs 539:1]
                                        Store(_match_arm_value_97):
                                            connect dest_reg_19, _match_arm_value_97.dest @[instruction.rs 539:1]
                                    connect dest_reg_16, dest_reg_19 @[unit.rs 127:1]
                            wire flag_reg_6: Ty1 @[instruction.rs 720:32]
                            wire _bundle_literal_expr_109: Ty1
                            connect _bundle_literal_expr_109.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_6, _bundle_literal_expr_109 @[instruction.rs 721:17]
                            match dest_reg_16.flag_regs[0]: @[instruction.rs 723:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_98):
                                    wire _bundle_literal_expr_110: Ty1
                                    connect _bundle_literal_expr_110.value, tail(UInt<32>(0hFE), 24)
                                    connect flag_reg_6, _bundle_literal_expr_110 @[instruction.rs 725:21]
                            wire flag_reg_7: Ty1 @[instruction.rs 720:32]
                            wire _bundle_literal_expr_111: Ty1
                            connect _bundle_literal_expr_111.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_7, _bundle_literal_expr_111 @[instruction.rs 721:17]
                            match dest_reg_16.flag_regs[1]: @[instruction.rs 723:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_99):
                                    wire _bundle_literal_expr_112: Ty1
                                    connect _bundle_literal_expr_112.value, tail(UInt<32>(0hFF), 24)
                                    connect flag_reg_7, _bundle_literal_expr_112 @[instruction.rs 725:21]
                            when eq(dest_reg_16.normal_regs[0].value, rename_table_special_1_src_0.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_special_1_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_87) @[reg_alloc.rs 131:41]
                            when eq(dest_reg_16.normal_regs[1].value, rename_table_special_1_src_0.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_special_1_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_87) @[reg_alloc.rs 131:41]
                            when eq(flag_reg_6.value, rename_table_special_1_src_0.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_special_1_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_87) @[reg_alloc.rs 131:41]
                            when eq(flag_reg_7.value, rename_table_special_1_src_0.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_special_1_src_0.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_87) @[reg_alloc.rs 131:41]
        connect rename_table_special_mem_r4.clk, cd.clk @[reg_alloc.rs 103:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<8>
        connect rename_table_special_mem_r4.addr, UInt<8>(0h0) @[reg_alloc.rs 104:17]
        connect rename_table_special_mem_r4.en, UInt<1>(0h0) @[reg_alloc.rs 105:17]
        wire rename_table_special_1_src_1: Ty45 @[reg_alloc.rs 108:21]
        wire _bundle_literal_expr_113: Ty1
        connect _bundle_literal_expr_113.value, tail(UInt<32>(0h0), 24)
        connect rename_table_special_1_src_1.addr, _bundle_literal_expr_113 @[reg_alloc.rs 111:17]
        connect rename_table_special_1_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlNone) @[reg_alloc.rs 112:17]
        when and(geq(rename_table_special_1_src_1.addr.value, UInt<32>(0hFE)), lt(rename_table_special_1_src_1.addr.value, UInt<32>(0h100))): @[reg_alloc.rs 114:17]
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<33>
            connect rename_table_special_mem_r4.addr, sub(rename_table_special_1_src_1.addr.value, UInt<32>(0hFE)) @[reg_alloc.rs 115:21]
            connect rename_table_special_mem_r4.en, UInt<1>(0h1) @[reg_alloc.rs 116:21]
            connect rename_table_special_1_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, rename_table_special_mem_r4.data) @[reg_alloc.rs 117:21]
            match fetch_decode_interface.decoded_insns[0].data: @[reg_alloc.rs 120:25]
                HdlNone:
                    skip
                HdlSome(_match_arm_value_100):
                    match renamed_mops_out_reg[0]: @[reg_alloc.rs 124:29]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_101):
                            wire dest_reg_20: Ty4 @[unit.rs 127:1]
                            match _match_arm_value_100.mop: @[unit.rs 127:1]
                                AluBranch(_match_arm_value_102):
                                    wire dest_reg_21: Ty4 @[instruction.rs 477:1]
                                    match _match_arm_value_102: @[instruction.rs 477:1]
                                        AddSub(_match_arm_value_103):
                                            connect dest_reg_21, _match_arm_value_103.alu_common.common.dest @[instruction.rs 477:1]
                                        AddSubI(_match_arm_value_104):
                                            connect dest_reg_21, _match_arm_value_104.alu_common.common.dest @[instruction.rs 477:1]
                                        Logical(_match_arm_value_105):
                                            connect dest_reg_21, _match_arm_value_105.alu_common.common.dest @[instruction.rs 477:1]
                                    connect dest_reg_20, dest_reg_21 @[unit.rs 127:1]
                                L2RegisterFile(_match_arm_value_106):
                                    wire dest_reg_22: Ty4 @[instruction.rs 504:1]
                                    match _match_arm_value_106: @[instruction.rs 504:1]
                                        ReadL2Reg(_match_arm_value_107):
                                            connect dest_reg_22, _match_arm_value_107.common.dest @[instruction.rs 504:1]
                                        WriteL2Reg(_match_arm_value_108):
                                            connect dest_reg_22, _match_arm_value_108.common.dest @[instruction.rs 504:1]
                                    connect dest_reg_20, dest_reg_22 @[unit.rs 127:1]
                                LoadStore(_match_arm_value_109):
                                    wire dest_reg_23: Ty4 @[instruction.rs 539:1]
                                    match _match_arm_value_109: @[instruction.rs 539:1]
                                        Load(_match_arm_value_110):
                                            connect dest_reg_23, _match_arm_value_110.dest @[instruction.rs 539:1]
                                        Store(_match_arm_value_111):
                                            connect dest_reg_23, _match_arm_value_111.dest @[instruction.rs 539:1]
                                    connect dest_reg_20, dest_reg_23 @[unit.rs 127:1]
                            wire flag_reg_8: Ty1 @[instruction.rs 720:32]
                            wire _bundle_literal_expr_114: Ty1
                            connect _bundle_literal_expr_114.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_8, _bundle_literal_expr_114 @[instruction.rs 721:17]
                            match dest_reg_20.flag_regs[0]: @[instruction.rs 723:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_112):
                                    wire _bundle_literal_expr_115: Ty1
                                    connect _bundle_literal_expr_115.value, tail(UInt<32>(0hFE), 24)
                                    connect flag_reg_8, _bundle_literal_expr_115 @[instruction.rs 725:21]
                            wire flag_reg_9: Ty1 @[instruction.rs 720:32]
                            wire _bundle_literal_expr_116: Ty1
                            connect _bundle_literal_expr_116.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_9, _bundle_literal_expr_116 @[instruction.rs 721:17]
                            match dest_reg_20.flag_regs[1]: @[instruction.rs 723:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_113):
                                    wire _bundle_literal_expr_117: Ty1
                                    connect _bundle_literal_expr_117.value, tail(UInt<32>(0hFF), 24)
                                    connect flag_reg_9, _bundle_literal_expr_117 @[instruction.rs 725:21]
                            when eq(dest_reg_20.normal_regs[0].value, rename_table_special_1_src_1.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_special_1_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_101) @[reg_alloc.rs 131:41]
                            when eq(dest_reg_20.normal_regs[1].value, rename_table_special_1_src_1.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_special_1_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_101) @[reg_alloc.rs 131:41]
                            when eq(flag_reg_8.value, rename_table_special_1_src_1.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_special_1_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_101) @[reg_alloc.rs 131:41]
                            when eq(flag_reg_9.value, rename_table_special_1_src_1.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_special_1_src_1.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_101) @[reg_alloc.rs 131:41]
        connect rename_table_special_mem_r5.clk, cd.clk @[reg_alloc.rs 103:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<8>
        connect rename_table_special_mem_r5.addr, UInt<8>(0h0) @[reg_alloc.rs 104:17]
        connect rename_table_special_mem_r5.en, UInt<1>(0h0) @[reg_alloc.rs 105:17]
        wire rename_table_special_1_src_2: Ty45 @[reg_alloc.rs 108:21]
        wire _bundle_literal_expr_118: Ty1
        connect _bundle_literal_expr_118.value, tail(UInt<32>(0h0), 24)
        connect rename_table_special_1_src_2.addr, _bundle_literal_expr_118 @[reg_alloc.rs 111:17]
        connect rename_table_special_1_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlNone) @[reg_alloc.rs 112:17]
        when and(geq(rename_table_special_1_src_2.addr.value, UInt<32>(0hFE)), lt(rename_table_special_1_src_2.addr.value, UInt<32>(0h100))): @[reg_alloc.rs 114:17]
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<33>
            connect rename_table_special_mem_r5.addr, sub(rename_table_special_1_src_2.addr.value, UInt<32>(0hFE)) @[reg_alloc.rs 115:21]
            connect rename_table_special_mem_r5.en, UInt<1>(0h1) @[reg_alloc.rs 116:21]
            connect rename_table_special_1_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, rename_table_special_mem_r5.data) @[reg_alloc.rs 117:21]
            match fetch_decode_interface.decoded_insns[0].data: @[reg_alloc.rs 120:25]
                HdlNone:
                    skip
                HdlSome(_match_arm_value_114):
                    match renamed_mops_out_reg[0]: @[reg_alloc.rs 124:29]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_115):
                            wire dest_reg_24: Ty4 @[unit.rs 127:1]
                            match _match_arm_value_114.mop: @[unit.rs 127:1]
                                AluBranch(_match_arm_value_116):
                                    wire dest_reg_25: Ty4 @[instruction.rs 477:1]
                                    match _match_arm_value_116: @[instruction.rs 477:1]
                                        AddSub(_match_arm_value_117):
                                            connect dest_reg_25, _match_arm_value_117.alu_common.common.dest @[instruction.rs 477:1]
                                        AddSubI(_match_arm_value_118):
                                            connect dest_reg_25, _match_arm_value_118.alu_common.common.dest @[instruction.rs 477:1]
                                        Logical(_match_arm_value_119):
                                            connect dest_reg_25, _match_arm_value_119.alu_common.common.dest @[instruction.rs 477:1]
                                    connect dest_reg_24, dest_reg_25 @[unit.rs 127:1]
                                L2RegisterFile(_match_arm_value_120):
                                    wire dest_reg_26: Ty4 @[instruction.rs 504:1]
                                    match _match_arm_value_120: @[instruction.rs 504:1]
                                        ReadL2Reg(_match_arm_value_121):
                                            connect dest_reg_26, _match_arm_value_121.common.dest @[instruction.rs 504:1]
                                        WriteL2Reg(_match_arm_value_122):
                                            connect dest_reg_26, _match_arm_value_122.common.dest @[instruction.rs 504:1]
                                    connect dest_reg_24, dest_reg_26 @[unit.rs 127:1]
                                LoadStore(_match_arm_value_123):
                                    wire dest_reg_27: Ty4 @[instruction.rs 539:1]
                                    match _match_arm_value_123: @[instruction.rs 539:1]
                                        Load(_match_arm_value_124):
                                            connect dest_reg_27, _match_arm_value_124.dest @[instruction.rs 539:1]
                                        Store(_match_arm_value_125):
                                            connect dest_reg_27, _match_arm_value_125.dest @[instruction.rs 539:1]
                                    connect dest_reg_24, dest_reg_27 @[unit.rs 127:1]
                            wire flag_reg_10: Ty1 @[instruction.rs 720:32]
                            wire _bundle_literal_expr_119: Ty1
                            connect _bundle_literal_expr_119.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_10, _bundle_literal_expr_119 @[instruction.rs 721:17]
                            match dest_reg_24.flag_regs[0]: @[instruction.rs 723:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_126):
                                    wire _bundle_literal_expr_120: Ty1
                                    connect _bundle_literal_expr_120.value, tail(UInt<32>(0hFE), 24)
                                    connect flag_reg_10, _bundle_literal_expr_120 @[instruction.rs 725:21]
                            wire flag_reg_11: Ty1 @[instruction.rs 720:32]
                            wire _bundle_literal_expr_121: Ty1
                            connect _bundle_literal_expr_121.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_11, _bundle_literal_expr_121 @[instruction.rs 721:17]
                            match dest_reg_24.flag_regs[1]: @[instruction.rs 723:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_127):
                                    wire _bundle_literal_expr_122: Ty1
                                    connect _bundle_literal_expr_122.value, tail(UInt<32>(0hFF), 24)
                                    connect flag_reg_11, _bundle_literal_expr_122 @[instruction.rs 725:21]
                            when eq(dest_reg_24.normal_regs[0].value, rename_table_special_1_src_2.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_special_1_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_115) @[reg_alloc.rs 131:41]
                            when eq(dest_reg_24.normal_regs[1].value, rename_table_special_1_src_2.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_special_1_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_115) @[reg_alloc.rs 131:41]
                            when eq(flag_reg_10.value, rename_table_special_1_src_2.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_special_1_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_115) @[reg_alloc.rs 131:41]
                            when eq(flag_reg_11.value, rename_table_special_1_src_2.addr.value): @[reg_alloc.rs 130:37]
                                connect rename_table_special_1_src_2.data, {|HdlNone, HdlSome: Ty25|}(HdlSome, _match_arm_value_115) @[reg_alloc.rs 131:41]
        match fetch_decode_interface.decoded_insns[1].data: @[reg_alloc.rs 159:9]
            HdlNone:
                skip
            HdlSome(_match_arm_value_128):
                wire unit_kind_1: Ty46 @[unit.rs 127:1]
                match _match_arm_value_128.mop: @[unit.rs 127:1]
                    AluBranch(_match_arm_value_129):
                        connect unit_kind_1, {|AluBranch, L2RegisterFile, LoadStore|}(AluBranch) @[unit.rs 127:1]
                    L2RegisterFile(_match_arm_value_130):
                        connect unit_kind_1, {|AluBranch, L2RegisterFile, LoadStore|}(L2RegisterFile) @[unit.rs 127:1]
                    LoadStore(_match_arm_value_131):
                        connect unit_kind_1, {|AluBranch, L2RegisterFile, LoadStore|}(LoadStore) @[unit.rs 127:1]
                wire available_units_for_kind_1: UInt<1>[2] @[unit.rs 127:1]
                match unit_kind_1: @[unit.rs 127:1]
                    AluBranch:
                        connect available_units_for_kind_1[0], UInt<1>(0h1) @[unit.rs 127:1]
                        connect available_units_for_kind_1[1], UInt<1>(0h1) @[unit.rs 127:1]
                    L2RegisterFile:
                        connect available_units_for_kind_1[0], UInt<1>(0h0) @[unit.rs 127:1]
                        connect available_units_for_kind_1[1], UInt<1>(0h0) @[unit.rs 127:1]
                    LoadStore:
                        connect available_units_for_kind_1[0], UInt<1>(0h0) @[unit.rs 127:1]
                        connect available_units_for_kind_1[1], UInt<1>(0h0) @[unit.rs 127:1]
                connect available_units[1], available_units_for_kind_1 @[reg_alloc.rs 160:13]
                match renamed_mops_out_reg[1]: @[reg_alloc.rs 165:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_132):
                        wire dest_reg_28: Ty4 @[unit.rs 127:1]
                        match _match_arm_value_128.mop: @[unit.rs 127:1]
                            AluBranch(_match_arm_value_133):
                                wire dest_reg_29: Ty4 @[instruction.rs 477:1]
                                match _match_arm_value_133: @[instruction.rs 477:1]
                                    AddSub(_match_arm_value_134):
                                        connect dest_reg_29, _match_arm_value_134.alu_common.common.dest @[instruction.rs 477:1]
                                    AddSubI(_match_arm_value_135):
                                        connect dest_reg_29, _match_arm_value_135.alu_common.common.dest @[instruction.rs 477:1]
                                    Logical(_match_arm_value_136):
                                        connect dest_reg_29, _match_arm_value_136.alu_common.common.dest @[instruction.rs 477:1]
                                connect dest_reg_28, dest_reg_29 @[unit.rs 127:1]
                            L2RegisterFile(_match_arm_value_137):
                                wire dest_reg_30: Ty4 @[instruction.rs 504:1]
                                match _match_arm_value_137: @[instruction.rs 504:1]
                                    ReadL2Reg(_match_arm_value_138):
                                        connect dest_reg_30, _match_arm_value_138.common.dest @[instruction.rs 504:1]
                                    WriteL2Reg(_match_arm_value_139):
                                        connect dest_reg_30, _match_arm_value_139.common.dest @[instruction.rs 504:1]
                                connect dest_reg_28, dest_reg_30 @[unit.rs 127:1]
                            LoadStore(_match_arm_value_140):
                                wire dest_reg_31: Ty4 @[instruction.rs 539:1]
                                match _match_arm_value_140: @[instruction.rs 539:1]
                                    Load(_match_arm_value_141):
                                        connect dest_reg_31, _match_arm_value_141.dest @[instruction.rs 539:1]
                                    Store(_match_arm_value_142):
                                        connect dest_reg_31, _match_arm_value_142.dest @[instruction.rs 539:1]
                                connect dest_reg_28, dest_reg_31 @[unit.rs 127:1]
                        wire mapped_regs_4: Ty42 @[unit.rs 127:1]
                        match _match_arm_value_128.mop: @[unit.rs 127:1]
                            AluBranch(_match_arm_value_143):
                                wire mapped_regs_5: Ty37 @[instruction.rs 477:1]
                                match _match_arm_value_143: @[instruction.rs 477:1]
                                    AddSub(_match_arm_value_144):
                                        wire renamed_src_reg_1_0: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_123: Ty1
                                        connect _bundle_literal_expr_123.value, _match_arm_value_144.alu_common.common.src[0]
                                        connect rename_table_normal_1_src_0.addr, _bundle_literal_expr_123 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_124: Ty1
                                        connect _bundle_literal_expr_124.value, _match_arm_value_144.alu_common.common.src[0]
                                        connect rename_table_special_1_src_0.addr, _bundle_literal_expr_124 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_1_src_0.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_1_src_0.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_125: Ty25
                                                        wire _bundle_literal_expr_126: Ty23
                                                        connect _bundle_literal_expr_126.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_125.unit_num, _bundle_literal_expr_126
                                                        wire _bundle_literal_expr_127: Ty24
                                                        connect _bundle_literal_expr_127.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_125.unit_out_reg, _bundle_literal_expr_127
                                                        connect renamed_src_reg_1_0, _bundle_literal_expr_125 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_145):
                                                        connect renamed_src_reg_1_0, _match_arm_value_145 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_146):
                                                connect renamed_src_reg_1_0, _match_arm_value_146 @[reg_alloc.rs 185:33]
                                        wire renamed_src_reg_1_1: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_128: Ty1
                                        connect _bundle_literal_expr_128.value, _match_arm_value_144.alu_common.common.src[1]
                                        connect rename_table_normal_1_src_1.addr, _bundle_literal_expr_128 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_129: Ty1
                                        connect _bundle_literal_expr_129.value, _match_arm_value_144.alu_common.common.src[1]
                                        connect rename_table_special_1_src_1.addr, _bundle_literal_expr_129 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_1_src_1.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_1_src_1.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_130: Ty25
                                                        wire _bundle_literal_expr_131: Ty23
                                                        connect _bundle_literal_expr_131.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_130.unit_num, _bundle_literal_expr_131
                                                        wire _bundle_literal_expr_132: Ty24
                                                        connect _bundle_literal_expr_132.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_130.unit_out_reg, _bundle_literal_expr_132
                                                        connect renamed_src_reg_1_1, _bundle_literal_expr_130 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_147):
                                                        connect renamed_src_reg_1_1, _match_arm_value_147 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_148):
                                                connect renamed_src_reg_1_1, _match_arm_value_148 @[reg_alloc.rs 185:33]
                                        wire renamed_src_reg_1_2: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_133: Ty1
                                        connect _bundle_literal_expr_133.value, _match_arm_value_144.alu_common.common.src[2]
                                        connect rename_table_normal_1_src_2.addr, _bundle_literal_expr_133 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_134: Ty1
                                        connect _bundle_literal_expr_134.value, _match_arm_value_144.alu_common.common.src[2]
                                        connect rename_table_special_1_src_2.addr, _bundle_literal_expr_134 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_1_src_2.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_1_src_2.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_135: Ty25
                                                        wire _bundle_literal_expr_136: Ty23
                                                        connect _bundle_literal_expr_136.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_135.unit_num, _bundle_literal_expr_136
                                                        wire _bundle_literal_expr_137: Ty24
                                                        connect _bundle_literal_expr_137.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_135.unit_out_reg, _bundle_literal_expr_137
                                                        connect renamed_src_reg_1_2, _bundle_literal_expr_135 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_149):
                                                        connect renamed_src_reg_1_2, _match_arm_value_149 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_150):
                                                connect renamed_src_reg_1_2, _match_arm_value_150 @[reg_alloc.rs 185:33]
                                        wire _bundle_literal_expr_138: Ty35
                                        wire _bundle_literal_expr_139: Ty34
                                        wire _bundle_literal_expr_140: Ty33
                                        connect _bundle_literal_expr_140.prefix_pad, _match_arm_value_144.alu_common.common.prefix_pad
                                        connect _bundle_literal_expr_140.dest, _match_arm_value_132.unit_out_reg
                                        wire _array_literal_expr_38: UInt<6>[3]
                                        wire _array_literal_expr_39: UInt<6>[3]
                                        wire _cast_bundle_to_bits_expr_39: Ty47
                                        connect _cast_bundle_to_bits_expr_39.unit_num, renamed_src_reg_1_0.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_39.unit_out_reg, renamed_src_reg_1_0.unit_out_reg.value
                                        wire _cast_to_bits_expr_39: UInt<6>
                                        connect _cast_to_bits_expr_39, cat(_cast_bundle_to_bits_expr_39.unit_out_reg, _cast_bundle_to_bits_expr_39.unit_num)
                                        connect _array_literal_expr_39[0], _cast_to_bits_expr_39
                                        wire _cast_bundle_to_bits_expr_40: Ty47
                                        connect _cast_bundle_to_bits_expr_40.unit_num, renamed_src_reg_1_1.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_40.unit_out_reg, renamed_src_reg_1_1.unit_out_reg.value
                                        wire _cast_to_bits_expr_40: UInt<6>
                                        connect _cast_to_bits_expr_40, cat(_cast_bundle_to_bits_expr_40.unit_out_reg, _cast_bundle_to_bits_expr_40.unit_num)
                                        connect _array_literal_expr_39[1], _cast_to_bits_expr_40
                                        wire _cast_bundle_to_bits_expr_41: Ty47
                                        connect _cast_bundle_to_bits_expr_41.unit_num, renamed_src_reg_1_2.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_41.unit_out_reg, renamed_src_reg_1_2.unit_out_reg.value
                                        wire _cast_to_bits_expr_41: UInt<6>
                                        connect _cast_to_bits_expr_41, cat(_cast_bundle_to_bits_expr_41.unit_out_reg, _cast_bundle_to_bits_expr_41.unit_num)
                                        connect _array_literal_expr_39[2], _cast_to_bits_expr_41
                                        connect _array_literal_expr_38[0], _array_literal_expr_39[0]
                                        wire _array_literal_expr_40: UInt<6>[3]
                                        wire _cast_bundle_to_bits_expr_42: Ty47
                                        connect _cast_bundle_to_bits_expr_42.unit_num, renamed_src_reg_1_0.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_42.unit_out_reg, renamed_src_reg_1_0.unit_out_reg.value
                                        wire _cast_to_bits_expr_42: UInt<6>
                                        connect _cast_to_bits_expr_42, cat(_cast_bundle_to_bits_expr_42.unit_out_reg, _cast_bundle_to_bits_expr_42.unit_num)
                                        connect _array_literal_expr_40[0], _cast_to_bits_expr_42
                                        wire _cast_bundle_to_bits_expr_43: Ty47
                                        connect _cast_bundle_to_bits_expr_43.unit_num, renamed_src_reg_1_1.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_43.unit_out_reg, renamed_src_reg_1_1.unit_out_reg.value
                                        wire _cast_to_bits_expr_43: UInt<6>
                                        connect _cast_to_bits_expr_43, cat(_cast_bundle_to_bits_expr_43.unit_out_reg, _cast_bundle_to_bits_expr_43.unit_num)
                                        connect _array_literal_expr_40[1], _cast_to_bits_expr_43
                                        wire _cast_bundle_to_bits_expr_44: Ty47
                                        connect _cast_bundle_to_bits_expr_44.unit_num, renamed_src_reg_1_2.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_44.unit_out_reg, renamed_src_reg_1_2.unit_out_reg.value
                                        wire _cast_to_bits_expr_44: UInt<6>
                                        connect _cast_to_bits_expr_44, cat(_cast_bundle_to_bits_expr_44.unit_out_reg, _cast_bundle_to_bits_expr_44.unit_num)
                                        connect _array_literal_expr_40[2], _cast_to_bits_expr_44
                                        connect _array_literal_expr_38[1], _array_literal_expr_40[1]
                                        wire _array_literal_expr_41: UInt<6>[3]
                                        wire _cast_bundle_to_bits_expr_45: Ty47
                                        connect _cast_bundle_to_bits_expr_45.unit_num, renamed_src_reg_1_0.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_45.unit_out_reg, renamed_src_reg_1_0.unit_out_reg.value
                                        wire _cast_to_bits_expr_45: UInt<6>
                                        connect _cast_to_bits_expr_45, cat(_cast_bundle_to_bits_expr_45.unit_out_reg, _cast_bundle_to_bits_expr_45.unit_num)
                                        connect _array_literal_expr_41[0], _cast_to_bits_expr_45
                                        wire _cast_bundle_to_bits_expr_46: Ty47
                                        connect _cast_bundle_to_bits_expr_46.unit_num, renamed_src_reg_1_1.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_46.unit_out_reg, renamed_src_reg_1_1.unit_out_reg.value
                                        wire _cast_to_bits_expr_46: UInt<6>
                                        connect _cast_to_bits_expr_46, cat(_cast_bundle_to_bits_expr_46.unit_out_reg, _cast_bundle_to_bits_expr_46.unit_num)
                                        connect _array_literal_expr_41[1], _cast_to_bits_expr_46
                                        wire _cast_bundle_to_bits_expr_47: Ty47
                                        connect _cast_bundle_to_bits_expr_47.unit_num, renamed_src_reg_1_2.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_47.unit_out_reg, renamed_src_reg_1_2.unit_out_reg.value
                                        wire _cast_to_bits_expr_47: UInt<6>
                                        connect _cast_to_bits_expr_47, cat(_cast_bundle_to_bits_expr_47.unit_out_reg, _cast_bundle_to_bits_expr_47.unit_num)
                                        connect _array_literal_expr_41[2], _cast_to_bits_expr_47
                                        connect _array_literal_expr_38[2], _array_literal_expr_41[2]
                                        connect _bundle_literal_expr_140.src, _array_literal_expr_38
                                        wire _bundle_literal_expr_141: Ty48
                                        connect _bundle_literal_expr_141.imm_low, _match_arm_value_144.alu_common.common.imm_low
                                        wire _array_literal_expr_42: UInt<8>[0]
                                        invalidate _array_literal_expr_42
                                        connect _bundle_literal_expr_141.reversed_src, _array_literal_expr_42
                                        connect _bundle_literal_expr_141.imm_sign, _match_arm_value_144.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_48: Ty49
                                        connect _cast_bundle_to_bits_expr_48.imm_low, _bundle_literal_expr_141.imm_low
                                        connect _cast_bundle_to_bits_expr_48.reversed_src, UInt<0>(0)
                                        connect _cast_bundle_to_bits_expr_48.imm_sign, asUInt(_bundle_literal_expr_141.imm_sign)
                                        wire _cast_to_bits_expr_48: UInt<26>
                                        connect _cast_to_bits_expr_48, cat(_cast_bundle_to_bits_expr_48.imm_sign, cat(_cast_bundle_to_bits_expr_48.reversed_src, _cast_bundle_to_bits_expr_48.imm_low))
                                        connect _bundle_literal_expr_140.imm_low, bits(asSInt(_cast_to_bits_expr_48), 24, 0)
                                        wire _bundle_literal_expr_142: Ty48
                                        connect _bundle_literal_expr_142.imm_low, _match_arm_value_144.alu_common.common.imm_low
                                        wire _array_literal_expr_43: UInt<8>[0]
                                        invalidate _array_literal_expr_43
                                        connect _bundle_literal_expr_142.reversed_src, _array_literal_expr_43
                                        connect _bundle_literal_expr_142.imm_sign, _match_arm_value_144.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_49: Ty49
                                        connect _cast_bundle_to_bits_expr_49.imm_low, _bundle_literal_expr_142.imm_low
                                        connect _cast_bundle_to_bits_expr_49.reversed_src, UInt<0>(0)
                                        connect _cast_bundle_to_bits_expr_49.imm_sign, asUInt(_bundle_literal_expr_142.imm_sign)
                                        wire _cast_to_bits_expr_49: UInt<26>
                                        connect _cast_to_bits_expr_49, cat(_cast_bundle_to_bits_expr_49.imm_sign, cat(_cast_bundle_to_bits_expr_49.reversed_src, _cast_bundle_to_bits_expr_49.imm_low))
                                        connect _bundle_literal_expr_140.imm_sign, shr(asSInt(_cast_to_bits_expr_49), 25)
                                        wire _bundle_literal_expr_143: Ty2
                                        invalidate _bundle_literal_expr_143
                                        connect _bundle_literal_expr_140._phantom, _bundle_literal_expr_143
                                        connect _bundle_literal_expr_139.common, _bundle_literal_expr_140
                                        connect _bundle_literal_expr_139.output_integer_mode, _match_arm_value_144.alu_common.output_integer_mode
                                        connect _bundle_literal_expr_138.alu_common, _bundle_literal_expr_139
                                        connect _bundle_literal_expr_138.invert_src0, _match_arm_value_144.invert_src0
                                        connect _bundle_literal_expr_138.invert_carry_in, _match_arm_value_144.invert_carry_in
                                        connect _bundle_literal_expr_138.invert_carry_out, _match_arm_value_144.invert_carry_out
                                        connect _bundle_literal_expr_138.add_pc, _match_arm_value_144.add_pc
                                        connect mapped_regs_5, {|AddSub: Ty35, AddSubI: Ty35, Logical: Ty36|}(AddSub, _bundle_literal_expr_138) @[instruction.rs 477:1]
                                    AddSubI(_match_arm_value_151):
                                        wire renamed_src_reg_1_0_1: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_144: Ty1
                                        connect _bundle_literal_expr_144.value, _match_arm_value_151.alu_common.common.src[0]
                                        connect rename_table_normal_1_src_0.addr, _bundle_literal_expr_144 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_145: Ty1
                                        connect _bundle_literal_expr_145.value, _match_arm_value_151.alu_common.common.src[0]
                                        connect rename_table_special_1_src_0.addr, _bundle_literal_expr_145 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_1_src_0.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_1_src_0.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_146: Ty25
                                                        wire _bundle_literal_expr_147: Ty23
                                                        connect _bundle_literal_expr_147.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_146.unit_num, _bundle_literal_expr_147
                                                        wire _bundle_literal_expr_148: Ty24
                                                        connect _bundle_literal_expr_148.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_146.unit_out_reg, _bundle_literal_expr_148
                                                        connect renamed_src_reg_1_0_1, _bundle_literal_expr_146 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_152):
                                                        connect renamed_src_reg_1_0_1, _match_arm_value_152 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_153):
                                                connect renamed_src_reg_1_0_1, _match_arm_value_153 @[reg_alloc.rs 185:33]
                                        wire renamed_src_reg_1_1_1: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_149: Ty1
                                        connect _bundle_literal_expr_149.value, _match_arm_value_151.alu_common.common.src[1]
                                        connect rename_table_normal_1_src_1.addr, _bundle_literal_expr_149 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_150: Ty1
                                        connect _bundle_literal_expr_150.value, _match_arm_value_151.alu_common.common.src[1]
                                        connect rename_table_special_1_src_1.addr, _bundle_literal_expr_150 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_1_src_1.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_1_src_1.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_151: Ty25
                                                        wire _bundle_literal_expr_152: Ty23
                                                        connect _bundle_literal_expr_152.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_151.unit_num, _bundle_literal_expr_152
                                                        wire _bundle_literal_expr_153: Ty24
                                                        connect _bundle_literal_expr_153.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_151.unit_out_reg, _bundle_literal_expr_153
                                                        connect renamed_src_reg_1_1_1, _bundle_literal_expr_151 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_154):
                                                        connect renamed_src_reg_1_1_1, _match_arm_value_154 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_155):
                                                connect renamed_src_reg_1_1_1, _match_arm_value_155 @[reg_alloc.rs 185:33]
                                        wire _bundle_literal_expr_154: Ty35
                                        wire _bundle_literal_expr_155: Ty34
                                        wire _bundle_literal_expr_156: Ty33
                                        connect _bundle_literal_expr_156.prefix_pad, _match_arm_value_151.alu_common.common.prefix_pad
                                        connect _bundle_literal_expr_156.dest, _match_arm_value_132.unit_out_reg
                                        wire _array_literal_expr_44: UInt<6>[3]
                                        wire _array_literal_expr_45: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_50: Ty47
                                        connect _cast_bundle_to_bits_expr_50.unit_num, renamed_src_reg_1_0_1.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_50.unit_out_reg, renamed_src_reg_1_0_1.unit_out_reg.value
                                        wire _cast_to_bits_expr_50: UInt<6>
                                        connect _cast_to_bits_expr_50, cat(_cast_bundle_to_bits_expr_50.unit_out_reg, _cast_bundle_to_bits_expr_50.unit_num)
                                        connect _array_literal_expr_45[0], _cast_to_bits_expr_50
                                        wire _cast_bundle_to_bits_expr_51: Ty47
                                        connect _cast_bundle_to_bits_expr_51.unit_num, renamed_src_reg_1_1_1.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_51.unit_out_reg, renamed_src_reg_1_1_1.unit_out_reg.value
                                        wire _cast_to_bits_expr_51: UInt<6>
                                        connect _cast_to_bits_expr_51, cat(_cast_bundle_to_bits_expr_51.unit_out_reg, _cast_bundle_to_bits_expr_51.unit_num)
                                        connect _array_literal_expr_45[1], _cast_to_bits_expr_51
                                        connect _array_literal_expr_44[0], _array_literal_expr_45[0]
                                        wire _array_literal_expr_46: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_52: Ty47
                                        connect _cast_bundle_to_bits_expr_52.unit_num, renamed_src_reg_1_0_1.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_52.unit_out_reg, renamed_src_reg_1_0_1.unit_out_reg.value
                                        wire _cast_to_bits_expr_52: UInt<6>
                                        connect _cast_to_bits_expr_52, cat(_cast_bundle_to_bits_expr_52.unit_out_reg, _cast_bundle_to_bits_expr_52.unit_num)
                                        connect _array_literal_expr_46[0], _cast_to_bits_expr_52
                                        wire _cast_bundle_to_bits_expr_53: Ty47
                                        connect _cast_bundle_to_bits_expr_53.unit_num, renamed_src_reg_1_1_1.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_53.unit_out_reg, renamed_src_reg_1_1_1.unit_out_reg.value
                                        wire _cast_to_bits_expr_53: UInt<6>
                                        connect _cast_to_bits_expr_53, cat(_cast_bundle_to_bits_expr_53.unit_out_reg, _cast_bundle_to_bits_expr_53.unit_num)
                                        connect _array_literal_expr_46[1], _cast_to_bits_expr_53
                                        connect _array_literal_expr_44[1], _array_literal_expr_46[1]
                                        wire _bundle_literal_expr_157: Ty50
                                        connect _bundle_literal_expr_157.imm_low, _match_arm_value_151.alu_common.common.imm_low
                                        wire _array_literal_expr_47: UInt<8>[1]
                                        connect _array_literal_expr_47[0], _match_arm_value_151.alu_common.common.src[2]
                                        connect _bundle_literal_expr_157.reversed_src, _array_literal_expr_47
                                        connect _bundle_literal_expr_157.imm_sign, _match_arm_value_151.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_54: Ty51
                                        connect _cast_bundle_to_bits_expr_54.imm_low, _bundle_literal_expr_157.imm_low
                                        connect _cast_bundle_to_bits_expr_54.reversed_src, _bundle_literal_expr_157.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_54.imm_sign, asUInt(_bundle_literal_expr_157.imm_sign)
                                        wire _cast_to_bits_expr_54: UInt<34>
                                        connect _cast_to_bits_expr_54, cat(_cast_bundle_to_bits_expr_54.imm_sign, cat(_cast_bundle_to_bits_expr_54.reversed_src, _cast_bundle_to_bits_expr_54.imm_low))
                                        wire _cast_bits_to_bundle_expr_6: Ty50
                                        wire _cast_bits_to_bundle_expr_flattened_6: Ty51
                                        connect _cast_bits_to_bundle_expr_flattened_6.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_54)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_6.imm_low, _cast_bits_to_bundle_expr_flattened_6.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_6.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_54)), 32, 25)
                                        wire _cast_bits_to_array_expr_6: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_6: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_6[0], bits(_cast_bits_to_bundle_expr_flattened_6.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_6[0], _cast_bits_to_array_expr_flattened_6[0]
                                        connect _cast_bits_to_bundle_expr_6.reversed_src, _cast_bits_to_array_expr_6
                                        connect _cast_bits_to_bundle_expr_flattened_6.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_54)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_6.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_6.imm_sign)
                                        connect _array_literal_expr_44[2], tail(_cast_bits_to_bundle_expr_6.reversed_src[0], 2)
                                        connect _bundle_literal_expr_156.src, _array_literal_expr_44
                                        wire _bundle_literal_expr_158: Ty50
                                        connect _bundle_literal_expr_158.imm_low, _match_arm_value_151.alu_common.common.imm_low
                                        wire _array_literal_expr_48: UInt<8>[1]
                                        connect _array_literal_expr_48[0], _match_arm_value_151.alu_common.common.src[2]
                                        connect _bundle_literal_expr_158.reversed_src, _array_literal_expr_48
                                        connect _bundle_literal_expr_158.imm_sign, _match_arm_value_151.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_55: Ty51
                                        connect _cast_bundle_to_bits_expr_55.imm_low, _bundle_literal_expr_158.imm_low
                                        connect _cast_bundle_to_bits_expr_55.reversed_src, _bundle_literal_expr_158.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_55.imm_sign, asUInt(_bundle_literal_expr_158.imm_sign)
                                        wire _cast_to_bits_expr_55: UInt<34>
                                        connect _cast_to_bits_expr_55, cat(_cast_bundle_to_bits_expr_55.imm_sign, cat(_cast_bundle_to_bits_expr_55.reversed_src, _cast_bundle_to_bits_expr_55.imm_low))
                                        connect _bundle_literal_expr_156.imm_low, bits(asSInt(_cast_to_bits_expr_55), 24, 0)
                                        wire _bundle_literal_expr_159: Ty50
                                        connect _bundle_literal_expr_159.imm_low, _match_arm_value_151.alu_common.common.imm_low
                                        wire _array_literal_expr_49: UInt<8>[1]
                                        connect _array_literal_expr_49[0], _match_arm_value_151.alu_common.common.src[2]
                                        connect _bundle_literal_expr_159.reversed_src, _array_literal_expr_49
                                        connect _bundle_literal_expr_159.imm_sign, _match_arm_value_151.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_56: Ty51
                                        connect _cast_bundle_to_bits_expr_56.imm_low, _bundle_literal_expr_159.imm_low
                                        connect _cast_bundle_to_bits_expr_56.reversed_src, _bundle_literal_expr_159.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_56.imm_sign, asUInt(_bundle_literal_expr_159.imm_sign)
                                        wire _cast_to_bits_expr_56: UInt<34>
                                        connect _cast_to_bits_expr_56, cat(_cast_bundle_to_bits_expr_56.imm_sign, cat(_cast_bundle_to_bits_expr_56.reversed_src, _cast_bundle_to_bits_expr_56.imm_low))
                                        connect _bundle_literal_expr_156.imm_sign, shr(asSInt(_cast_to_bits_expr_56), 33)
                                        wire _bundle_literal_expr_160: Ty2
                                        invalidate _bundle_literal_expr_160
                                        connect _bundle_literal_expr_156._phantom, _bundle_literal_expr_160
                                        connect _bundle_literal_expr_155.common, _bundle_literal_expr_156
                                        connect _bundle_literal_expr_155.output_integer_mode, _match_arm_value_151.alu_common.output_integer_mode
                                        connect _bundle_literal_expr_154.alu_common, _bundle_literal_expr_155
                                        connect _bundle_literal_expr_154.invert_src0, _match_arm_value_151.invert_src0
                                        connect _bundle_literal_expr_154.invert_carry_in, _match_arm_value_151.invert_carry_in
                                        connect _bundle_literal_expr_154.invert_carry_out, _match_arm_value_151.invert_carry_out
                                        connect _bundle_literal_expr_154.add_pc, _match_arm_value_151.add_pc
                                        connect mapped_regs_5, {|AddSub: Ty35, AddSubI: Ty35, Logical: Ty36|}(AddSubI, _bundle_literal_expr_154) @[instruction.rs 477:1]
                                    Logical(_match_arm_value_156):
                                        wire renamed_src_reg_1_0_2: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_161: Ty1
                                        connect _bundle_literal_expr_161.value, _match_arm_value_156.alu_common.common.src[0]
                                        connect rename_table_normal_1_src_0.addr, _bundle_literal_expr_161 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_162: Ty1
                                        connect _bundle_literal_expr_162.value, _match_arm_value_156.alu_common.common.src[0]
                                        connect rename_table_special_1_src_0.addr, _bundle_literal_expr_162 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_1_src_0.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_1_src_0.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_163: Ty25
                                                        wire _bundle_literal_expr_164: Ty23
                                                        connect _bundle_literal_expr_164.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_163.unit_num, _bundle_literal_expr_164
                                                        wire _bundle_literal_expr_165: Ty24
                                                        connect _bundle_literal_expr_165.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_163.unit_out_reg, _bundle_literal_expr_165
                                                        connect renamed_src_reg_1_0_2, _bundle_literal_expr_163 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_157):
                                                        connect renamed_src_reg_1_0_2, _match_arm_value_157 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_158):
                                                connect renamed_src_reg_1_0_2, _match_arm_value_158 @[reg_alloc.rs 185:33]
                                        wire renamed_src_reg_1_1_2: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_166: Ty1
                                        connect _bundle_literal_expr_166.value, _match_arm_value_156.alu_common.common.src[1]
                                        connect rename_table_normal_1_src_1.addr, _bundle_literal_expr_166 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_167: Ty1
                                        connect _bundle_literal_expr_167.value, _match_arm_value_156.alu_common.common.src[1]
                                        connect rename_table_special_1_src_1.addr, _bundle_literal_expr_167 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_1_src_1.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_1_src_1.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_168: Ty25
                                                        wire _bundle_literal_expr_169: Ty23
                                                        connect _bundle_literal_expr_169.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_168.unit_num, _bundle_literal_expr_169
                                                        wire _bundle_literal_expr_170: Ty24
                                                        connect _bundle_literal_expr_170.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_168.unit_out_reg, _bundle_literal_expr_170
                                                        connect renamed_src_reg_1_1_2, _bundle_literal_expr_168 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_159):
                                                        connect renamed_src_reg_1_1_2, _match_arm_value_159 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_160):
                                                connect renamed_src_reg_1_1_2, _match_arm_value_160 @[reg_alloc.rs 185:33]
                                        wire _bundle_literal_expr_171: Ty36
                                        wire _bundle_literal_expr_172: Ty34
                                        wire _bundle_literal_expr_173: Ty33
                                        connect _bundle_literal_expr_173.prefix_pad, _match_arm_value_156.alu_common.common.prefix_pad
                                        connect _bundle_literal_expr_173.dest, _match_arm_value_132.unit_out_reg
                                        wire _array_literal_expr_50: UInt<6>[3]
                                        wire _array_literal_expr_51: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_57: Ty47
                                        connect _cast_bundle_to_bits_expr_57.unit_num, renamed_src_reg_1_0_2.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_57.unit_out_reg, renamed_src_reg_1_0_2.unit_out_reg.value
                                        wire _cast_to_bits_expr_57: UInt<6>
                                        connect _cast_to_bits_expr_57, cat(_cast_bundle_to_bits_expr_57.unit_out_reg, _cast_bundle_to_bits_expr_57.unit_num)
                                        connect _array_literal_expr_51[0], _cast_to_bits_expr_57
                                        wire _cast_bundle_to_bits_expr_58: Ty47
                                        connect _cast_bundle_to_bits_expr_58.unit_num, renamed_src_reg_1_1_2.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_58.unit_out_reg, renamed_src_reg_1_1_2.unit_out_reg.value
                                        wire _cast_to_bits_expr_58: UInt<6>
                                        connect _cast_to_bits_expr_58, cat(_cast_bundle_to_bits_expr_58.unit_out_reg, _cast_bundle_to_bits_expr_58.unit_num)
                                        connect _array_literal_expr_51[1], _cast_to_bits_expr_58
                                        connect _array_literal_expr_50[0], _array_literal_expr_51[0]
                                        wire _array_literal_expr_52: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_59: Ty47
                                        connect _cast_bundle_to_bits_expr_59.unit_num, renamed_src_reg_1_0_2.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_59.unit_out_reg, renamed_src_reg_1_0_2.unit_out_reg.value
                                        wire _cast_to_bits_expr_59: UInt<6>
                                        connect _cast_to_bits_expr_59, cat(_cast_bundle_to_bits_expr_59.unit_out_reg, _cast_bundle_to_bits_expr_59.unit_num)
                                        connect _array_literal_expr_52[0], _cast_to_bits_expr_59
                                        wire _cast_bundle_to_bits_expr_60: Ty47
                                        connect _cast_bundle_to_bits_expr_60.unit_num, renamed_src_reg_1_1_2.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_60.unit_out_reg, renamed_src_reg_1_1_2.unit_out_reg.value
                                        wire _cast_to_bits_expr_60: UInt<6>
                                        connect _cast_to_bits_expr_60, cat(_cast_bundle_to_bits_expr_60.unit_out_reg, _cast_bundle_to_bits_expr_60.unit_num)
                                        connect _array_literal_expr_52[1], _cast_to_bits_expr_60
                                        connect _array_literal_expr_50[1], _array_literal_expr_52[1]
                                        wire _bundle_literal_expr_174: Ty50
                                        connect _bundle_literal_expr_174.imm_low, _match_arm_value_156.alu_common.common.imm_low
                                        wire _array_literal_expr_53: UInt<8>[1]
                                        connect _array_literal_expr_53[0], _match_arm_value_156.alu_common.common.src[2]
                                        connect _bundle_literal_expr_174.reversed_src, _array_literal_expr_53
                                        connect _bundle_literal_expr_174.imm_sign, _match_arm_value_156.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_61: Ty51
                                        connect _cast_bundle_to_bits_expr_61.imm_low, _bundle_literal_expr_174.imm_low
                                        connect _cast_bundle_to_bits_expr_61.reversed_src, _bundle_literal_expr_174.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_61.imm_sign, asUInt(_bundle_literal_expr_174.imm_sign)
                                        wire _cast_to_bits_expr_61: UInt<34>
                                        connect _cast_to_bits_expr_61, cat(_cast_bundle_to_bits_expr_61.imm_sign, cat(_cast_bundle_to_bits_expr_61.reversed_src, _cast_bundle_to_bits_expr_61.imm_low))
                                        wire _cast_bits_to_bundle_expr_7: Ty50
                                        wire _cast_bits_to_bundle_expr_flattened_7: Ty51
                                        connect _cast_bits_to_bundle_expr_flattened_7.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_61)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_7.imm_low, _cast_bits_to_bundle_expr_flattened_7.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_7.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_61)), 32, 25)
                                        wire _cast_bits_to_array_expr_7: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_7: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_7[0], bits(_cast_bits_to_bundle_expr_flattened_7.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_7[0], _cast_bits_to_array_expr_flattened_7[0]
                                        connect _cast_bits_to_bundle_expr_7.reversed_src, _cast_bits_to_array_expr_7
                                        connect _cast_bits_to_bundle_expr_flattened_7.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_61)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_7.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_7.imm_sign)
                                        connect _array_literal_expr_50[2], tail(_cast_bits_to_bundle_expr_7.reversed_src[0], 2)
                                        connect _bundle_literal_expr_173.src, _array_literal_expr_50
                                        wire _bundle_literal_expr_175: Ty50
                                        connect _bundle_literal_expr_175.imm_low, _match_arm_value_156.alu_common.common.imm_low
                                        wire _array_literal_expr_54: UInt<8>[1]
                                        connect _array_literal_expr_54[0], _match_arm_value_156.alu_common.common.src[2]
                                        connect _bundle_literal_expr_175.reversed_src, _array_literal_expr_54
                                        connect _bundle_literal_expr_175.imm_sign, _match_arm_value_156.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_62: Ty51
                                        connect _cast_bundle_to_bits_expr_62.imm_low, _bundle_literal_expr_175.imm_low
                                        connect _cast_bundle_to_bits_expr_62.reversed_src, _bundle_literal_expr_175.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_62.imm_sign, asUInt(_bundle_literal_expr_175.imm_sign)
                                        wire _cast_to_bits_expr_62: UInt<34>
                                        connect _cast_to_bits_expr_62, cat(_cast_bundle_to_bits_expr_62.imm_sign, cat(_cast_bundle_to_bits_expr_62.reversed_src, _cast_bundle_to_bits_expr_62.imm_low))
                                        connect _bundle_literal_expr_173.imm_low, bits(asSInt(_cast_to_bits_expr_62), 24, 0)
                                        wire _bundle_literal_expr_176: Ty50
                                        connect _bundle_literal_expr_176.imm_low, _match_arm_value_156.alu_common.common.imm_low
                                        wire _array_literal_expr_55: UInt<8>[1]
                                        connect _array_literal_expr_55[0], _match_arm_value_156.alu_common.common.src[2]
                                        connect _bundle_literal_expr_176.reversed_src, _array_literal_expr_55
                                        connect _bundle_literal_expr_176.imm_sign, _match_arm_value_156.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_63: Ty51
                                        connect _cast_bundle_to_bits_expr_63.imm_low, _bundle_literal_expr_176.imm_low
                                        connect _cast_bundle_to_bits_expr_63.reversed_src, _bundle_literal_expr_176.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_63.imm_sign, asUInt(_bundle_literal_expr_176.imm_sign)
                                        wire _cast_to_bits_expr_63: UInt<34>
                                        connect _cast_to_bits_expr_63, cat(_cast_bundle_to_bits_expr_63.imm_sign, cat(_cast_bundle_to_bits_expr_63.reversed_src, _cast_bundle_to_bits_expr_63.imm_low))
                                        connect _bundle_literal_expr_173.imm_sign, shr(asSInt(_cast_to_bits_expr_63), 33)
                                        wire _bundle_literal_expr_177: Ty2
                                        invalidate _bundle_literal_expr_177
                                        connect _bundle_literal_expr_173._phantom, _bundle_literal_expr_177
                                        connect _bundle_literal_expr_172.common, _bundle_literal_expr_173
                                        connect _bundle_literal_expr_172.output_integer_mode, _match_arm_value_156.alu_common.output_integer_mode
                                        connect _bundle_literal_expr_171.alu_common, _bundle_literal_expr_172
                                        connect _bundle_literal_expr_171.lut, _match_arm_value_156.lut
                                        connect mapped_regs_5, {|AddSub: Ty35, AddSubI: Ty35, Logical: Ty36|}(Logical, _bundle_literal_expr_171) @[instruction.rs 477:1]
                                connect mapped_regs_4, {|AluBranch: Ty37, L2RegisterFile: Ty40, LoadStore: Ty41|}(AluBranch, mapped_regs_5) @[unit.rs 127:1]
                            L2RegisterFile(_match_arm_value_161):
                                wire mapped_regs_6: Ty40 @[instruction.rs 504:1]
                                match _match_arm_value_161: @[instruction.rs 504:1]
                                    ReadL2Reg(_match_arm_value_162):
                                        wire _bundle_literal_expr_178: Ty39
                                        wire _bundle_literal_expr_179: Ty38
                                        connect _bundle_literal_expr_179.prefix_pad, _match_arm_value_162.common.prefix_pad
                                        connect _bundle_literal_expr_179.dest, _match_arm_value_132.unit_out_reg
                                        wire _array_literal_expr_56: UInt<6>[3]
                                        connect _array_literal_expr_56[0], pad(UInt<0>(0h0), 6)
                                        connect _array_literal_expr_56[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_180: Ty50
                                        connect _bundle_literal_expr_180.imm_low, _match_arm_value_162.common.imm_low
                                        wire _array_literal_expr_57: UInt<8>[1]
                                        connect _array_literal_expr_57[0], _match_arm_value_162.common.src[2]
                                        connect _bundle_literal_expr_180.reversed_src, _array_literal_expr_57
                                        connect _bundle_literal_expr_180.imm_sign, _match_arm_value_162.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_64: Ty51
                                        connect _cast_bundle_to_bits_expr_64.imm_low, _bundle_literal_expr_180.imm_low
                                        connect _cast_bundle_to_bits_expr_64.reversed_src, _bundle_literal_expr_180.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_64.imm_sign, asUInt(_bundle_literal_expr_180.imm_sign)
                                        wire _cast_to_bits_expr_64: UInt<34>
                                        connect _cast_to_bits_expr_64, cat(_cast_bundle_to_bits_expr_64.imm_sign, cat(_cast_bundle_to_bits_expr_64.reversed_src, _cast_bundle_to_bits_expr_64.imm_low))
                                        wire _cast_bits_to_bundle_expr_8: Ty50
                                        wire _cast_bits_to_bundle_expr_flattened_8: Ty51
                                        connect _cast_bits_to_bundle_expr_flattened_8.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_64)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_8.imm_low, _cast_bits_to_bundle_expr_flattened_8.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_8.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_64)), 32, 25)
                                        wire _cast_bits_to_array_expr_8: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_8: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_8[0], bits(_cast_bits_to_bundle_expr_flattened_8.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_8[0], _cast_bits_to_array_expr_flattened_8[0]
                                        connect _cast_bits_to_bundle_expr_8.reversed_src, _cast_bits_to_array_expr_8
                                        connect _cast_bits_to_bundle_expr_flattened_8.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_64)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_8.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_8.imm_sign)
                                        connect _array_literal_expr_56[2], tail(_cast_bits_to_bundle_expr_8.reversed_src[0], 2)
                                        connect _bundle_literal_expr_179.src, _array_literal_expr_56
                                        wire _bundle_literal_expr_181: Ty50
                                        connect _bundle_literal_expr_181.imm_low, _match_arm_value_162.common.imm_low
                                        wire _array_literal_expr_58: UInt<8>[1]
                                        connect _array_literal_expr_58[0], _match_arm_value_162.common.src[2]
                                        connect _bundle_literal_expr_181.reversed_src, _array_literal_expr_58
                                        connect _bundle_literal_expr_181.imm_sign, _match_arm_value_162.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_65: Ty51
                                        connect _cast_bundle_to_bits_expr_65.imm_low, _bundle_literal_expr_181.imm_low
                                        connect _cast_bundle_to_bits_expr_65.reversed_src, _bundle_literal_expr_181.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_65.imm_sign, asUInt(_bundle_literal_expr_181.imm_sign)
                                        wire _cast_to_bits_expr_65: UInt<34>
                                        connect _cast_to_bits_expr_65, cat(_cast_bundle_to_bits_expr_65.imm_sign, cat(_cast_bundle_to_bits_expr_65.reversed_src, _cast_bundle_to_bits_expr_65.imm_low))
                                        connect _bundle_literal_expr_179.imm_low, bits(asSInt(_cast_to_bits_expr_65), 24, 0)
                                        wire _bundle_literal_expr_182: Ty50
                                        connect _bundle_literal_expr_182.imm_low, _match_arm_value_162.common.imm_low
                                        wire _array_literal_expr_59: UInt<8>[1]
                                        connect _array_literal_expr_59[0], _match_arm_value_162.common.src[2]
                                        connect _bundle_literal_expr_182.reversed_src, _array_literal_expr_59
                                        connect _bundle_literal_expr_182.imm_sign, _match_arm_value_162.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_66: Ty51
                                        connect _cast_bundle_to_bits_expr_66.imm_low, _bundle_literal_expr_182.imm_low
                                        connect _cast_bundle_to_bits_expr_66.reversed_src, _bundle_literal_expr_182.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_66.imm_sign, asUInt(_bundle_literal_expr_182.imm_sign)
                                        wire _cast_to_bits_expr_66: UInt<34>
                                        connect _cast_to_bits_expr_66, cat(_cast_bundle_to_bits_expr_66.imm_sign, cat(_cast_bundle_to_bits_expr_66.reversed_src, _cast_bundle_to_bits_expr_66.imm_low))
                                        connect _bundle_literal_expr_179.imm_sign, shr(asSInt(_cast_to_bits_expr_66), 33)
                                        wire _bundle_literal_expr_183: Ty2
                                        invalidate _bundle_literal_expr_183
                                        connect _bundle_literal_expr_179._phantom, _bundle_literal_expr_183
                                        connect _bundle_literal_expr_178.common, _bundle_literal_expr_179
                                        connect mapped_regs_6, {|ReadL2Reg: Ty39, WriteL2Reg: Ty39|}(ReadL2Reg, _bundle_literal_expr_178) @[instruction.rs 504:1]
                                    WriteL2Reg(_match_arm_value_163):
                                        wire renamed_src_reg_1_0_3: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_184: Ty1
                                        connect _bundle_literal_expr_184.value, _match_arm_value_163.common.src[0]
                                        connect rename_table_normal_1_src_0.addr, _bundle_literal_expr_184 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_185: Ty1
                                        connect _bundle_literal_expr_185.value, _match_arm_value_163.common.src[0]
                                        connect rename_table_special_1_src_0.addr, _bundle_literal_expr_185 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_1_src_0.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_1_src_0.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_186: Ty25
                                                        wire _bundle_literal_expr_187: Ty23
                                                        connect _bundle_literal_expr_187.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_186.unit_num, _bundle_literal_expr_187
                                                        wire _bundle_literal_expr_188: Ty24
                                                        connect _bundle_literal_expr_188.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_186.unit_out_reg, _bundle_literal_expr_188
                                                        connect renamed_src_reg_1_0_3, _bundle_literal_expr_186 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_164):
                                                        connect renamed_src_reg_1_0_3, _match_arm_value_164 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_165):
                                                connect renamed_src_reg_1_0_3, _match_arm_value_165 @[reg_alloc.rs 185:33]
                                        wire _bundle_literal_expr_189: Ty39
                                        wire _bundle_literal_expr_190: Ty38
                                        connect _bundle_literal_expr_190.prefix_pad, _match_arm_value_163.common.prefix_pad
                                        connect _bundle_literal_expr_190.dest, _match_arm_value_132.unit_out_reg
                                        wire _array_literal_expr_60: UInt<6>[3]
                                        wire _array_literal_expr_61: UInt<6>[1]
                                        wire _cast_bundle_to_bits_expr_67: Ty47
                                        connect _cast_bundle_to_bits_expr_67.unit_num, renamed_src_reg_1_0_3.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_67.unit_out_reg, renamed_src_reg_1_0_3.unit_out_reg.value
                                        wire _cast_to_bits_expr_67: UInt<6>
                                        connect _cast_to_bits_expr_67, cat(_cast_bundle_to_bits_expr_67.unit_out_reg, _cast_bundle_to_bits_expr_67.unit_num)
                                        connect _array_literal_expr_61[0], _cast_to_bits_expr_67
                                        connect _array_literal_expr_60[0], _array_literal_expr_61[0]
                                        connect _array_literal_expr_60[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_191: Ty50
                                        connect _bundle_literal_expr_191.imm_low, _match_arm_value_163.common.imm_low
                                        wire _array_literal_expr_62: UInt<8>[1]
                                        connect _array_literal_expr_62[0], _match_arm_value_163.common.src[2]
                                        connect _bundle_literal_expr_191.reversed_src, _array_literal_expr_62
                                        connect _bundle_literal_expr_191.imm_sign, _match_arm_value_163.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_68: Ty51
                                        connect _cast_bundle_to_bits_expr_68.imm_low, _bundle_literal_expr_191.imm_low
                                        connect _cast_bundle_to_bits_expr_68.reversed_src, _bundle_literal_expr_191.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_68.imm_sign, asUInt(_bundle_literal_expr_191.imm_sign)
                                        wire _cast_to_bits_expr_68: UInt<34>
                                        connect _cast_to_bits_expr_68, cat(_cast_bundle_to_bits_expr_68.imm_sign, cat(_cast_bundle_to_bits_expr_68.reversed_src, _cast_bundle_to_bits_expr_68.imm_low))
                                        wire _cast_bits_to_bundle_expr_9: Ty50
                                        wire _cast_bits_to_bundle_expr_flattened_9: Ty51
                                        connect _cast_bits_to_bundle_expr_flattened_9.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_68)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_9.imm_low, _cast_bits_to_bundle_expr_flattened_9.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_9.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_68)), 32, 25)
                                        wire _cast_bits_to_array_expr_9: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_9: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_9[0], bits(_cast_bits_to_bundle_expr_flattened_9.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_9[0], _cast_bits_to_array_expr_flattened_9[0]
                                        connect _cast_bits_to_bundle_expr_9.reversed_src, _cast_bits_to_array_expr_9
                                        connect _cast_bits_to_bundle_expr_flattened_9.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_68)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_9.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_9.imm_sign)
                                        connect _array_literal_expr_60[2], tail(_cast_bits_to_bundle_expr_9.reversed_src[0], 2)
                                        connect _bundle_literal_expr_190.src, _array_literal_expr_60
                                        wire _bundle_literal_expr_192: Ty50
                                        connect _bundle_literal_expr_192.imm_low, _match_arm_value_163.common.imm_low
                                        wire _array_literal_expr_63: UInt<8>[1]
                                        connect _array_literal_expr_63[0], _match_arm_value_163.common.src[2]
                                        connect _bundle_literal_expr_192.reversed_src, _array_literal_expr_63
                                        connect _bundle_literal_expr_192.imm_sign, _match_arm_value_163.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_69: Ty51
                                        connect _cast_bundle_to_bits_expr_69.imm_low, _bundle_literal_expr_192.imm_low
                                        connect _cast_bundle_to_bits_expr_69.reversed_src, _bundle_literal_expr_192.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_69.imm_sign, asUInt(_bundle_literal_expr_192.imm_sign)
                                        wire _cast_to_bits_expr_69: UInt<34>
                                        connect _cast_to_bits_expr_69, cat(_cast_bundle_to_bits_expr_69.imm_sign, cat(_cast_bundle_to_bits_expr_69.reversed_src, _cast_bundle_to_bits_expr_69.imm_low))
                                        connect _bundle_literal_expr_190.imm_low, bits(asSInt(_cast_to_bits_expr_69), 24, 0)
                                        wire _bundle_literal_expr_193: Ty50
                                        connect _bundle_literal_expr_193.imm_low, _match_arm_value_163.common.imm_low
                                        wire _array_literal_expr_64: UInt<8>[1]
                                        connect _array_literal_expr_64[0], _match_arm_value_163.common.src[2]
                                        connect _bundle_literal_expr_193.reversed_src, _array_literal_expr_64
                                        connect _bundle_literal_expr_193.imm_sign, _match_arm_value_163.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_70: Ty51
                                        connect _cast_bundle_to_bits_expr_70.imm_low, _bundle_literal_expr_193.imm_low
                                        connect _cast_bundle_to_bits_expr_70.reversed_src, _bundle_literal_expr_193.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_70.imm_sign, asUInt(_bundle_literal_expr_193.imm_sign)
                                        wire _cast_to_bits_expr_70: UInt<34>
                                        connect _cast_to_bits_expr_70, cat(_cast_bundle_to_bits_expr_70.imm_sign, cat(_cast_bundle_to_bits_expr_70.reversed_src, _cast_bundle_to_bits_expr_70.imm_low))
                                        connect _bundle_literal_expr_190.imm_sign, shr(asSInt(_cast_to_bits_expr_70), 33)
                                        wire _bundle_literal_expr_194: Ty2
                                        invalidate _bundle_literal_expr_194
                                        connect _bundle_literal_expr_190._phantom, _bundle_literal_expr_194
                                        connect _bundle_literal_expr_189.common, _bundle_literal_expr_190
                                        connect mapped_regs_6, {|ReadL2Reg: Ty39, WriteL2Reg: Ty39|}(WriteL2Reg, _bundle_literal_expr_189) @[instruction.rs 504:1]
                                connect mapped_regs_4, {|AluBranch: Ty37, L2RegisterFile: Ty40, LoadStore: Ty41|}(L2RegisterFile, mapped_regs_6) @[unit.rs 127:1]
                            LoadStore(_match_arm_value_166):
                                wire mapped_regs_7: Ty41 @[instruction.rs 539:1]
                                match _match_arm_value_166: @[instruction.rs 539:1]
                                    Load(_match_arm_value_167):
                                        wire _bundle_literal_expr_195: Ty38
                                        connect _bundle_literal_expr_195.prefix_pad, _match_arm_value_167.prefix_pad
                                        connect _bundle_literal_expr_195.dest, _match_arm_value_132.unit_out_reg
                                        wire _array_literal_expr_65: UInt<6>[3]
                                        connect _array_literal_expr_65[0], pad(UInt<0>(0h0), 6)
                                        connect _array_literal_expr_65[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_196: Ty50
                                        connect _bundle_literal_expr_196.imm_low, _match_arm_value_167.imm_low
                                        wire _array_literal_expr_66: UInt<8>[1]
                                        connect _array_literal_expr_66[0], _match_arm_value_167.src[2]
                                        connect _bundle_literal_expr_196.reversed_src, _array_literal_expr_66
                                        connect _bundle_literal_expr_196.imm_sign, _match_arm_value_167.imm_sign
                                        wire _cast_bundle_to_bits_expr_71: Ty51
                                        connect _cast_bundle_to_bits_expr_71.imm_low, _bundle_literal_expr_196.imm_low
                                        connect _cast_bundle_to_bits_expr_71.reversed_src, _bundle_literal_expr_196.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_71.imm_sign, asUInt(_bundle_literal_expr_196.imm_sign)
                                        wire _cast_to_bits_expr_71: UInt<34>
                                        connect _cast_to_bits_expr_71, cat(_cast_bundle_to_bits_expr_71.imm_sign, cat(_cast_bundle_to_bits_expr_71.reversed_src, _cast_bundle_to_bits_expr_71.imm_low))
                                        wire _cast_bits_to_bundle_expr_10: Ty50
                                        wire _cast_bits_to_bundle_expr_flattened_10: Ty51
                                        connect _cast_bits_to_bundle_expr_flattened_10.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_71)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_10.imm_low, _cast_bits_to_bundle_expr_flattened_10.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_10.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_71)), 32, 25)
                                        wire _cast_bits_to_array_expr_10: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_10: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_10[0], bits(_cast_bits_to_bundle_expr_flattened_10.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_10[0], _cast_bits_to_array_expr_flattened_10[0]
                                        connect _cast_bits_to_bundle_expr_10.reversed_src, _cast_bits_to_array_expr_10
                                        connect _cast_bits_to_bundle_expr_flattened_10.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_71)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_10.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_10.imm_sign)
                                        connect _array_literal_expr_65[2], tail(_cast_bits_to_bundle_expr_10.reversed_src[0], 2)
                                        connect _bundle_literal_expr_195.src, _array_literal_expr_65
                                        wire _bundle_literal_expr_197: Ty50
                                        connect _bundle_literal_expr_197.imm_low, _match_arm_value_167.imm_low
                                        wire _array_literal_expr_67: UInt<8>[1]
                                        connect _array_literal_expr_67[0], _match_arm_value_167.src[2]
                                        connect _bundle_literal_expr_197.reversed_src, _array_literal_expr_67
                                        connect _bundle_literal_expr_197.imm_sign, _match_arm_value_167.imm_sign
                                        wire _cast_bundle_to_bits_expr_72: Ty51
                                        connect _cast_bundle_to_bits_expr_72.imm_low, _bundle_literal_expr_197.imm_low
                                        connect _cast_bundle_to_bits_expr_72.reversed_src, _bundle_literal_expr_197.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_72.imm_sign, asUInt(_bundle_literal_expr_197.imm_sign)
                                        wire _cast_to_bits_expr_72: UInt<34>
                                        connect _cast_to_bits_expr_72, cat(_cast_bundle_to_bits_expr_72.imm_sign, cat(_cast_bundle_to_bits_expr_72.reversed_src, _cast_bundle_to_bits_expr_72.imm_low))
                                        connect _bundle_literal_expr_195.imm_low, bits(asSInt(_cast_to_bits_expr_72), 24, 0)
                                        wire _bundle_literal_expr_198: Ty50
                                        connect _bundle_literal_expr_198.imm_low, _match_arm_value_167.imm_low
                                        wire _array_literal_expr_68: UInt<8>[1]
                                        connect _array_literal_expr_68[0], _match_arm_value_167.src[2]
                                        connect _bundle_literal_expr_198.reversed_src, _array_literal_expr_68
                                        connect _bundle_literal_expr_198.imm_sign, _match_arm_value_167.imm_sign
                                        wire _cast_bundle_to_bits_expr_73: Ty51
                                        connect _cast_bundle_to_bits_expr_73.imm_low, _bundle_literal_expr_198.imm_low
                                        connect _cast_bundle_to_bits_expr_73.reversed_src, _bundle_literal_expr_198.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_73.imm_sign, asUInt(_bundle_literal_expr_198.imm_sign)
                                        wire _cast_to_bits_expr_73: UInt<34>
                                        connect _cast_to_bits_expr_73, cat(_cast_bundle_to_bits_expr_73.imm_sign, cat(_cast_bundle_to_bits_expr_73.reversed_src, _cast_bundle_to_bits_expr_73.imm_low))
                                        connect _bundle_literal_expr_195.imm_sign, shr(asSInt(_cast_to_bits_expr_73), 33)
                                        wire _bundle_literal_expr_199: Ty2
                                        invalidate _bundle_literal_expr_199
                                        connect _bundle_literal_expr_195._phantom, _bundle_literal_expr_199
                                        connect mapped_regs_7, {|Load: Ty38, Store: Ty38|}(Load, _bundle_literal_expr_195) @[instruction.rs 539:1]
                                    Store(_match_arm_value_168):
                                        wire renamed_src_reg_1_0_4: Ty25 @[reg_alloc.rs 178:33]
                                        wire _bundle_literal_expr_200: Ty1
                                        connect _bundle_literal_expr_200.value, _match_arm_value_168.src[0]
                                        connect rename_table_normal_1_src_0.addr, _bundle_literal_expr_200 @[reg_alloc.rs 181:29]
                                        wire _bundle_literal_expr_201: Ty1
                                        connect _bundle_literal_expr_201.value, _match_arm_value_168.src[0]
                                        connect rename_table_special_1_src_0.addr, _bundle_literal_expr_201 @[reg_alloc.rs 182:29]
                                        match rename_table_normal_1_src_0.data: @[reg_alloc.rs 184:29]
                                            HdlNone:
                                                match rename_table_special_1_src_0.data: @[reg_alloc.rs 186:36]
                                                    HdlNone:
                                                        wire _bundle_literal_expr_202: Ty25
                                                        wire _bundle_literal_expr_203: Ty23
                                                        connect _bundle_literal_expr_203.adj_value, tail(UInt<64>(0h0), 62)
                                                        connect _bundle_literal_expr_202.unit_num, _bundle_literal_expr_203
                                                        wire _bundle_literal_expr_204: Ty24
                                                        connect _bundle_literal_expr_204.value, tail(UInt<8>(0h0), 4)
                                                        connect _bundle_literal_expr_202.unit_out_reg, _bundle_literal_expr_204
                                                        connect renamed_src_reg_1_0_4, _bundle_literal_expr_202 @[reg_alloc.rs 191:33]
                                                    HdlSome(_match_arm_value_169):
                                                        connect renamed_src_reg_1_0_4, _match_arm_value_169 @[reg_alloc.rs 189:33]
                                            HdlSome(_match_arm_value_170):
                                                connect renamed_src_reg_1_0_4, _match_arm_value_170 @[reg_alloc.rs 185:33]
                                        wire _bundle_literal_expr_205: Ty38
                                        connect _bundle_literal_expr_205.prefix_pad, _match_arm_value_168.prefix_pad
                                        connect _bundle_literal_expr_205.dest, _match_arm_value_132.unit_out_reg
                                        wire _array_literal_expr_69: UInt<6>[3]
                                        wire _array_literal_expr_70: UInt<6>[1]
                                        wire _cast_bundle_to_bits_expr_74: Ty47
                                        connect _cast_bundle_to_bits_expr_74.unit_num, renamed_src_reg_1_0_4.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_74.unit_out_reg, renamed_src_reg_1_0_4.unit_out_reg.value
                                        wire _cast_to_bits_expr_74: UInt<6>
                                        connect _cast_to_bits_expr_74, cat(_cast_bundle_to_bits_expr_74.unit_out_reg, _cast_bundle_to_bits_expr_74.unit_num)
                                        connect _array_literal_expr_70[0], _cast_to_bits_expr_74
                                        connect _array_literal_expr_69[0], _array_literal_expr_70[0]
                                        connect _array_literal_expr_69[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_206: Ty50
                                        connect _bundle_literal_expr_206.imm_low, _match_arm_value_168.imm_low
                                        wire _array_literal_expr_71: UInt<8>[1]
                                        connect _array_literal_expr_71[0], _match_arm_value_168.src[2]
                                        connect _bundle_literal_expr_206.reversed_src, _array_literal_expr_71
                                        connect _bundle_literal_expr_206.imm_sign, _match_arm_value_168.imm_sign
                                        wire _cast_bundle_to_bits_expr_75: Ty51
                                        connect _cast_bundle_to_bits_expr_75.imm_low, _bundle_literal_expr_206.imm_low
                                        connect _cast_bundle_to_bits_expr_75.reversed_src, _bundle_literal_expr_206.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_75.imm_sign, asUInt(_bundle_literal_expr_206.imm_sign)
                                        wire _cast_to_bits_expr_75: UInt<34>
                                        connect _cast_to_bits_expr_75, cat(_cast_bundle_to_bits_expr_75.imm_sign, cat(_cast_bundle_to_bits_expr_75.reversed_src, _cast_bundle_to_bits_expr_75.imm_low))
                                        wire _cast_bits_to_bundle_expr_11: Ty50
                                        wire _cast_bits_to_bundle_expr_flattened_11: Ty51
                                        connect _cast_bits_to_bundle_expr_flattened_11.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_75)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_11.imm_low, _cast_bits_to_bundle_expr_flattened_11.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_11.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_75)), 32, 25)
                                        wire _cast_bits_to_array_expr_11: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_11: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_11[0], bits(_cast_bits_to_bundle_expr_flattened_11.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_11[0], _cast_bits_to_array_expr_flattened_11[0]
                                        connect _cast_bits_to_bundle_expr_11.reversed_src, _cast_bits_to_array_expr_11
                                        connect _cast_bits_to_bundle_expr_flattened_11.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_75)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_11.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_11.imm_sign)
                                        connect _array_literal_expr_69[2], tail(_cast_bits_to_bundle_expr_11.reversed_src[0], 2)
                                        connect _bundle_literal_expr_205.src, _array_literal_expr_69
                                        wire _bundle_literal_expr_207: Ty50
                                        connect _bundle_literal_expr_207.imm_low, _match_arm_value_168.imm_low
                                        wire _array_literal_expr_72: UInt<8>[1]
                                        connect _array_literal_expr_72[0], _match_arm_value_168.src[2]
                                        connect _bundle_literal_expr_207.reversed_src, _array_literal_expr_72
                                        connect _bundle_literal_expr_207.imm_sign, _match_arm_value_168.imm_sign
                                        wire _cast_bundle_to_bits_expr_76: Ty51
                                        connect _cast_bundle_to_bits_expr_76.imm_low, _bundle_literal_expr_207.imm_low
                                        connect _cast_bundle_to_bits_expr_76.reversed_src, _bundle_literal_expr_207.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_76.imm_sign, asUInt(_bundle_literal_expr_207.imm_sign)
                                        wire _cast_to_bits_expr_76: UInt<34>
                                        connect _cast_to_bits_expr_76, cat(_cast_bundle_to_bits_expr_76.imm_sign, cat(_cast_bundle_to_bits_expr_76.reversed_src, _cast_bundle_to_bits_expr_76.imm_low))
                                        connect _bundle_literal_expr_205.imm_low, bits(asSInt(_cast_to_bits_expr_76), 24, 0)
                                        wire _bundle_literal_expr_208: Ty50
                                        connect _bundle_literal_expr_208.imm_low, _match_arm_value_168.imm_low
                                        wire _array_literal_expr_73: UInt<8>[1]
                                        connect _array_literal_expr_73[0], _match_arm_value_168.src[2]
                                        connect _bundle_literal_expr_208.reversed_src, _array_literal_expr_73
                                        connect _bundle_literal_expr_208.imm_sign, _match_arm_value_168.imm_sign
                                        wire _cast_bundle_to_bits_expr_77: Ty51
                                        connect _cast_bundle_to_bits_expr_77.imm_low, _bundle_literal_expr_208.imm_low
                                        connect _cast_bundle_to_bits_expr_77.reversed_src, _bundle_literal_expr_208.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_77.imm_sign, asUInt(_bundle_literal_expr_208.imm_sign)
                                        wire _cast_to_bits_expr_77: UInt<34>
                                        connect _cast_to_bits_expr_77, cat(_cast_bundle_to_bits_expr_77.imm_sign, cat(_cast_bundle_to_bits_expr_77.reversed_src, _cast_bundle_to_bits_expr_77.imm_low))
                                        connect _bundle_literal_expr_205.imm_sign, shr(asSInt(_cast_to_bits_expr_77), 33)
                                        wire _bundle_literal_expr_209: Ty2
                                        invalidate _bundle_literal_expr_209
                                        connect _bundle_literal_expr_205._phantom, _bundle_literal_expr_209
                                        connect mapped_regs_7, {|Load: Ty38, Store: Ty38|}(Store, _bundle_literal_expr_205) @[instruction.rs 539:1]
                                connect mapped_regs_4, {|AluBranch: Ty37, L2RegisterFile: Ty40, LoadStore: Ty41|}(LoadStore, mapped_regs_7) @[unit.rs 127:1]
                        connect renamed_mops[1], {|HdlNone, HdlSome: Ty42|}(HdlSome, mapped_regs_4) @[reg_alloc.rs 167:17]
        wire selected_unit_index_leaf_1_0: Ty32 @[reg_alloc.rs 209:25]
        connect selected_unit_index_leaf_1_0, {|HdlNone, HdlSome: UInt<2>|}(HdlNone) @[reg_alloc.rs 212:21]
        wire unit_index_1_0: UInt<2> @[reg_alloc.rs 218:25]
        ; connect different types:
        ; lhs: UInt<2>
        ; rhs: UInt<64>
        connect unit_index_1_0, UInt<64>(0h0) @[reg_alloc.rs 221:21]
        when available_units[1][0]: @[reg_alloc.rs 223:21]
            connect selected_unit_index_leaf_1_0, {|HdlNone, HdlSome: UInt<2>|}(HdlSome, unit_index_1_0) @[reg_alloc.rs 224:25]
        wire selected_unit_index_leaf_1_1: Ty32 @[reg_alloc.rs 209:25]
        connect selected_unit_index_leaf_1_1, {|HdlNone, HdlSome: UInt<2>|}(HdlNone) @[reg_alloc.rs 212:21]
        wire unit_index_1_1: UInt<2> @[reg_alloc.rs 218:25]
        ; connect different types:
        ; lhs: UInt<2>
        ; rhs: UInt<64>
        connect unit_index_1_1, UInt<64>(0h1) @[reg_alloc.rs 221:21]
        when available_units[1][1]: @[reg_alloc.rs 223:21]
            connect selected_unit_index_leaf_1_1, {|HdlNone, HdlSome: UInt<2>|}(HdlSome, unit_index_1_1) @[reg_alloc.rs 224:25]
        wire selected_unit_index_node_1_0: Ty32 @[reg_alloc.rs 231:25]
        connect selected_unit_index_node_1_0, selected_unit_index_leaf_1_0 @[reg_alloc.rs 235:21]
        match selected_unit_index_leaf_1_0: @[reg_alloc.rs 237:21]
            HdlNone:
                connect selected_unit_index_node_1_0, selected_unit_index_leaf_1_1 @[reg_alloc.rs 238:25]
            HdlSome(_match_arm_value_171):
                skip
        connect selected_unit_indexes[1], selected_unit_index_node_1_0 @[reg_alloc.rs 201:9]
        match selected_unit_indexes[0]: @[reg_alloc.rs 251:13]
            HdlNone:
                skip
            HdlSome(_match_arm_value_172):
                connect available_units[1][_match_arm_value_172], UInt<1>(0h0) @[reg_alloc.rs 252:17]
        wire _array_literal_expr_74: Ty44[2]
        connect _array_literal_expr_74[0], {|HdlNone, HdlSome: Ty25|}(HdlNone)
        connect _array_literal_expr_74[1], {|HdlNone, HdlSome: Ty25|}(HdlNone)
        connect renamed_mops_out_reg, _array_literal_expr_74 @[reg_alloc.rs 259:5]
        inst unit_0 of alu_branch @[reg_alloc.rs 271:13]
        connect unit_0.cd, cd @[reg_alloc.rs 273:9]
        inst unit_0_free_regs_tracker of unit_free_regs_tracker @[reg_alloc.rs 286:13]
        connect unit_0_free_regs_tracker.cd, cd @[reg_alloc.rs 288:9]
        wire _uninit_expr: Ty55
        invalidate _uninit_expr
        connect unit_0_free_regs_tracker.free_in[0].data, _uninit_expr @[reg_alloc.rs 290:9]
        connect unit_0_free_regs_tracker.alloc_out[0].ready, UInt<1>(0h0) @[reg_alloc.rs 294:9]
        connect unit_0.`input`.data, {|HdlNone, HdlSome: Ty37|}(HdlNone) @[reg_alloc.rs 295:9]
        match unit_0_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 298:13]
            HdlNone:
                connect available_units[0][0], UInt<1>(0h0) @[reg_alloc.rs 300:17]
            HdlSome(_match_arm_value_173):
                skip
        when not(unit_0.`input`.ready): @[reg_alloc.rs 303:13]
            connect available_units[0][0], UInt<1>(0h0) @[reg_alloc.rs 305:17]
        match selected_unit_indexes[0]: @[reg_alloc.rs 308:13]
            HdlNone:
                skip
            HdlSome(_match_arm_value_174):
                when eq(_match_arm_value_174, UInt<64>(0h0)): @[reg_alloc.rs 310:17]
                    wire and_then_out: Ty52 @[reg_alloc.rs 314:25]
                    connect unit_0_free_regs_tracker.alloc_out[0].ready, UInt<1>(0h1) @[reg_alloc.rs 311:21]
                    match renamed_mops[0]: @[reg_alloc.rs 314:25]
                        HdlNone:
                            connect and_then_out, {|HdlNone, HdlSome: Ty37|}(HdlNone) @[reg_alloc.rs 314:25]
                        HdlSome(_match_arm_value_175):
                            wire alu_branch_mop: Ty52 @[unit.rs 127:1]
                            connect alu_branch_mop, {|HdlNone, HdlSome: Ty37|}(HdlNone) @[unit.rs 127:1]
                            match _match_arm_value_175: @[unit.rs 127:1]
                                AluBranch(_match_arm_value_176):
                                    connect alu_branch_mop, {|HdlNone, HdlSome: Ty37|}(HdlSome, _match_arm_value_176) @[unit.rs 127:1]
                                L2RegisterFile(_match_arm_value_177):
                                    skip
                                LoadStore(_match_arm_value_178):
                                    skip
                            connect and_then_out, alu_branch_mop @[reg_alloc.rs 314:25]
                    match and_then_out: @[reg_alloc.rs 313:21]
                        HdlNone:
                            wire _uninit_expr_1: Ty37
                            invalidate _uninit_expr_1
                            connect unit_0.`input`.data, {|HdlNone, HdlSome: Ty37|}(HdlSome, _uninit_expr_1) @[reg_alloc.rs 318:25]
                        HdlSome(_match_arm_value_179):
                            connect unit_0.`input`.data, {|HdlNone, HdlSome: Ty37|}(HdlSome, _match_arm_value_179) @[reg_alloc.rs 316:25]
                    match unit_0_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 325:21]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_180):
                            wire _bundle_literal_expr_210: Ty25
                            wire _bundle_literal_expr_211: Ty23
                            connect _bundle_literal_expr_211.adj_value, tail(UInt<64>(0h1), 62)
                            connect _bundle_literal_expr_210.unit_num, _bundle_literal_expr_211
                            wire _bundle_literal_expr_212: Ty24
                            connect _bundle_literal_expr_212.value, _match_arm_value_180
                            connect _bundle_literal_expr_210.unit_out_reg, _bundle_literal_expr_212
                            connect renamed_mops_out_reg[0], {|HdlNone, HdlSome: Ty25|}(HdlSome, _bundle_literal_expr_210) @[reg_alloc.rs 331:25]
        match unit_0_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 298:13]
            HdlNone:
                connect available_units[1][0], UInt<1>(0h0) @[reg_alloc.rs 300:17]
            HdlSome(_match_arm_value_181):
                skip
        when not(unit_0.`input`.ready): @[reg_alloc.rs 303:13]
            connect available_units[1][0], UInt<1>(0h0) @[reg_alloc.rs 305:17]
        match selected_unit_indexes[1]: @[reg_alloc.rs 308:13]
            HdlNone:
                skip
            HdlSome(_match_arm_value_182):
                when eq(_match_arm_value_182, UInt<64>(0h0)): @[reg_alloc.rs 310:17]
                    wire and_then_out_1: Ty52 @[reg_alloc.rs 314:25]
                    connect unit_0_free_regs_tracker.alloc_out[0].ready, UInt<1>(0h1) @[reg_alloc.rs 311:21]
                    match renamed_mops[1]: @[reg_alloc.rs 314:25]
                        HdlNone:
                            connect and_then_out_1, {|HdlNone, HdlSome: Ty37|}(HdlNone) @[reg_alloc.rs 314:25]
                        HdlSome(_match_arm_value_183):
                            wire alu_branch_mop_1: Ty52 @[unit.rs 127:1]
                            connect alu_branch_mop_1, {|HdlNone, HdlSome: Ty37|}(HdlNone) @[unit.rs 127:1]
                            match _match_arm_value_183: @[unit.rs 127:1]
                                AluBranch(_match_arm_value_184):
                                    connect alu_branch_mop_1, {|HdlNone, HdlSome: Ty37|}(HdlSome, _match_arm_value_184) @[unit.rs 127:1]
                                L2RegisterFile(_match_arm_value_185):
                                    skip
                                LoadStore(_match_arm_value_186):
                                    skip
                            connect and_then_out_1, alu_branch_mop_1 @[reg_alloc.rs 314:25]
                    match and_then_out_1: @[reg_alloc.rs 313:21]
                        HdlNone:
                            wire _uninit_expr_2: Ty37
                            invalidate _uninit_expr_2
                            connect unit_0.`input`.data, {|HdlNone, HdlSome: Ty37|}(HdlSome, _uninit_expr_2) @[reg_alloc.rs 318:25]
                        HdlSome(_match_arm_value_187):
                            connect unit_0.`input`.data, {|HdlNone, HdlSome: Ty37|}(HdlSome, _match_arm_value_187) @[reg_alloc.rs 316:25]
                    match unit_0_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 325:21]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_188):
                            wire _bundle_literal_expr_213: Ty25
                            wire _bundle_literal_expr_214: Ty23
                            connect _bundle_literal_expr_214.adj_value, tail(UInt<64>(0h1), 62)
                            connect _bundle_literal_expr_213.unit_num, _bundle_literal_expr_214
                            wire _bundle_literal_expr_215: Ty24
                            connect _bundle_literal_expr_215.value, _match_arm_value_188
                            connect _bundle_literal_expr_213.unit_out_reg, _bundle_literal_expr_215
                            connect renamed_mops_out_reg[1], {|HdlNone, HdlSome: Ty25|}(HdlSome, _bundle_literal_expr_213) @[reg_alloc.rs 331:25]
        inst unit_1 of alu_branch_1 @[reg_alloc.rs 271:13]
        connect unit_1.cd, cd @[reg_alloc.rs 273:9]
        inst unit_1_free_regs_tracker of unit_free_regs_tracker_1 @[reg_alloc.rs 286:13]
        connect unit_1_free_regs_tracker.cd, cd @[reg_alloc.rs 288:9]
        wire _uninit_expr_3: Ty55
        invalidate _uninit_expr_3
        connect unit_1_free_regs_tracker.free_in[0].data, _uninit_expr_3 @[reg_alloc.rs 290:9]
        connect unit_1_free_regs_tracker.alloc_out[0].ready, UInt<1>(0h0) @[reg_alloc.rs 294:9]
        connect unit_1.`input`.data, {|HdlNone, HdlSome: Ty37|}(HdlNone) @[reg_alloc.rs 295:9]
        match unit_1_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 298:13]
            HdlNone:
                connect available_units[0][1], UInt<1>(0h0) @[reg_alloc.rs 300:17]
            HdlSome(_match_arm_value_189):
                skip
        when not(unit_1.`input`.ready): @[reg_alloc.rs 303:13]
            connect available_units[0][1], UInt<1>(0h0) @[reg_alloc.rs 305:17]
        match selected_unit_indexes[0]: @[reg_alloc.rs 308:13]
            HdlNone:
                skip
            HdlSome(_match_arm_value_190):
                when eq(_match_arm_value_190, UInt<64>(0h1)): @[reg_alloc.rs 310:17]
                    wire and_then_out_2: Ty52 @[reg_alloc.rs 314:25]
                    connect unit_1_free_regs_tracker.alloc_out[0].ready, UInt<1>(0h1) @[reg_alloc.rs 311:21]
                    match renamed_mops[0]: @[reg_alloc.rs 314:25]
                        HdlNone:
                            connect and_then_out_2, {|HdlNone, HdlSome: Ty37|}(HdlNone) @[reg_alloc.rs 314:25]
                        HdlSome(_match_arm_value_191):
                            wire alu_branch_mop_2: Ty52 @[unit.rs 127:1]
                            connect alu_branch_mop_2, {|HdlNone, HdlSome: Ty37|}(HdlNone) @[unit.rs 127:1]
                            match _match_arm_value_191: @[unit.rs 127:1]
                                AluBranch(_match_arm_value_192):
                                    connect alu_branch_mop_2, {|HdlNone, HdlSome: Ty37|}(HdlSome, _match_arm_value_192) @[unit.rs 127:1]
                                L2RegisterFile(_match_arm_value_193):
                                    skip
                                LoadStore(_match_arm_value_194):
                                    skip
                            connect and_then_out_2, alu_branch_mop_2 @[reg_alloc.rs 314:25]
                    match and_then_out_2: @[reg_alloc.rs 313:21]
                        HdlNone:
                            wire _uninit_expr_4: Ty37
                            invalidate _uninit_expr_4
                            connect unit_1.`input`.data, {|HdlNone, HdlSome: Ty37|}(HdlSome, _uninit_expr_4) @[reg_alloc.rs 318:25]
                        HdlSome(_match_arm_value_195):
                            connect unit_1.`input`.data, {|HdlNone, HdlSome: Ty37|}(HdlSome, _match_arm_value_195) @[reg_alloc.rs 316:25]
                    match unit_1_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 325:21]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_196):
                            wire _bundle_literal_expr_216: Ty25
                            wire _bundle_literal_expr_217: Ty23
                            connect _bundle_literal_expr_217.adj_value, tail(UInt<64>(0h2), 62)
                            connect _bundle_literal_expr_216.unit_num, _bundle_literal_expr_217
                            wire _bundle_literal_expr_218: Ty24
                            connect _bundle_literal_expr_218.value, _match_arm_value_196
                            connect _bundle_literal_expr_216.unit_out_reg, _bundle_literal_expr_218
                            connect renamed_mops_out_reg[0], {|HdlNone, HdlSome: Ty25|}(HdlSome, _bundle_literal_expr_216) @[reg_alloc.rs 331:25]
        match unit_1_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 298:13]
            HdlNone:
                connect available_units[1][1], UInt<1>(0h0) @[reg_alloc.rs 300:17]
            HdlSome(_match_arm_value_197):
                skip
        when not(unit_1.`input`.ready): @[reg_alloc.rs 303:13]
            connect available_units[1][1], UInt<1>(0h0) @[reg_alloc.rs 305:17]
        match selected_unit_indexes[1]: @[reg_alloc.rs 308:13]
            HdlNone:
                skip
            HdlSome(_match_arm_value_198):
                when eq(_match_arm_value_198, UInt<64>(0h1)): @[reg_alloc.rs 310:17]
                    wire and_then_out_3: Ty52 @[reg_alloc.rs 314:25]
                    connect unit_1_free_regs_tracker.alloc_out[0].ready, UInt<1>(0h1) @[reg_alloc.rs 311:21]
                    match renamed_mops[1]: @[reg_alloc.rs 314:25]
                        HdlNone:
                            connect and_then_out_3, {|HdlNone, HdlSome: Ty37|}(HdlNone) @[reg_alloc.rs 314:25]
                        HdlSome(_match_arm_value_199):
                            wire alu_branch_mop_3: Ty52 @[unit.rs 127:1]
                            connect alu_branch_mop_3, {|HdlNone, HdlSome: Ty37|}(HdlNone) @[unit.rs 127:1]
                            match _match_arm_value_199: @[unit.rs 127:1]
                                AluBranch(_match_arm_value_200):
                                    connect alu_branch_mop_3, {|HdlNone, HdlSome: Ty37|}(HdlSome, _match_arm_value_200) @[unit.rs 127:1]
                                L2RegisterFile(_match_arm_value_201):
                                    skip
                                LoadStore(_match_arm_value_202):
                                    skip
                            connect and_then_out_3, alu_branch_mop_3 @[reg_alloc.rs 314:25]
                    match and_then_out_3: @[reg_alloc.rs 313:21]
                        HdlNone:
                            wire _uninit_expr_5: Ty37
                            invalidate _uninit_expr_5
                            connect unit_1.`input`.data, {|HdlNone, HdlSome: Ty37|}(HdlSome, _uninit_expr_5) @[reg_alloc.rs 318:25]
                        HdlSome(_match_arm_value_203):
                            connect unit_1.`input`.data, {|HdlNone, HdlSome: Ty37|}(HdlSome, _match_arm_value_203) @[reg_alloc.rs 316:25]
                    match unit_1_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 325:21]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_204):
                            wire _bundle_literal_expr_219: Ty25
                            wire _bundle_literal_expr_220: Ty23
                            connect _bundle_literal_expr_220.adj_value, tail(UInt<64>(0h2), 62)
                            connect _bundle_literal_expr_219.unit_num, _bundle_literal_expr_220
                            wire _bundle_literal_expr_221: Ty24
                            connect _bundle_literal_expr_221.value, _match_arm_value_204
                            connect _bundle_literal_expr_219.unit_out_reg, _bundle_literal_expr_221
                            connect renamed_mops_out_reg[1], {|HdlNone, HdlSome: Ty25|}(HdlSome, _bundle_literal_expr_219) @[reg_alloc.rs 331:25]
    module alu_branch: @[alu_branch.rs 15:1]
        input cd: Ty0 @[alu_branch.rs 18:29]
        input `input`: Ty53 @[alu_branch.rs 21:11]
        connect `input`.ready, UInt<1>(0h1) @[alu_branch.rs 23:5]
    module unit_free_regs_tracker: @[unit_free_regs_tracker.rs 7:1]
        input cd: Ty0 @[unit_free_regs_tracker.rs 14:29]
        input free_in: Ty56[1] @[unit_free_regs_tracker.rs 17:11]
        output alloc_out: Ty56[1] @[unit_free_regs_tracker.rs 20:11]
        wire _array_literal_expr: UInt<1>[16]
        connect _array_literal_expr[0], UInt<1>(0h0)
        connect _array_literal_expr[1], UInt<1>(0h0)
        connect _array_literal_expr[2], UInt<1>(0h0)
        connect _array_literal_expr[3], UInt<1>(0h0)
        connect _array_literal_expr[4], UInt<1>(0h0)
        connect _array_literal_expr[5], UInt<1>(0h0)
        connect _array_literal_expr[6], UInt<1>(0h0)
        connect _array_literal_expr[7], UInt<1>(0h0)
        connect _array_literal_expr[8], UInt<1>(0h0)
        connect _array_literal_expr[9], UInt<1>(0h0)
        connect _array_literal_expr[10], UInt<1>(0h0)
        connect _array_literal_expr[11], UInt<1>(0h0)
        connect _array_literal_expr[12], UInt<1>(0h0)
        connect _array_literal_expr[13], UInt<1>(0h0)
        connect _array_literal_expr[14], UInt<1>(0h0)
        connect _array_literal_expr[15], UInt<1>(0h0)
        regreset allocated_reg: UInt<1>[16], cd.clk, cd.rst, _array_literal_expr @[unit_free_regs_tracker.rs 27:25]
        connect free_in[0].ready, UInt<1>(0h1) @[unit_free_regs_tracker.rs 29:9]
        wire firing_data: Ty55 @[ready_valid.rs 30:27]
        connect firing_data, {|HdlNone, HdlSome: UInt<4>|}(HdlNone) @[ready_valid.rs 31:9]
        when free_in[0].ready: @[ready_valid.rs 33:9]
            connect firing_data, free_in[0].data @[ready_valid.rs 34:13]
        match firing_data: @[unit_free_regs_tracker.rs 31:9]
            HdlNone:
                skip
            HdlSome(_match_arm_value):
                connect allocated_reg[_match_arm_value], UInt<1>(0h0) @[unit_free_regs_tracker.rs 32:13]
        wire reduced_count_0_2: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_2, add(not(allocated_reg[0]), not(allocated_reg[1])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_2: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_2, or(or(neq(add(not(allocated_reg[0]), not(allocated_reg[1])), reduced_count_0_2), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_2: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[0]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_1: UInt<0>[1]
            connect _array_literal_expr_1[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_0_2[0], _array_literal_expr_1[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_2: UInt<0>[1]
            connect _array_literal_expr_2[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_2[0], add(_array_literal_expr_2[sub(UInt<64>(0h0), not(allocated_reg[0]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_2_4: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_2_4, add(not(allocated_reg[2]), not(allocated_reg[3])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_2_4: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_2_4, or(or(neq(add(not(allocated_reg[2]), not(allocated_reg[3])), reduced_count_2_4), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_2_4: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[2]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_3: UInt<0>[1]
            connect _array_literal_expr_3[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_2_4[0], _array_literal_expr_3[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_4: UInt<0>[1]
            connect _array_literal_expr_4[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_2_4[0], add(_array_literal_expr_4[sub(UInt<64>(0h0), not(allocated_reg[2]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_0_4: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_4, add(reduced_count_0_2, reduced_count_2_4) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_4: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_4, or(or(neq(add(reduced_count_0_2, reduced_count_2_4), reduced_count_0_4), reduced_count_overflowed_0_2), reduced_count_overflowed_2_4) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_4: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_0_2, gt(reduced_count_0_2, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_0_4[0], reduced_alloc_nums_0_2[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_4[0], add(reduced_alloc_nums_2_4[sub(UInt<64>(0h0), reduced_count_0_2)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_4_6: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_4_6, add(not(allocated_reg[4]), not(allocated_reg[5])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_4_6: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_4_6, or(or(neq(add(not(allocated_reg[4]), not(allocated_reg[5])), reduced_count_4_6), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_4_6: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[4]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_5: UInt<0>[1]
            connect _array_literal_expr_5[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_4_6[0], _array_literal_expr_5[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_6: UInt<0>[1]
            connect _array_literal_expr_6[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_4_6[0], add(_array_literal_expr_6[sub(UInt<64>(0h0), not(allocated_reg[4]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_6_8: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_6_8, add(not(allocated_reg[6]), not(allocated_reg[7])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_6_8: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_6_8, or(or(neq(add(not(allocated_reg[6]), not(allocated_reg[7])), reduced_count_6_8), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_6_8: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[6]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_7: UInt<0>[1]
            connect _array_literal_expr_7[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_6_8[0], _array_literal_expr_7[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_8: UInt<0>[1]
            connect _array_literal_expr_8[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_6_8[0], add(_array_literal_expr_8[sub(UInt<64>(0h0), not(allocated_reg[6]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_4_8: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_4_8, add(reduced_count_4_6, reduced_count_6_8) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_4_8: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_4_8, or(or(neq(add(reduced_count_4_6, reduced_count_6_8), reduced_count_4_8), reduced_count_overflowed_4_6), reduced_count_overflowed_6_8) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_4_8: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_4_6, gt(reduced_count_4_6, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_4_8[0], reduced_alloc_nums_4_6[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_4_8[0], add(reduced_alloc_nums_6_8[sub(UInt<64>(0h0), reduced_count_4_6)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_0_8: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_8, add(reduced_count_0_4, reduced_count_4_8) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_8: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_8, or(or(neq(add(reduced_count_0_4, reduced_count_4_8), reduced_count_0_8), reduced_count_overflowed_0_4), reduced_count_overflowed_4_8) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_8: UInt<3>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_0_4, gt(reduced_count_0_4, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<2>
            connect reduced_alloc_nums_0_8[0], reduced_alloc_nums_0_4[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_8[0], add(reduced_alloc_nums_4_8[sub(UInt<64>(0h0), reduced_count_0_4)], UInt<64>(0h4)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_8_10: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_8_10, add(not(allocated_reg[8]), not(allocated_reg[9])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_8_10: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_8_10, or(or(neq(add(not(allocated_reg[8]), not(allocated_reg[9])), reduced_count_8_10), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_8_10: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[8]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_9: UInt<0>[1]
            connect _array_literal_expr_9[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_8_10[0], _array_literal_expr_9[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_10: UInt<0>[1]
            connect _array_literal_expr_10[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_8_10[0], add(_array_literal_expr_10[sub(UInt<64>(0h0), not(allocated_reg[8]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_10_12: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_10_12, add(not(allocated_reg[10]), not(allocated_reg[11])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_10_12: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_10_12, or(or(neq(add(not(allocated_reg[10]), not(allocated_reg[11])), reduced_count_10_12), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_10_12: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[10]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_11: UInt<0>[1]
            connect _array_literal_expr_11[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_10_12[0], _array_literal_expr_11[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_12: UInt<0>[1]
            connect _array_literal_expr_12[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_10_12[0], add(_array_literal_expr_12[sub(UInt<64>(0h0), not(allocated_reg[10]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_8_12: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_8_12, add(reduced_count_8_10, reduced_count_10_12) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_8_12: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_8_12, or(or(neq(add(reduced_count_8_10, reduced_count_10_12), reduced_count_8_12), reduced_count_overflowed_8_10), reduced_count_overflowed_10_12) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_8_12: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_8_10, gt(reduced_count_8_10, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_8_12[0], reduced_alloc_nums_8_10[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_8_12[0], add(reduced_alloc_nums_10_12[sub(UInt<64>(0h0), reduced_count_8_10)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_12_14: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_12_14, add(not(allocated_reg[12]), not(allocated_reg[13])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_12_14: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_12_14, or(or(neq(add(not(allocated_reg[12]), not(allocated_reg[13])), reduced_count_12_14), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_12_14: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[12]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_13: UInt<0>[1]
            connect _array_literal_expr_13[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_12_14[0], _array_literal_expr_13[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_14: UInt<0>[1]
            connect _array_literal_expr_14[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_12_14[0], add(_array_literal_expr_14[sub(UInt<64>(0h0), not(allocated_reg[12]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_14_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_14_16, add(not(allocated_reg[14]), not(allocated_reg[15])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_14_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_14_16, or(or(neq(add(not(allocated_reg[14]), not(allocated_reg[15])), reduced_count_14_16), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_14_16: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[14]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_15: UInt<0>[1]
            connect _array_literal_expr_15[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_14_16[0], _array_literal_expr_15[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_16: UInt<0>[1]
            connect _array_literal_expr_16[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_14_16[0], add(_array_literal_expr_16[sub(UInt<64>(0h0), not(allocated_reg[14]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_12_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_12_16, add(reduced_count_12_14, reduced_count_14_16) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_12_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_12_16, or(or(neq(add(reduced_count_12_14, reduced_count_14_16), reduced_count_12_16), reduced_count_overflowed_12_14), reduced_count_overflowed_14_16) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_12_16: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_12_14, gt(reduced_count_12_14, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_12_16[0], reduced_alloc_nums_12_14[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_12_16[0], add(reduced_alloc_nums_14_16[sub(UInt<64>(0h0), reduced_count_12_14)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_8_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_8_16, add(reduced_count_8_12, reduced_count_12_16) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_8_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_8_16, or(or(neq(add(reduced_count_8_12, reduced_count_12_16), reduced_count_8_16), reduced_count_overflowed_8_12), reduced_count_overflowed_12_16) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_8_16: UInt<3>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_8_12, gt(reduced_count_8_12, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<2>
            connect reduced_alloc_nums_8_16[0], reduced_alloc_nums_8_12[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_8_16[0], add(reduced_alloc_nums_12_16[sub(UInt<64>(0h0), reduced_count_8_12)], UInt<64>(0h4)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_0_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_16, add(reduced_count_0_8, reduced_count_8_16) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_16, or(or(neq(add(reduced_count_0_8, reduced_count_8_16), reduced_count_0_16), reduced_count_overflowed_0_8), reduced_count_overflowed_8_16) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_16: UInt<4>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_0_8, gt(reduced_count_0_8, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<4>
            ; rhs: UInt<3>
            connect reduced_alloc_nums_0_16[0], reduced_alloc_nums_0_8[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<4>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_16[0], add(reduced_alloc_nums_8_16[sub(UInt<64>(0h0), reduced_count_0_8)], UInt<64>(0h8)) @[unit_free_regs_tracker.rs 83:21]
        wire firing_data_1: Ty55 @[ready_valid.rs 30:27]
        connect firing_data_1, {|HdlNone, HdlSome: UInt<4>|}(HdlNone) @[ready_valid.rs 31:9]
        when alloc_out[0].ready: @[ready_valid.rs 33:9]
            connect firing_data_1, alloc_out[0].data @[ready_valid.rs 34:13]
        match firing_data_1: @[unit_free_regs_tracker.rs 102:9]
            HdlNone:
                skip
            HdlSome(_match_arm_value_1):
                connect allocated_reg[_match_arm_value_1], UInt<1>(0h1) @[unit_free_regs_tracker.rs 103:13]
        when or(reduced_count_overflowed_0_16, gt(reduced_count_0_16, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 106:9]
            connect alloc_out[0].data, {|HdlNone, HdlSome: UInt<4>|}(HdlSome, reduced_alloc_nums_0_16[0]) @[unit_free_regs_tracker.rs 107:13]
        else:
            connect alloc_out[0].data, {|HdlNone, HdlSome: UInt<4>|}(HdlNone) @[unit_free_regs_tracker.rs 112:13]
    module alu_branch_1: @[alu_branch.rs 15:1]
        input cd: Ty0 @[alu_branch.rs 18:29]
        input `input`: Ty53 @[alu_branch.rs 21:11]
        connect `input`.ready, UInt<1>(0h1) @[alu_branch.rs 23:5]
    module unit_free_regs_tracker_1: @[unit_free_regs_tracker.rs 7:1]
        input cd: Ty0 @[unit_free_regs_tracker.rs 14:29]
        input free_in: Ty56[1] @[unit_free_regs_tracker.rs 17:11]
        output alloc_out: Ty56[1] @[unit_free_regs_tracker.rs 20:11]
        wire _array_literal_expr: UInt<1>[16]
        connect _array_literal_expr[0], UInt<1>(0h0)
        connect _array_literal_expr[1], UInt<1>(0h0)
        connect _array_literal_expr[2], UInt<1>(0h0)
        connect _array_literal_expr[3], UInt<1>(0h0)
        connect _array_literal_expr[4], UInt<1>(0h0)
        connect _array_literal_expr[5], UInt<1>(0h0)
        connect _array_literal_expr[6], UInt<1>(0h0)
        connect _array_literal_expr[7], UInt<1>(0h0)
        connect _array_literal_expr[8], UInt<1>(0h0)
        connect _array_literal_expr[9], UInt<1>(0h0)
        connect _array_literal_expr[10], UInt<1>(0h0)
        connect _array_literal_expr[11], UInt<1>(0h0)
        connect _array_literal_expr[12], UInt<1>(0h0)
        connect _array_literal_expr[13], UInt<1>(0h0)
        connect _array_literal_expr[14], UInt<1>(0h0)
        connect _array_literal_expr[15], UInt<1>(0h0)
        regreset allocated_reg: UInt<1>[16], cd.clk, cd.rst, _array_literal_expr @[unit_free_regs_tracker.rs 27:25]
        connect free_in[0].ready, UInt<1>(0h1) @[unit_free_regs_tracker.rs 29:9]
        wire firing_data: Ty55 @[ready_valid.rs 30:27]
        connect firing_data, {|HdlNone, HdlSome: UInt<4>|}(HdlNone) @[ready_valid.rs 31:9]
        when free_in[0].ready: @[ready_valid.rs 33:9]
            connect firing_data, free_in[0].data @[ready_valid.rs 34:13]
        match firing_data: @[unit_free_regs_tracker.rs 31:9]
            HdlNone:
                skip
            HdlSome(_match_arm_value):
                connect allocated_reg[_match_arm_value], UInt<1>(0h0) @[unit_free_regs_tracker.rs 32:13]
        wire reduced_count_0_2: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_2, add(not(allocated_reg[0]), not(allocated_reg[1])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_2: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_2, or(or(neq(add(not(allocated_reg[0]), not(allocated_reg[1])), reduced_count_0_2), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_2: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[0]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_1: UInt<0>[1]
            connect _array_literal_expr_1[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_0_2[0], _array_literal_expr_1[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_2: UInt<0>[1]
            connect _array_literal_expr_2[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_2[0], add(_array_literal_expr_2[sub(UInt<64>(0h0), not(allocated_reg[0]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_2_4: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_2_4, add(not(allocated_reg[2]), not(allocated_reg[3])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_2_4: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_2_4, or(or(neq(add(not(allocated_reg[2]), not(allocated_reg[3])), reduced_count_2_4), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_2_4: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[2]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_3: UInt<0>[1]
            connect _array_literal_expr_3[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_2_4[0], _array_literal_expr_3[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_4: UInt<0>[1]
            connect _array_literal_expr_4[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_2_4[0], add(_array_literal_expr_4[sub(UInt<64>(0h0), not(allocated_reg[2]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_0_4: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_4, add(reduced_count_0_2, reduced_count_2_4) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_4: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_4, or(or(neq(add(reduced_count_0_2, reduced_count_2_4), reduced_count_0_4), reduced_count_overflowed_0_2), reduced_count_overflowed_2_4) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_4: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_0_2, gt(reduced_count_0_2, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_0_4[0], reduced_alloc_nums_0_2[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_4[0], add(reduced_alloc_nums_2_4[sub(UInt<64>(0h0), reduced_count_0_2)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_4_6: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_4_6, add(not(allocated_reg[4]), not(allocated_reg[5])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_4_6: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_4_6, or(or(neq(add(not(allocated_reg[4]), not(allocated_reg[5])), reduced_count_4_6), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_4_6: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[4]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_5: UInt<0>[1]
            connect _array_literal_expr_5[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_4_6[0], _array_literal_expr_5[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_6: UInt<0>[1]
            connect _array_literal_expr_6[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_4_6[0], add(_array_literal_expr_6[sub(UInt<64>(0h0), not(allocated_reg[4]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_6_8: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_6_8, add(not(allocated_reg[6]), not(allocated_reg[7])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_6_8: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_6_8, or(or(neq(add(not(allocated_reg[6]), not(allocated_reg[7])), reduced_count_6_8), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_6_8: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[6]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_7: UInt<0>[1]
            connect _array_literal_expr_7[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_6_8[0], _array_literal_expr_7[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_8: UInt<0>[1]
            connect _array_literal_expr_8[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_6_8[0], add(_array_literal_expr_8[sub(UInt<64>(0h0), not(allocated_reg[6]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_4_8: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_4_8, add(reduced_count_4_6, reduced_count_6_8) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_4_8: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_4_8, or(or(neq(add(reduced_count_4_6, reduced_count_6_8), reduced_count_4_8), reduced_count_overflowed_4_6), reduced_count_overflowed_6_8) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_4_8: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_4_6, gt(reduced_count_4_6, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_4_8[0], reduced_alloc_nums_4_6[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_4_8[0], add(reduced_alloc_nums_6_8[sub(UInt<64>(0h0), reduced_count_4_6)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_0_8: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_8, add(reduced_count_0_4, reduced_count_4_8) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_8: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_8, or(or(neq(add(reduced_count_0_4, reduced_count_4_8), reduced_count_0_8), reduced_count_overflowed_0_4), reduced_count_overflowed_4_8) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_8: UInt<3>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_0_4, gt(reduced_count_0_4, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<2>
            connect reduced_alloc_nums_0_8[0], reduced_alloc_nums_0_4[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_8[0], add(reduced_alloc_nums_4_8[sub(UInt<64>(0h0), reduced_count_0_4)], UInt<64>(0h4)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_8_10: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_8_10, add(not(allocated_reg[8]), not(allocated_reg[9])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_8_10: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_8_10, or(or(neq(add(not(allocated_reg[8]), not(allocated_reg[9])), reduced_count_8_10), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_8_10: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[8]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_9: UInt<0>[1]
            connect _array_literal_expr_9[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_8_10[0], _array_literal_expr_9[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_10: UInt<0>[1]
            connect _array_literal_expr_10[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_8_10[0], add(_array_literal_expr_10[sub(UInt<64>(0h0), not(allocated_reg[8]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_10_12: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_10_12, add(not(allocated_reg[10]), not(allocated_reg[11])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_10_12: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_10_12, or(or(neq(add(not(allocated_reg[10]), not(allocated_reg[11])), reduced_count_10_12), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_10_12: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[10]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_11: UInt<0>[1]
            connect _array_literal_expr_11[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_10_12[0], _array_literal_expr_11[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_12: UInt<0>[1]
            connect _array_literal_expr_12[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_10_12[0], add(_array_literal_expr_12[sub(UInt<64>(0h0), not(allocated_reg[10]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_8_12: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_8_12, add(reduced_count_8_10, reduced_count_10_12) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_8_12: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_8_12, or(or(neq(add(reduced_count_8_10, reduced_count_10_12), reduced_count_8_12), reduced_count_overflowed_8_10), reduced_count_overflowed_10_12) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_8_12: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_8_10, gt(reduced_count_8_10, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_8_12[0], reduced_alloc_nums_8_10[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_8_12[0], add(reduced_alloc_nums_10_12[sub(UInt<64>(0h0), reduced_count_8_10)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_12_14: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_12_14, add(not(allocated_reg[12]), not(allocated_reg[13])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_12_14: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_12_14, or(or(neq(add(not(allocated_reg[12]), not(allocated_reg[13])), reduced_count_12_14), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_12_14: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[12]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_13: UInt<0>[1]
            connect _array_literal_expr_13[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_12_14[0], _array_literal_expr_13[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_14: UInt<0>[1]
            connect _array_literal_expr_14[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_12_14[0], add(_array_literal_expr_14[sub(UInt<64>(0h0), not(allocated_reg[12]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_14_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_14_16, add(not(allocated_reg[14]), not(allocated_reg[15])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_14_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_14_16, or(or(neq(add(not(allocated_reg[14]), not(allocated_reg[15])), reduced_count_14_16), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_14_16: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[14]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_15: UInt<0>[1]
            connect _array_literal_expr_15[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_14_16[0], _array_literal_expr_15[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_16: UInt<0>[1]
            connect _array_literal_expr_16[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_14_16[0], add(_array_literal_expr_16[sub(UInt<64>(0h0), not(allocated_reg[14]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_12_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_12_16, add(reduced_count_12_14, reduced_count_14_16) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_12_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_12_16, or(or(neq(add(reduced_count_12_14, reduced_count_14_16), reduced_count_12_16), reduced_count_overflowed_12_14), reduced_count_overflowed_14_16) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_12_16: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_12_14, gt(reduced_count_12_14, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_12_16[0], reduced_alloc_nums_12_14[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_12_16[0], add(reduced_alloc_nums_14_16[sub(UInt<64>(0h0), reduced_count_12_14)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_8_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_8_16, add(reduced_count_8_12, reduced_count_12_16) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_8_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_8_16, or(or(neq(add(reduced_count_8_12, reduced_count_12_16), reduced_count_8_16), reduced_count_overflowed_8_12), reduced_count_overflowed_12_16) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_8_16: UInt<3>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_8_12, gt(reduced_count_8_12, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<2>
            connect reduced_alloc_nums_8_16[0], reduced_alloc_nums_8_12[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_8_16[0], add(reduced_alloc_nums_12_16[sub(UInt<64>(0h0), reduced_count_8_12)], UInt<64>(0h4)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_0_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_16, add(reduced_count_0_8, reduced_count_8_16) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_16, or(or(neq(add(reduced_count_0_8, reduced_count_8_16), reduced_count_0_16), reduced_count_overflowed_0_8), reduced_count_overflowed_8_16) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_16: UInt<4>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_0_8, gt(reduced_count_0_8, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<4>
            ; rhs: UInt<3>
            connect reduced_alloc_nums_0_16[0], reduced_alloc_nums_0_8[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<4>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_16[0], add(reduced_alloc_nums_8_16[sub(UInt<64>(0h0), reduced_count_0_8)], UInt<64>(0h8)) @[unit_free_regs_tracker.rs 83:21]
        wire firing_data_1: Ty55 @[ready_valid.rs 30:27]
        connect firing_data_1, {|HdlNone, HdlSome: UInt<4>|}(HdlNone) @[ready_valid.rs 31:9]
        when alloc_out[0].ready: @[ready_valid.rs 33:9]
            connect firing_data_1, alloc_out[0].data @[ready_valid.rs 34:13]
        match firing_data_1: @[unit_free_regs_tracker.rs 102:9]
            HdlNone:
                skip
            HdlSome(_match_arm_value_1):
                connect allocated_reg[_match_arm_value_1], UInt<1>(0h1) @[unit_free_regs_tracker.rs 103:13]
        when or(reduced_count_overflowed_0_16, gt(reduced_count_0_16, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 106:9]
            connect alloc_out[0].data, {|HdlNone, HdlSome: UInt<4>|}(HdlSome, reduced_alloc_nums_0_16[0]) @[unit_free_regs_tracker.rs 107:13]
        else:
            connect alloc_out[0].data, {|HdlNone, HdlSome: UInt<4>|}(HdlNone) @[unit_free_regs_tracker.rs 112:13]
",
    };
    let sim_debug = format!("{sim:#?}");
    println!("#######\n{sim_debug}\n#######");
    if sim_debug != include_str!("expected/reg_alloc.txt") {
        panic!();
    }
}
